@@ -13,9 +13,54 @@
 
 namespace ngbla 
 {
-  // Interface to lapack functions
+
+  class T_Lapack { };
+  static constexpr T_Lapack Lapack;
+
+  template <typename TA>
+  class LapackExpr : public Expr<LapackExpr<TA> >
+  {
+    const TA & a;
+  public:
+    LapackExpr (const TA & aa) : a(aa) { ; }
+    const TA & A() const { return a; }
+    size_t Height() const { return a.Height(); }
+    size_t Width() const { return a.Width(); }
+  };
+  
+  template <typename TA>
+  INLINE LapackExpr<TA> operator| (const Expr<TA> & a, T_Lapack /* tl */)
+  {
+    return LapackExpr<TA> (a.Spec());
+  }
+
+  
+  template <typename TOP, typename T, typename TB>
+  class assign_trait<TOP, T, LapackExpr<TB>, int>
+  {
+  public:
+    static INLINE T & Assign (MatExpr<T> & self, const Expr<LapackExpr<TB>> & v)
+    {
+    #ifdef LAPACK
+      if constexpr (std::is_same_v<TOP,typename MatExpr<T>::As>)
+                     LapackMultAdd (v.Spec().A().A(), v.Spec().A().B(), 1.0, self.Spec(), 0.0);
+      if constexpr (std::is_same_v<TOP,typename MatExpr<T>::AsAdd>)
+                     LapackMultAdd (v.Spec().A().A(), v.Spec().A().B(), 1.0, self.Spec(), 1.0);
+      if constexpr (std::is_same_v<TOP,typename MatExpr<T>::AsSub>)
+                     LapackMultAdd (v.Spec().A().A(), v.Spec().A().B(), -1.0, self.Spec(), 1.0);
+      return self.Spec();
+    #else // LAPACK
+      throw Exception("No Lapack");
+    #endif // LAPACK
+    }
+  };  
 
 
+  
+
+
+
+  
 #ifdef LAPACK
 
   extern "C" {
@@ -45,6 +90,63 @@ namespace ngbla
   }
 
 
+  // Interface to lapack functions
+  NGS_DLL_HEADER int sgemm(char *transa, char *transb, integer *m, integer *
+		  n, integer *k, real *alpha, real *a, integer *lda, 
+		  real *b, integer *ldb, real *beta, real *c__, 
+		  integer *ldc);
+
+  NGS_DLL_HEADER int dgemm(char *transa, char *transb, integer *m, integer *
+		  n, integer *k, doublereal *alpha, doublereal *a, integer *lda, 
+		  doublereal *b, integer *ldb, doublereal *beta, doublereal *c__, 
+		  integer *ldc);
+
+  NGS_DLL_HEADER int sgemm(char *transa, char *transb, integer *m, integer *
+		  n, integer *k, real *alpha, real *a, integer *lda, 
+		  real *b, integer *ldb, real *beta, real *c__, 
+		  integer *ldc);
+
+  NGS_DLL_HEADER int zgemm(char *transa, char *transb, integer *m, integer *
+		    n, integer *k, doublecomplex *alpha, doublecomplex *a, integer *lda, 
+		    doublecomplex *b, integer *ldb, doublecomplex *beta, doublecomplex *
+		    c__, integer *ldc);
+
+  NGS_DLL_HEADER int dger(integer *m, integer *n, doublereal *alpha,
+                   doublereal *x, integer *incx, doublereal *y, integer *incy,
+                   doublereal *a, integer *lda);
+  
+  NGS_DLL_HEADER int dgetrf(integer* n, integer* m, double* a, integer* lda, integer* ipiv, integer* info);
+  NGS_DLL_HEADER int dgetri(integer* n, double* a, integer* lda, integer* ipiv,
+                            double* hwork, integer* lwork, integer* info);
+  NGS_DLL_HEADER int dgetrs(char *trans, integer *n, integer *nrhs, 
+                            doublereal *a, integer *lda, integer *ipiv, doublereal *b, integer *
+                            ldb, integer *info);
+  
+
+  inline int gemm(char *transa, char *transb, integer *m, integer *
+      n, integer *k, real *alpha, real *a, integer *lda,
+      real *b, integer *ldb, real *beta, real *c__,
+      integer *ldc)
+  {
+    return sgemm (transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c__, ldc);
+  }
+
+  inline int gemm(char *transa, char *transb, integer *m, integer *
+      n, integer *k, doublereal *alpha, doublereal *a, integer *lda,
+      doublereal *b, integer *ldb, doublereal *beta, doublereal *c__,
+      integer *ldc)
+  {
+    return dgemm (transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c__, ldc);
+  }
+
+  inline int gemm(char *transa, char *transb, integer *m, integer *
+      n, integer *k, doublecomplex *alpha, doublecomplex *a, integer *lda,
+      doublecomplex *b, integer *ldb, doublecomplex *beta, doublecomplex *
+      c__, integer *ldc)
+  {
+    return zgemm (transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c__, ldc);
+  }
+
   
   // BLAS 1
 
@@ -74,7 +176,7 @@ namespace ngbla
     integer m = a.Width();
     integer n = a.Height();
     double alpha = 1;
-    integer lda = a.Width();
+    integer lda = max(size_t(1), a.Width());
     integer incx = 1;
     double beta = 0;
     integer incy = 1;
@@ -89,7 +191,7 @@ namespace ngbla
     integer m = a.Width();
     integer n = a.Height();
     Complex alpha(1,0);
-    integer lda = a.Width();
+    integer lda = max(size_t(1), a.Width());
     integer incx = 1;
     Complex beta(0, 0);
     integer incy = 1;
@@ -106,7 +208,7 @@ namespace ngbla
     integer m = a.Width();
     integer n = a.Height();
     double alpha = 1;
-    integer lda = a.Width();
+    integer lda = max(size_t(1), a.Width());
     integer incx = 1;
     double beta = 0;
     integer incy = 1;
@@ -114,11 +216,6 @@ namespace ngbla
   }
 
 
-  /*
-  extern "C"
-  void dger_ (int & m, int & n, double & alpha, double & x, int & incx, double & y, int & incy, double & a, int & lda);
-  */
-
   inline void LapackAddxyt (ngbla::FlatMatrix<double> a,
                             double fac,
                             ngbla::FlatVector<double> x,
@@ -127,30 +224,15 @@ namespace ngbla
     integer m = a.Width();
     integer n = a.Height();
     double alpha = fac;
-    integer lda = a.Width();
+    integer lda = max(size_t(1), a.Width());
     integer incx = 1;
     integer incy = 1;
 
-    dger_ (&m, &n, &alpha, &y(0), &incx, &x(0), &incy, &a(0,0), &lda);
+    dger (&m, &n, &alpha, &y(0), &incx, &x(0), &incy, &a(0,0), &lda);
   }
 
 
 
-  inline int gemm(char *transa, char *transb, integer *m, integer *
-		  n, integer *k, doublereal *alpha, doublereal *a, integer *lda, 
-		  doublereal *b, integer *ldb, doublereal *beta, doublereal *c__, 
-		  integer *ldc)
-  {
-    return dgemm_ (transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c__, ldc);
-  }
-
-  inline int gemm(char *transa, char *transb, integer *m, integer *
-		    n, integer *k, doublecomplex *alpha, doublecomplex *a, integer *lda, 
-		    doublecomplex *b, integer *ldb, doublecomplex *beta, doublecomplex *
-		    c__, integer *ldc)
-  {
-    return zgemm_ (transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c__, ldc);
-  }
 
 
   template <typename SCAL>
@@ -167,9 +249,9 @@ namespace ngbla
     integer k = transa ? a.Height() : a.Width();
     SCAL alpha = 1.0;
     SCAL beta = 0;
-    integer lda = a.Dist();
-    integer ldb = b.Dist();
-    integer ldc = c.Dist();
+    integer lda = max(size_t(1), a.Dist());
+    integer ldb = max(size_t(1), b.Dist());
+    integer ldc = max(size_t(1), c.Dist());
 
     gemm (&transb_, &transa_, &n, &m, &k, &alpha, 
 	  &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
@@ -191,9 +273,9 @@ namespace ngbla
     integer k = transa ? a.Height() : a.Width();
     SCAL alpha = aalpha;
     SCAL beta = abeta;
-    integer lda = a.Dist();
-    integer ldb = b.Dist();
-    integer ldc = c.Dist();
+    integer lda = max(size_t(1), a.Dist());
+    integer ldb = max(size_t(1), b.Dist());
+    integer ldc = max(size_t(1), c.Dist());
 
     gemm (&transb_, &transa_, &n, &m, &k, &alpha, 
 	  &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
@@ -280,7 +362,12 @@ namespace ngbla
 
 
 
-
+  inline void LapackMultAdd (SliceMatrix<float> a, 
+			     SliceMatrix<float> b, 
+			     float alpha,
+			     SliceMatrix<float> c,
+			     float beta)
+  { BASE_LapackMultAdd<float> (a, false, b, false, alpha, c, beta); }
 
   inline void LapackMultAdd (SliceMatrix<double> a, 
 			     SliceMatrix<double> b, 
@@ -289,6 +376,30 @@ namespace ngbla
 			     double beta)
   { BASE_LapackMultAdd<double> (a, false, b, false, alpha, c, beta); }
 
+  inline void LapackMultAdd (SliceMatrix<double,ColMajor> a, 
+			     SliceMatrix<double> b, 
+			     double alpha,
+			     SliceMatrix<double> c,
+			     double beta)
+  { BASE_LapackMultAdd<double> (Trans(a), true, b, false, alpha, c, beta); }
+
+  inline void LapackMultAdd (SliceMatrix<double> a, 
+			     SliceMatrix<double,ColMajor> b, 
+			     double alpha,
+			     SliceMatrix<double> c,
+			     double beta)
+  { BASE_LapackMultAdd<double> (a, false, Trans(b), true, alpha, c, beta); }
+
+  inline void LapackMultAdd (SliceMatrix<double,ColMajor> a, 
+			     SliceMatrix<double,ColMajor> b, 
+			     double alpha,
+			     SliceMatrix<double> c,
+			     double beta)
+  { BASE_LapackMultAdd<double> (Trans(a), true, Trans(b), true, alpha, c, beta); }
+
+
+
+  /*
   template <typename TA>
   inline void LapackMultAdd (SliceMatrix<double> a, 
 			     TransExpr<TA> b, 
@@ -312,7 +423,7 @@ namespace ngbla
 			     SliceMatrix<double> c,
 			     double beta = 1.0)
   { BASE_LapackMultAdd<double> (a.A(), true, b.A(), true, alpha, c, beta); }
-
+  */
 
 
   inline void LapackMultAdd (SliceMatrix<Complex> a, 
@@ -322,6 +433,28 @@ namespace ngbla
 			     Complex beta = 1.0)
   { BASE_LapackMultAdd<Complex> (a, false, b, false, alpha, c, beta); }
 
+  inline void LapackMultAdd (SliceMatrix<Complex,ColMajor> a, 
+			     SliceMatrix<Complex> b, 
+			     Complex alpha,
+			     SliceMatrix<Complex> c,
+			     Complex beta)
+  { BASE_LapackMultAdd<Complex> (Trans(a), true, b, false, alpha, c, beta); }
+
+  inline void LapackMultAdd (SliceMatrix<Complex> a, 
+			     SliceMatrix<Complex,ColMajor> b, 
+			     Complex alpha,
+			     SliceMatrix<Complex> c,
+			     Complex beta)
+  { BASE_LapackMultAdd<Complex> (a, false, Trans(b), true, alpha, c, beta); }
+
+  inline void LapackMultAdd (SliceMatrix<Complex,ColMajor> a, 
+			     SliceMatrix<Complex,ColMajor> b, 
+			     Complex alpha,
+			     SliceMatrix<Complex> c,
+			     Complex beta)
+  { BASE_LapackMultAdd<Complex> (Trans(a), true, Trans(b), true, alpha, c, beta); }
+
+
   template <typename TA>
   inline void LapackMultAdd (SliceMatrix<Complex> a, 
 			     TransExpr<TA> b, 
@@ -348,9 +481,49 @@ namespace ngbla
 
 
 
+  template <typename TA, typename TB, typename Talpha, typename Tc, typename Tbeta>
+  inline void LapackMultAdd (MinusExpr<TA> a, 
+			     const TB & b, 
+			     Talpha alpha,			     
+			     const Tc & c,
+			     Tbeta beta)
+  { LapackMultAdd (a.A(), b, -alpha, c, beta); }
 
 
 
+  
+
+
+  // we don't have a lapack function for that 
+  void LapackMultAdd (SliceMatrix<double> a, 
+                      SliceMatrix<Complex,ColMajor> b, 
+                      Complex alpha,
+                      SliceMatrix<Complex> c,
+                      Complex beta);
+  // ---> moved implementation to ng_blas
+  /*
+  {
+    if (beta == 0.0)
+      c = alpha * (a * b);
+    else
+      {
+        c *= beta;
+        c += alpha * (a * b);
+      }
+    // BASE_LapackMultAdd<double> (Trans(a), true, Trans(b), true, alpha, c, beta);
+  }
+  */
+
+  template <typename TA, typename TB, typename ALPHA, typename BETA, typename TC>
+  inline void LapackMultAdd (TA a, 
+                             TB b, 
+                             ALPHA alpha,
+                             SliceMatrix<TC,ColMajor> c,
+                             BETA beta)
+  {
+    LapackMultAdd (Trans(b), Trans(a), alpha, Trans(c), beta);
+  }
+  
 
   /*
 
@@ -403,7 +576,7 @@ namespace ngbla
 
 
 
-    // old functions for compatibilty 
+    // old functions for compatibility 
   inline void LapackMultABt (SliceMatrix<double> a, 
 			     SliceMatrix<double> b, 
 			     SliceMatrix<double> c)
@@ -415,11 +588,11 @@ namespace ngbla
     integer k = a.Width();
     double alpha = 1.0;
     double beta = 0;
-    integer lda = a.Dist();
-    integer ldb = b.Dist();
-    integer ldc = c.Dist();
+    integer lda = max(size_t(1), a.Dist());
+    integer ldb = max(size_t(1), b.Dist());
+    integer ldc = max(size_t(1), c.Dist());
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
  
@@ -434,10 +607,10 @@ namespace ngbla
     integer k = a.Width();
     double alpha = 1.0;
     double beta = 0;
-    integer lda = a.Dist();
-    integer ldb = b.Dist();
-    integer ldc = c.Dist();
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    integer lda = max(size_t(1), a.Dist());
+    integer ldb = max(size_t(1), b.Dist());
+    integer ldc = max(size_t(1), c.Dist());
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
   inline void LapackMultAtB (ngbla::FlatMatrix<double> a, 
@@ -451,11 +624,11 @@ namespace ngbla
     integer k = a.Height();
     double alpha = 1.0;
     double beta = 0;
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Dist(); // c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Dist()); // c.Width();
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
   inline void LapackMultAtBt (ngbla::FlatMatrix<double> a, 
@@ -469,11 +642,11 @@ namespace ngbla
     integer k = a.Height();
     double alpha = 1.0;
     double beta = 0;
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
   inline void LapackMultABt (ngbla::FlatMatrix<ngbla::Complex> a, 
@@ -489,11 +662,11 @@ namespace ngbla
     integer k = a.Width();
     Complex alpha(1,0); // double alpha[2] =  { 1.0, 0.0 };
     Complex beta(0,0);  // double beta[2] = { 0.0, 0.0 };
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    zgemm_ (&transa, &transb, &n, &m, &k, &alpha,  
+    zgemm (&transa, &transb, &n, &m, &k, &alpha,  
             &b(0,0), &ldb, 
             &a(0,0), &lda, &beta, 
             &c(0,0), &ldc);
@@ -514,11 +687,11 @@ namespace ngbla
     integer k = a.Width();
     double alpha = fac;
     double beta = 1.0;
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
 
@@ -534,11 +707,11 @@ namespace ngbla
     integer k = a.Width();
     double alpha = fac;
     double beta = 1.0;
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
   inline void LapackMultAddAtB (ngbla::FlatMatrix<double> a, 
@@ -553,11 +726,11 @@ namespace ngbla
     integer k = a.Height();
     double alpha = fac;
     double beta = 1.0;
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    dgemm_ (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
+    dgemm (&transa, &transb, &n, &m, &k, &alpha, &b(0,0), &ldb, &a(0,0), &lda, &beta, &c(0,0), &ldc);
   }
 
 
@@ -574,11 +747,11 @@ namespace ngbla
     integer k = a.Width();
     Complex alpha(fac, 0);
     Complex beta(1,0);
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    zgemm_ (&transa, &transb, &n, &m, &k, &alpha, 
+    zgemm (&transa, &transb, &n, &m, &k, &alpha, 
             &b(0,0), &ldb, 
             &a(0,0), &lda, &beta, 
             &c(0,0), &ldc);
@@ -599,11 +772,11 @@ namespace ngbla
     integer k = a.Height();
     Complex alpha(fac, 0); // double alpha[2] = { fac, 0 };
     Complex beta(1,0);     // double beta[2] = { 1.0, 0 };
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    zgemm_ (&transa, &transb, &n, &m, &k, &alpha, 
+    zgemm (&transa, &transb, &n, &m, &k, &alpha, 
             &b(0,0), &ldb, 
             &a(0,0), &lda, &beta, 
             &c(0,0), &ldc);
@@ -624,11 +797,11 @@ namespace ngbla
     integer k = a.Width();
     Complex alpha(fac, 0); // double alpha[2] = { fac, 0 };
     Complex beta(1,0);    // double beta[2] = { 1.0, 0 };
-    integer lda = a.Width();
-    integer ldb = b.Width();
-    integer ldc = c.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
+    integer ldc = max(size_t(1), c.Width());
 
-    zgemm_ (&transa, &transb, &n, &m, &k, &alpha, 
+    zgemm (&transa, &transb, &n, &m, &k, &alpha, 
             &b(0,0), &ldb, 
             &a(0,0), &lda, &beta, 
             &c(0,0), &ldc);
@@ -664,44 +837,97 @@ namespace ngbla
   */
 
 
+  template <ORDERING ORD>
+  class LapackLU
+  {
+    Matrix <double, ORD> a;
+    ArrayMem<integer,100> ipiv;
+    
+  public:
+    LapackLU (Matrix<double,ORD> _a)
+      : a(std::move(_a)), ipiv(a.Height())
+    {
+      integer m = a.Height();
+      if (m == 0) return;
+      integer n = a.Width();
+      integer lda = a.Dist();
 
-  inline void LapackInverse (ngbla::FlatMatrix<double> a)
+      integer info;
+      dgetrf(&n, &m, &a(0,0), &lda, &ipiv[0], &info);
+    }
+    
+    template <typename Db>
+    void Solve (VectorView<double,Db> b) const
+    {
+      /*
+      int dgetrs_(char *trans, integer *n, integer *nrhs, 
+                  doublereal *a, integer *lda, integer *ipiv, doublereal *b, integer *
+                  ldb, integer *info);
+      */
+      char transa =  (ORD == ColMajor) ? 'N' : 'T';
+      integer n = a.Height();
+      integer nrhs = 1;
+      integer lda = a.Dist();
+      integer ldb = b.Size();
+      integer info;
+      dgetrs(&transa, &n, &nrhs, a.Data(), &lda, ipiv.Data(), b.Data(), &ldb, &info);
+    }
+    
+    Matrix <double,ORD> Inverse() &&
+    {
+      double hwork;
+      integer lwork = -1;
+      integer n = a.Height();      
+      integer lda = a.Dist();
+      integer info;
+      dgetri(&n, &a(0,0), &lda, &ipiv[0], &hwork, &lwork, &info);
+      lwork = integer(hwork);
+      ArrayMem<double,1000> work(lwork);
+      dgetri(&n, &a(0,0), &lda, &ipiv[0], &work[0], &lwork, &info);
+      return std::move(a);
+    }
+  };
+
+
+  
+
+  inline void LapackInverse (ngbla::SliceMatrix<double> a)
   {
     integer m = a.Height();
     if (m == 0) return;
     integer n = a.Width();
-    integer lda = a.Width();
+    integer lda = max(size_t(1), a.Dist());
 
-    Array<integer> ipiv(n);
+    ArrayMem<integer,100> ipiv(n);
     integer info;
     
-    dgetrf_ (&n, &m, &a(0,0), &lda, &ipiv[0], &info);
+    dgetrf(&n, &m, &a(0,0), &lda, &ipiv[0], &info);
 
     double hwork;
     integer lwork = -1;
-    dgetri_ (&n, &a(0,0), &lda, &ipiv[0], &hwork, &lwork, &info);
+    dgetri(&n, &a(0,0), &lda, &ipiv[0], &hwork, &lwork, &info);
     lwork = integer(hwork);
 
-    Array<double> work(lwork);
-    dgetri_ (&n, &a(0,0), &lda, &ipiv[0], &work[0], &lwork, &info);
+    ArrayMem<double,1000> work(lwork);
+    dgetri(&n, &a(0,0), &lda, &ipiv[0], &work[0], &lwork, &info);
   }
 
 
 
-  inline void LapackInverseSPD (ngbla::FlatMatrix<double> a)
+  inline void LapackInverseSPD (ngbla::SliceMatrix<double> a)
   {
     integer n = a.Width();
     if (n == 0) return;
-    integer lda = n;
+    integer lda = max(size_t(1), a.Dist());
 
     integer info;
-    char uplo = 'L';
+    char uplo = 'U';
 
     dpotrf_ (&uplo, &n, &a(0,0), &lda, &info);
     dpotri_ (&uplo, &n, &a(0,0), &lda, &info);
     for (int i = 0; i < n; i++)
       for (int j = 0; j < i; j++)
-	a(i,j) = a(j,i);
+	a(j,i) = a(i,j);
   }
 
 
@@ -725,8 +951,8 @@ namespace ngbla
   {
     integer m = a.Height();
     integer n = a.Width();
-    integer lda = a.Width();
-    integer ldb = b.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
     integer nrhs = b.Height();
 
     ArrayMem<integer,100> ipiv(n);
@@ -774,13 +1000,13 @@ namespace ngbla
   */
 
 
-  inline void LapackInverse (ngbla::FlatMatrix<ngbla::Complex> a)
+  inline void LapackInverse (ngbla::SliceMatrix<ngbla::Complex> a)
   {
     integer m = a.Height();
     if (m == 0) return;
 
     integer n = a.Width();
-    integer lda = a.Width();
+    integer lda = a.Dist();
     integer * ipiv = new integer[n];
     integer lwork = 100*n;
     Complex * work = new Complex[lwork];
@@ -816,8 +1042,8 @@ namespace ngbla
   {
     integer m = a.Height();
     integer n = a.Width();
-    integer lda = a.Width();
-    integer ldb = b.Width();
+    integer lda = max(size_t(1), a.Width());
+    integer ldb = max(size_t(1), b.Width());
     integer nrhs = b.Height();
     integer * ipiv = new integer[n];
     integer lwork = 100*n;
@@ -924,7 +1150,7 @@ namespace ngbla
         return;
       }
     
-    for ( int i = 0; i < lami.Size(); i++ )
+    for ( size_t i = 0; i < lami.Size(); i++ )
       lami(i) = ngbla::Complex (lami_re[i], lami_im[i]);
 
     delete[] work; 
@@ -945,11 +1171,11 @@ namespace ngbla
     integer nvl = 1; 
     integer nvr = eveci.Width() ; 
   
-    std::complex<double> * vl = 0; 
-    std::complex<double> * vr;//  = new std::complex<double> [nvr*n];
+    Complex * vl = 0;
+    Complex * vr;//  = new std::complex<double> [nvr*n];
   
     integer lwork = 8*n; 
-    std::complex<double> * work = new std::complex<double>[lwork]; 
+    Complex * work = new Complex [lwork];
     double *rwork = new double[8*n];  
     integer info = 0;
   
@@ -960,10 +1186,10 @@ namespace ngbla
     else
       {
         nvr = n;
-        vr =  new std::complex<double> [nvr*n];
+        vr =  new Complex [nvr*n];
       }
 
-    zgeev_(&jobvl, &jobvr, &n, (std::complex<double>*)(void*)&a(0,0), &n, (std::complex<double>*)(void*)&lami(0), vl, &nvl, vr, &nvr, work, &lwork, rwork, &info);
+    zgeev_(&jobvl, &jobvr, &n, &a(0,0), &n, &lami(0), vl, &nvl, vr, &nvr, work, &lwork, rwork, &info);
     //  alpha, beta, &vl, &nvl, vr, &nvr,  
     // 	     work , &lwork, rwork,  &info);
   
@@ -1013,14 +1239,14 @@ namespace ngbla
     char jobvr = 'N', jobvl= 'N';
     // bool balancing = 0; 
   
-    std::complex<double> * alpha= new std::complex<double>[n];
-    std::complex<double> * beta = new std::complex<double>[n]; 
-    std::complex<double> vl=0.; 
+    Complex * alpha= new Complex[n];
+    Complex * beta = new Complex[n];
+    Complex vl=0.;
   
     integer nvl = 1; 
-    std::complex<double> * vr = NULL;
+    Complex * vr = NULL;
   
-    std::complex<double> * work = new std::complex<double>[8*n]; 
+    Complex * work = new Complex[8*n];
     integer lwork = 8*n; 
     double *rwork = new double[8*n];  
   
@@ -1089,10 +1315,10 @@ namespace ngbla
     for(i=0;i<n;i++)
       {
         if(abs(beta[i]) >= 1.e-30) 
-          lami[i]=std::complex<double>(alpha[i]/beta[i]);     
+          lami[i]=Complex(alpha[i]/beta[i]);
         else 
           {
-            lami[i] = std::complex<double>(100.,100.);
+            lami[i] = Complex(100.,100.);
           }
       } 
   
@@ -1168,6 +1394,46 @@ namespace ngbla
     delete [] work; 
   }
 
+  // A = U * diag(S) * V
+  extern void LapackSVD (SliceMatrix<double, ColMajor> A,
+                         SliceMatrix<double, ColMajor> U,
+                         SliceMatrix<double, ColMajor> V,
+                         FlatVector<double> S,                         
+                         bool all);
+
+
+  // A = U * diag(S) * V
+  inline void LapackSVD (SliceMatrix<double> A,
+                         SliceMatrix<double> U,
+                         SliceMatrix<double> V,
+                         FlatVector<double> S,                         
+                         bool all)
+  {
+    LapackSVD (Trans(A), Trans(V), Trans(U), S, all);
+  }
+  
+
+  // A = U * diag(S) * V
+  extern void LapackSVD (SliceMatrix<Complex, ColMajor> A,
+                         SliceMatrix<Complex, ColMajor> U,
+                         SliceMatrix<Complex, ColMajor> V,
+                         FlatVector<double> S,                         
+                         bool all);
+
+
+  // A = U * diag(S) * V
+  inline void LapackSVD (SliceMatrix<Complex> A,
+                         SliceMatrix<Complex> U,
+                         SliceMatrix<Complex> V,
+                         FlatVector<double> S,                         
+                         bool all)
+  {
+    LapackSVD (Trans(A), Trans(V), Trans(U), S, all);
+  }
+  
+  
+
+  
 
 #else
 
@@ -1277,8 +1543,7 @@ namespace ngbla
   { c += fac * a * Trans (b); }
 
 
-
-
+  
   inline void LapackInverse (ngbla::FlatMatrix<double> a)
   { 
     CalcInverse (a);
