@@ -1,16 +1,17 @@
 #include <comp.hpp>
 #include <multigrid.hpp>
-#include <solve.hpp>
+// #include <solve.hpp>
 #include <parallelngs.hpp>
 
+
 namespace ngcomp
 {
   using namespace ngmg;
 
   
-  Preconditioner :: Preconditioner (const PDE * const apde, const Flags & aflags, 
-				    const string aname)
-    : NGS_Object(apde->GetMeshAccess(), aname), flags(aflags)
+  Preconditioner :: Preconditioner (shared_ptr<BilinearForm> bfa, const Flags & aflags,
+                                    const string aname)
+    : NGS_Object(bfa? bfa->GetMeshAccess(): nullptr, aflags, aname), bf(bfa)
   {
     test = flags.GetDefineFlag ("test");
     timing = flags.GetDefineFlag ("timing");
@@ -21,28 +22,26 @@ namespace ngcomp
     uselapack = flags.GetDefineFlag ("lapacktest");
     if ( uselapack ) test = 1;
 
-    if(test)
+    on_proc = int ( flags.GetNumFlag("only_on", -1));
+    if (!flags.GetDefineFlag ("not_register_for_auto_update"))
       {
-	string testresult_ok_name = flags.GetStringFlag ("testresultok","");
-	string testresult_min_name = flags.GetStringFlag ("testresultmin","");
-	string testresult_max_name = flags.GetStringFlag ("testresultmax","");
-	
-	if(testresult_ok_name != "") testresult_ok = &(const_cast<PDE*>(apde)->GetVariable(testresult_ok_name));
-	if(testresult_min_name != "") testresult_min = &(const_cast<PDE*>(apde)->GetVariable(testresult_min_name));
-	if(testresult_max_name != "") testresult_max = &(const_cast<PDE*>(apde)->GetVariable(testresult_max_name));
+        bfa->SetPreconditioner(this);
+        is_registered = true;
       }
 
-    on_proc = int ( flags.GetNumFlag("only_on", -1));
   }
   
+
+  
   Preconditioner :: ~Preconditioner ()
   {
-    ;
+    if (auto bfp = bf.lock(); is_registered && bfp)
+      bfp->UnsetPreconditioner(this);
   }
 
   void Preconditioner :: Test () const
   {
-    cout << "Compute eigenvalues" << endl;
+    cout << IM(1) << "Compute eigenvalues" << endl;
     const BaseMatrix & amat = GetAMatrix();
     const BaseMatrix & pre = GetMatrix();
     int eigenretval;
@@ -56,9 +55,9 @@ namespace ngcomp
         eigen.SetPrecision(1e-15);
         eigenretval = eigen.Calc();
         eigen.PrintEigenValues (*testout);
-        (cout) << " Min Eigenvalue : "  << eigen.EigenValue(1) << endl; 
-        (cout) << " Max Eigenvalue : " << eigen.MaxEigenValue() << endl; 
-        (cout) << " Condition   " << eigen.MaxEigenValue()/eigen.EigenValue(1) << endl; 
+        cout << IM(1) << " Min Eigenvalue : "  << eigen.EigenValue(1) << endl; 
+        cout << IM(1) << " Max Eigenvalue : " << eigen.MaxEigenValue() << endl; 
+        cout << IM(1) << " Condition   " << eigen.MaxEigenValue()/eigen.EigenValue(1) << endl; 
         (*testout) << " Min Eigenvalue : "  << eigen.EigenValue(1) << endl; 
         (*testout) << " Max Eigenvalue : " << eigen.MaxEigenValue() << endl; 
         
@@ -84,17 +83,17 @@ namespace ngcomp
 
 	for ( int i = 0; i < n; i++ )
 	  {
-	    FlatArray<int> rowindices = 
+	    FlatArray rowindices = 
 	      dynamic_cast<const BaseSparseMatrix&>(amat).GetRowIndices(i);
 	    if ( rowindices.Size() <= 1 )
-	      internaldofs.Set(i);
+	      internaldofs.SetBit(i);
 	    else
 	      n_elim++;
 	  }
 
         Matrix<Complex> mat(n_elim), mat2(n_elim), ev(n_elim);
-        BaseVector & v1 = *amat.CreateVector();
-        BaseVector & v2 = *amat.CreateVector();
+        BaseVector & v1 = *amat.CreateColVector();
+        BaseVector & v2 = *amat.CreateColVector();
         FlatVector<Complex> fv1 = v1.FVComplex();
         // FlatVector<Complex> fv2 = v2.FVComplex();
         
@@ -146,8 +145,8 @@ namespace ngcomp
     double time;
     starttime = clock();
 
-    BaseVector & vecf = *pre.CreateVector();
-    BaseVector & vecu = *pre.CreateVector();
+    BaseVector & vecf = *pre.CreateColVector();
+    BaseVector & vecu = *pre.CreateColVector();
 
     vecf = 1;
     int steps = 0;
@@ -177,34 +176,42 @@ namespace ngcomp
   }
 
 
+  void Preconditioner :: ThrowPreconditionerNotReady() const
+  {
+    throw Exception("Preconditioner not ready: either update manually, or use update from assembling");
+  }
 
 
 
-  MGPreconditioner :: MGPreconditioner (const PDE & pde, const Flags & aflags, const string aname)
-    : Preconditioner (&pde,aflags,aname)
+
+
+
+
+  MGPreconditioner :: MGPreconditioner (shared_ptr<BilinearForm> abfa, const Flags & aflags, const string aname)
+    : Preconditioner (abfa,aflags,aname)
   {
     // cout << "*** MGPreconditioner constructor called" << endl;
     mgtest = flags.GetDefineFlag ("mgtest");
     mgfile = flags.GetStringFlag ("mgfile","mgtest.out"); 
     mgnumber = int(flags.GetNumFlag("mgnumber",1)); 
-   
- 
-
-    const MeshAccess & ma = pde.GetMeshAccess();
-    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", NULL));
-
-    const LinearForm * lfconstraint = 
-      pde.GetLinearForm (flags.GetStringFlag ("constraint", ""),1);
-    const FESpace & fes = bfa->GetFESpace();
     
 
-    const BilinearForm * lo_bfa = bfa;
-    const FESpace * lo_fes = &fes;
+    shared_ptr<MeshAccess> ma = abfa->GetMeshAccess();
+    bfa = abfa;
+    // bfa -> SetPreconditioner (this);
 
-    if (&bfa->GetLowOrderBilinearForm())
+
+    // shared_ptr<LinearForm> lfconstraint = pde.GetLinearForm (flags.GetStringFlag ("constraint", ""),1);
+    shared_ptr<FESpace> fes = bfa->GetFESpace();
+    
+
+    shared_ptr<BilinearForm> lo_bfa = bfa;
+    shared_ptr<FESpace> lo_fes = fes;
+
+    if (bfa->GetLowOrderBilinearForm())
       {
-	lo_bfa = &bfa->GetLowOrderBilinearForm();
-	lo_fes = &fes.LowOrderFESpace();
+	lo_bfa = bfa->GetLowOrderBilinearForm();
+	lo_fes = fes->LowOrderFESpacePtr();
       }
     /*
     else if (id == 0 && ntasks > 1 )  // not supported anymore
@@ -214,24 +221,26 @@ namespace ngcomp
       }
     */
 
-    Smoother * sm = NULL;
+    shared_ptr<Smoother> sm = nullptr;
     //    const char * smoother = flags.GetStringFlag ("smoother", "point");
     smoothertype = flags.GetStringFlag ("smoother", "point");
 
     if (smoothertype == "point")
       {
-	sm = new GSSmoother (ma, *lo_bfa);
+	sm = make_shared<GSSmoother> (*ma, *lo_bfa);
       }
     else if (smoothertype == "line")
       {
-	sm = new AnisotropicSmoother (ma, *lo_bfa);
+	sm = make_shared<AnisotropicSmoother> (*ma, *lo_bfa);
       }
     else if (smoothertype == "block") 
       {
-	if (!lfconstraint)
-	  sm = new BlockSmoother (ma, *lo_bfa, flags);
-	else
-	  sm = new BlockSmoother (ma, *lo_bfa, *lfconstraint, flags);
+	// if (!lfconstraint)
+        sm = make_shared<BlockSmoother> (*ma, *lo_bfa, flags);
+          /*
+            else
+            sm = new BlockSmoother (*ma, *lo_bfa, *lfconstraint, flags);
+          */
       }
     /*
     else if (smoothertype == "potential")
@@ -245,57 +254,44 @@ namespace ngcomp
     if (!sm)
       throw Exception ("smoother could not be allocated"); 
 
-    const Prolongation * prol = lo_fes->GetProlongation();
+    auto prol = lo_fes->GetProlongation();
 
-    mgp = new MultigridPreconditioner (ma, *lo_fes, *lo_bfa, sm, 
-				       const_cast<Prolongation*> (prol));
-    mgp->SetOwnProlongation (0);
+    mgp = make_shared<MultigridPreconditioner> (lo_bfa, sm, prol);
     mgp->SetSmoothingSteps (int(flags.GetNumFlag ("smoothingsteps", 1)));
     mgp->SetCycle (int(flags.GetNumFlag ("cycle", 1)));
     mgp->SetIncreaseSmoothingSteps (int(flags.GetNumFlag ("increasesmoothingsteps", 1)));
     mgp->SetCoarseSmoothingSteps (int(flags.GetNumFlag ("coarsesmoothingsteps", 1)));
     mgp->SetUpdateAll( flags.GetDefineFlag( "updateall" ) );
+    mgp->SetHarmonicExtensionProlongation (flags.GetDefineFlag("he_prolongation"));    
+    mgp->SetUpdateAlways(flags.GetDefineFlag("updatealways"));
 
     MultigridPreconditioner::COARSETYPE ct = MultigridPreconditioner::EXACT_COARSE;
-    const char * coarse = flags.GetStringFlag ("coarsetype", "direct");
-    if (strcmp (coarse, "smoothing") == 0)
+    const string & coarse = flags.GetStringFlag ("coarsetype", "direct");
+    if (coarse == "smoothing")
       ct = MultigridPreconditioner::SMOOTHING_COARSE;
-    else if (strcmp (coarse, "cg") == 0)
+    else if (coarse == "cg")
       ct = MultigridPreconditioner::CG_COARSE;
     mgp->SetCoarseType (ct);
     
 
-    coarse_pre = 
-      pde.GetPreconditioner (flags.GetStringFlag ("coarseprecond", ""), 1);
-
-    if (coarse_pre)
-      mgp->SetCoarseType (MultigridPreconditioner::USER_COARSE);
+    // coarse_pre = pde.GetPreconditioner (flags.GetStringFlag ("coarseprecond", ""), 1);
+    // if (coarse_pre) mgp->SetCoarseType (MultigridPreconditioner::USER_COARSE);
 
     finesmoothingsteps = int (flags.GetNumFlag ("finesmoothingsteps", 1));
 
     tlp = 0;
-    inversetype = flags.GetStringFlag("inverse", "sparsecholesky");
+    inversetype = flags.GetStringFlag("inverse", GetInverseName (default_inversetype));
+    GetMemoryTracer().Track(*mgp, "MultiGridPreconditioner");
   }
 
- 
-  MGPreconditioner :: ~MGPreconditioner()
-  {
-    delete mgp;
-    delete tlp;
-  }
-
-  void MGPreconditioner :: FreeSmootherMem ()
-  {
-    if(mgp) mgp->FreeMem();
-    if(tlp) tlp->FreeMem();
-  }
-  
 
   void MGPreconditioner :: Update ()
   {
-    const BilinearForm * lo_bfa = &bfa->GetLowOrderBilinearForm();
+    static Timer t("MGPreconditioner::Update"); RegionTimer reg(t);
+    
+    shared_ptr<BilinearForm> lo_bfa = bfa->GetLowOrderBilinearForm();
 
-    INVERSETYPE invtype, loinvtype;
+    INVERSETYPE invtype, loinvtype = default_inversetype;
     invtype = dynamic_cast<const BaseSparseMatrix & > (bfa->GetMatrix()).SetInverseType (inversetype);
     if (lo_bfa)
       loinvtype = dynamic_cast<const BaseSparseMatrix & > (lo_bfa->GetMatrix()) .SetInverseType (inversetype);
@@ -305,28 +301,25 @@ namespace ngcomp
 
     if (coarse_pre)
       {
-	mgp->SetCoarseGridPreconditioner (&coarse_pre->GetMatrix());
-	mgp->SetOwnCoarseGridPreconditioner(false);
+	mgp->SetCoarseGridPreconditioner (shared_ptr<BaseMatrix> (const_cast<BaseMatrix*>(&coarse_pre->GetMatrix()), NOOP_Deleter));
       }
 
-    if (&bfa->GetLowOrderBilinearForm()) //  || ntasks > 1) not supported anymore
+    if (bfa->GetLowOrderBilinearForm()) //  || ntasks > 1) not supported anymore
       {
-	delete tlp;
-
-	Smoother * fine_smoother = NULL;
-
-	fine_smoother = new BlockSmoother (bfa->GetMeshAccess(), *bfa, flags);
-
-	tlp = new TwoLevelMatrix (&bfa->GetMatrix(),
-				  mgp,
-				  fine_smoother,
-				  bfa->GetMeshAccess().GetNLevels()-1);
-	
-	tlp -> SetSmoothingSteps (finesmoothingsteps);
+        static Timer t("MGPreconditioner::Update - fine precond"); RegionTimer reg(t);
+        auto fine_smoother = make_shared<BlockSmoother> (*bfa->GetMeshAccess(), *bfa, flags);
+        GetMemoryTracer().Track(*fine_smoother, "FineSmoother");
+        tlp = make_shared<TwoLevelMatrix> (&bfa->GetMatrix(),
+                                           &*mgp,
+                                           fine_smoother,
+                                           bfa->GetMeshAccess()->GetNLevels()-1);
+        tlp -> SetSmoothingSteps (finesmoothingsteps);
+        if (bfa->GetFESpace()->LowOrderEmbedding())
+          tlp->SetEmbedding (bfa->GetFESpace()->LowOrderEmbedding());
 	tlp -> Update();
       }
     else
-      tlp = 0;
+      tlp = nullptr;
 
     if (timing) Timing();
     if (test) Test();
@@ -340,10 +333,9 @@ namespace ngcomp
   
   void MGPreconditioner :: CleanUpLevel () 
   { 
-    if (&bfa->GetLowOrderBilinearForm())
+    if (bfa->GetLowOrderBilinearForm())
       {
-	delete tlp;
-	tlp = 0;
+	tlp = nullptr;
       }    
   }
 
@@ -356,7 +348,7 @@ namespace ngcomp
       return *mgp; 
   }
 
-  void MGPreconditioner::PrintReport (ostream & ost)
+  void MGPreconditioner::PrintReport (ostream & ost) const
   {
     ost << "Multigrid preconditioner" << endl
 	<< "bilinear-form = " << bfa->GetName() << endl
@@ -364,14 +356,13 @@ namespace ngcomp
   }
 
 
-  void MGPreconditioner::MemoryUsage (Array<MemoryUsageStruct*> & mu) const
+  Array<MemoryUsage> MGPreconditioner::GetMemoryUsage () const
   {
-    int olds = mu.Size();
+    auto mu = GetMatrix().GetMemoryUsage ();;
 
-    if (&GetMatrix()) GetMatrix().MemoryUsage (mu);;
-
-    for (int i = olds; i < mu.Size(); i++)
-      mu[i]->AddName (string(" mgpre ")); // +GetName());
+    for (int i = 0; i < mu.Size(); i++)
+      mu[i].AddName (string(" mgpre ")); // +GetName());
+    return mu;
   }
     
 
@@ -400,7 +391,7 @@ namespace ngcomp
 
     // double cond;
 
-    condout << bfa->GetFESpace().GetNDof() << "\t" << bfa->GetFESpace().GetOrder() << "\t" << eigen.EigenValue(mgnumber) << "\t" << eigen.MaxEigenValue() << "\t" 
+    condout << bfa->GetFESpace()->GetNDof() << "\t" << bfa->GetFESpace()->GetOrder() << "\t" << eigen.EigenValue(mgnumber) << "\t" << eigen.MaxEigenValue() << "\t" 
 	    << eigen.MaxEigenValue()/eigen.EigenValue(mgnumber) <<  "\t" << endl;
     
     if(testresult_ok) *testresult_ok = eigenretval;
@@ -409,9 +400,23 @@ namespace ngcomp
 
   }
 
+  void MGPreconditioner :: SetDirectSolverCluster(shared_ptr<Array<int>> cluster)
+  {
+    if(tlp)
+      {
+        if(auto blocksmoother = dynamic_cast<BlockSmoother*>(&tlp->GetSmoother()))
+          blocksmoother->SetDirectSolverCluster(cluster);
+      }
+    else
+      if(auto blocksmoother = dynamic_cast<BlockSmoother*>(&mgp->GetSmoother()))
+        blocksmoother->SetDirectSolverCluster(cluster);
+  }
 
-
-
+  void MGPreconditioner::SetCoarsePreconditioner(shared_ptr<Preconditioner> prec)
+  {
+      coarse_pre = prec;
+      mgp->SetCoarseType (MultigridPreconditioner::USER_COARSE);
+  }
 
   
   // ****************************** DirectPreconditioner **************************
@@ -419,53 +424,76 @@ namespace ngcomp
 
   class NGS_DLL_HEADER DirectPreconditioner : public Preconditioner
   {
-    const BilinearForm * bfa;
-    BaseMatrix * inverse;
+    shared_ptr<BilinearForm> bfa;
+    shared_ptr<BaseMatrix> inverse;
     string inversetype;
 
   public:
-    DirectPreconditioner (const PDE & pde, const Flags & aflags,
+    DirectPreconditioner (shared_ptr<BilinearForm> abfa, const Flags & aflags,
 			  const string aname = "directprecond")
-      : Preconditioner(&pde,aflags,aname)
+      : Preconditioner(abfa,aflags,aname), bfa(abfa)
     {
-      bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", NULL));
-      inverse = NULL;
-      inversetype = flags.GetStringFlag("inverse", "sparsecholesky");
+      // bfa -> SetPreconditioner (this);
+      inversetype = flags.GetStringFlag("inverse", GetInverseName (default_inversetype));
     }
-    
+
     ///
     virtual ~DirectPreconditioner()
     {
-      delete inverse;
+      ; //delete inverse;
     }
     
+    virtual void FinalizeLevel (const BaseMatrix * mat) 
+    {
+      Update();
+    }
     
     ///
     virtual void Update ()
     {
-      delete inverse;
+      // delete inverse;
+      if (GetTimeStamp() == bfa->GetTimeStamp()) return;
+      timestamp = bfa->GetTimeStamp();
+      
+      cout << IM(3) << "Update Direct Solver Preconditioner" << flush;
       
       try
-	{
+	{                                          
+          auto have_sparse_fact = dynamic_pointer_cast<SparseFactorization> (inverse);
+          if (have_sparse_fact && have_sparse_fact -> SupportsUpdate())
+            {
+              if (have_sparse_fact->GetAMatrix() == bfa->GetMatrixPtr())
+                {
+                  // cout << "have the same matrix, can update factorization" << endl;
+                  have_sparse_fact->Update();
+                  return;
+                }
+            }
+          
 	  bfa->GetMatrix().SetInverseType (inversetype);
-	  const BitArray * freedofs = 
-	    bfa->GetFESpace().GetFreeDofs (bfa->UsesEliminateInternal());
+	  shared_ptr<BitArray> freedofs = 
+	    bfa->GetFESpace()->GetFreeDofs (bfa->UsesEliminateInternal());
 	  inverse = bfa->GetMatrix().InverseMatrix(freedofs);
 	}
       catch (exception & e)
 	{
-	  throw Exception ("DirectPreconditioner: needs a sparse matrix (or has memory problems)");
+	  throw Exception (string("caught exception in DirectPreconditioner: \n") +
+                           e.what() + 
+                           "\nneeds a sparse matrix (or has memory problems)");
 	}
+      GetMemoryTracer().Track(*inverse, "Inverse");
     }
 
     virtual void CleanUpLevel ()
     {
-      delete inverse;
-      inverse = NULL;
+      // delete inverse;
+      inverse = nullptr;
     }
 
     virtual const BaseMatrix & GetMatrix() const
     {
+      if (!inverse)
+        ThrowPreconditionerNotReady();        
       return *inverse;
     }
 
@@ -489,64 +517,49 @@ namespace ngcomp
   // ****************************** LocalPreconditioner *******************************
 
 
-
-  LocalPreconditioner :: LocalPreconditioner (PDE * pde, Flags & aflags, 
-					      const string aname)
-    : Preconditioner (pde,aflags,aname),
-      coarse_pre(0)
+  /**
+     Local (Block-Jacobi or Block-Gauss-Seidel) preconditioner
+  */
+  class LocalPreconditioner : public Preconditioner
   {
-    bfa = pde->GetBilinearForm (flags.GetStringFlag ("bilinearform", NULL));
-    block = flags.GetDefineFlag ("block");
-    jacobi = NULL;
-    locprectest = flags.GetDefineFlag ("mgtest");
-    locprecfile = flags.GetStringFlag ("mgfile","locprectest.out"); 
+  protected:
+    ///
+    shared_ptr<BilinearForm> bfa;
+    ///
+    shared_ptr<BaseMatrix> jacobi;
+    ///
+    bool block;
+    bool locprectest; 
+    string locprecfile; 
 
-    string smoother = flags.GetStringFlag("smoother","");
-    if ( smoother == "block" )
-      block = true;
-
-    // coarse-grid preconditioner only used in parallel!!
-    ct = "NO_COARSE";
-    const char * coarse = flags.GetStringFlag ("coarsetype", "nocoarse");
-    if (strcmp (coarse, "smoothing") == 0)
-      ct = "SMOOTHING_COARSE";
-    else if (strcmp (coarse, "direct") == 0)
-      ct = "DIRECT_COARSE";
+    string ct;
+    shared_ptr<Preconditioner> coarse_pre;
+    function<shared_ptr<Table<DofId>>(FESpace&)> blockcreator;
+  public:
+    ///
+    LocalPreconditioner (shared_ptr<BilinearForm> bfa, const Flags & aflags,
+			 const string aname = "localprecond");
+    ///
+    virtual ~LocalPreconditioner() { ; }
+    ///
+    virtual bool IsComplex() const { return jacobi->IsComplex(); }
     
-
-    coarse_pre = 
-      pde->GetPreconditioner (flags.GetStringFlag ("coarseprecond", ""), 1);
-    if (coarse_pre)
-      ct = "USER_COARSE";
-
-  }
-   
-
-  LocalPreconditioner :: ~LocalPreconditioner()
-  {
-    delete jacobi;
-  }
-
-  void LocalPreconditioner :: Update ()
-  {
-    cout << IM(1) << "Update Local Preconditioner" << flush;
-    delete jacobi;
-    
-    // const BaseSparseMatrix& amatrix = dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix());
-    // 	if ( inversetype != "none" )
-    // 	amatrix.SetInverseType ( inversetype );
-	
-    int blocktype = int (flags.GetNumFlag ( "blocktype", -1));
-
-    // if (MyMPI_GetNTasks() != 1) return;
-    bool parallel = (this->on_proc == -1);
-    /*
-    if ( !parallel && id != this->on_proc )
-      {
+    ///
+    virtual void FinalizeLevel (const BaseMatrix * mat) 
+    {
+      cout << IM(3) << "Update Local Preconditioner" << flush;
+      timestamp = bfa->GetTimeStamp();
+      int blocktype = int (flags.GetNumFlag ( "blocktype", -1));
+      
+      // if (MyMPI_GetNTasks() != 1) return;
+      bool parallel = (this->on_proc == -1);
+      /*
+        if ( !parallel && id != this->on_proc )
+        {
 	jacobi = 0; 
 	return;
-      }
-    */
+        }
+      */
 
 //     BaseBlockJacobiPrecond::COARSE_TYPE bbct;
 //     switch ( ct  )
@@ -557,45 +570,108 @@ namespace ngcomp
 // 	bbct = DIRECT_COARSE; break;
 //       }
 
-    if ( block && blocktype == -1 ) blocktype = 0;
-    if ( blocktype >= 0 )
-      {
-	// new: blocktypes, specified in fespace
-	if (bfa->UsesEliminateInternal())
-	  flags.SetFlag("eliminate_internal");
-	Table<int> * blocks = bfa->GetFESpace().CreateSmoothingBlocks(flags);
-	jacobi = dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix())
-	  .CreateBlockJacobiPrecond(*blocks, 0, coarse_pre, parallel, bfa->GetFESpace().GetFreeDofs());
-	// dynamic_cast<BaseBlockJacobiPrecond&> (*jacobi) . InitCoarseType(ct, bfa->GetFESpace().GetFreeDofs());
-      }
-    else if (block)
-      {
-	cout << "\nFlag block deprecated: use -blocktype=<typeno> instead" << endl;
-      }
-    else
-      {
-	const BaseMatrix * mat = &bfa->GetMatrix();
+      if (blockcreator)
+        {
+          shared_ptr<Table<int>> blocks = blockcreator(*bfa->GetFESpace());
+          // cout << "created blocks: " << *blocks << endl;
+          jacobi = dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix())
+            .CreateBlockJacobiPrecond(blocks, 0, parallel, bfa->GetFESpace()->GetFreeDofs());
+          return;
+        }
+      
+      if ( block && blocktype == -1 ) blocktype = 0;
+      if ( blocktype >= 0 )
+        {
+          // new: blocktypes, specified in fespace
+          if (bfa->UsesEliminateInternal())
+            flags.SetFlag("eliminate_internal");
+          shared_ptr<Table<int>> blocks = bfa->GetFESpace()->CreateSmoothingBlocks(flags);
+          jacobi = dynamic_cast<const BaseSparseMatrix&> (bfa->GetMatrix())
+            .CreateBlockJacobiPrecond(blocks, 0, parallel, bfa->GetFESpace()->GetFreeDofs());
+        }
+      else if (block)
+        {
+          cout << "\nFlag block deprecated: use -blocktype=<typeno> instead" << endl;
+        }
+      else
+        {
+          shared_ptr<BaseMatrix> mat = bfa->GetMatrixPtr();
 #ifdef PARALLEL
-	if (dynamic_cast<const ParallelMatrix*> (mat))
-	  mat = &(dynamic_cast<const ParallelMatrix*> (mat)->GetMatrix());
+          if (dynamic_pointer_cast<ParallelMatrix> (mat))
+            mat = dynamic_pointer_cast<ParallelMatrix> (mat)->GetMatrix();
 #endif
-	jacobi = dynamic_cast<const BaseSparseMatrix&> (*mat)
-	  .CreateJacobiPrecond(bfa->GetFESpace().GetFreeDofs(bfa->UsesEliminateInternal()));
-      }
+          jacobi = dynamic_pointer_cast<BaseSparseMatrix> (mat)
+            -> CreateJacobiPrecond(bfa->GetFESpace()->GetFreeDofs(bfa->UsesEliminateInternal()));
+        }
+    }
+
+    virtual void Update ()
+    {
+      if (GetTimeStamp() < bfa->GetTimeStamp())
+        FinalizeLevel (&bfa->GetMatrix());
+      if (test) Test();
+      if(locprectest) LocPrecTest(); 
+    }
 
 
+    ///
+    virtual const BaseMatrix & GetMatrix() const
+    {
+      if (!jacobi)
+        ThrowPreconditionerNotReady();
+      return *jacobi;
+    }
     
-    if (test) Test();
-    if(locprectest) LocPrecTest(); 
-    //    cout << " done" << endl;
-  }
+    virtual shared_ptr<BaseMatrix> GetMatrixPtr()
+    {
+      if (!jacobi)
+        ThrowPreconditionerNotReady();
+      return jacobi;
+    }
 
-  const ngla::BaseMatrix & LocalPreconditioner :: GetMatrix() const
+    ///
+    virtual const BaseMatrix & GetAMatrix() const
+    {
+      return bfa->GetMatrix(); 
+    }
+    ///
+    virtual const char * ClassName() const
+    { return "Local Preconditioner"; }
+    void LocPrecTest () const;
+  };
+
+
+
+
+
+  LocalPreconditioner :: 
+  LocalPreconditioner  (shared_ptr<BilinearForm> abfa, const Flags & aflags,
+                        const string aname)
+
+    : Preconditioner (abfa,aflags,aname), bfa(abfa)
   {
-    return *jacobi;
+    // bfa -> SetPreconditioner (this);
+
+    block = flags.GetDefineFlag ("block");
+    locprectest = flags.GetDefineFlag ("mgtest");
+    locprecfile = flags.GetStringFlag ("mgfile","locprectest.out"); 
+
+    string smoother = flags.GetStringFlag("smoother","");
+    if ( smoother == "block" )
+      block = true;
+
+    // coarse-grid preconditioner only used in parallel!!
+    ct = "NO_COARSE";
+
+    if(flags.AnyFlagDefined("blockcreator"))
+      {
+        blockcreator = std::any_cast<function<shared_ptr<Table<DofId>>(const FESpace&)>>(flags.GetAnyFlag("blockcreator"));
+        cout << IM(3) << "local pre, got blockcreator" << endl;
+      }
   }
 
- void LocalPreconditioner :: LocPrecTest () const
+
+  void LocalPreconditioner :: LocPrecTest () const
   {
     cout << "Compute eigenvalues" << endl;
     const BaseMatrix & amat = GetAMatrix();
@@ -618,7 +694,7 @@ namespace ngcomp
 
     // double cond;
 
-    condout << bfa->GetFESpace().GetNDof() << "\t" << bfa->GetFESpace().GetOrder() << "\t" << eigen.EigenValue(1) << "\t" << eigen.MaxEigenValue() << "\t" 
+    condout << bfa->GetFESpace()->GetNDof() << "\t" << bfa->GetFESpace()->GetOrder() << "\t" << eigen.EigenValue(1) << "\t" << eigen.MaxEigenValue() << "\t" 
 	    << eigen.MaxEigenValue()/eigen.EigenValue(1) <<  "\t" << endl;
     
     if(testresult_ok) *testresult_ok = eigenretval;
@@ -636,21 +712,16 @@ namespace ngcomp
   // ****************************** TwoLevelPreconditioner *******************************
 
 
-  TwoLevelPreconditioner :: TwoLevelPreconditioner (PDE * apde, Flags & aflags, const string aname)
-    : Preconditioner(apde,aflags,aname)
-  {
-    pde = apde;
-    bfa = pde->GetBilinearForm (flags.GetStringFlag ("bilinearform", NULL));
-    cpre = pde->GetPreconditioner (flags.GetStringFlag ("coarsepreconditioner", NULL));
-    smoothingsteps = int (flags.GetNumFlag ("smoothingsteps", 1));
-    premat = NULL;
-  }
-
   TwoLevelPreconditioner :: ~TwoLevelPreconditioner()
   {
     delete premat;
   }
 
+  const BaseMatrix & TwoLevelPreconditioner :: GetMatrix() const
+  {
+    return *premat;
+  }
+  
   void TwoLevelPreconditioner :: Update ()
   {
     /*
@@ -679,14 +750,6 @@ namespace ngcomp
 
 
 
-ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, const string aname)
-    : Preconditioner(apde,aflags,aname)
-  {
-    dim = int (flags.GetNumFlag ("dim", 1));
-    cm = 0;
-    creal = apde->GetPreconditioner (flags.GetStringFlag ("realpreconditioner",""));
-  }
-
   ComplexPreconditioner :: ~ComplexPreconditioner()
   { 
     ;
@@ -698,16 +761,16 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
     switch (dim)
       {
       case 1:
-	cm = new Real2ComplexMatrix<double,Complex> (&creal->GetMatrix());
+	cm = new Real2ComplexMatrix<double,Complex> (creal->GetMatrixPtr());
 	break;
       case 2:
-	cm = new Real2ComplexMatrix<Vec<2,double>,Vec<2,Complex> > (&creal->GetMatrix());
+	cm = new Real2ComplexMatrix<Vec<2,double>,Vec<2,Complex> > (creal->GetMatrixPtr());
 	break;
       case 3:
-	cm = new Real2ComplexMatrix<Vec<3,double>,Vec<3,Complex> > (&creal->GetMatrix());
+	cm = new Real2ComplexMatrix<Vec<3,double>,Vec<3,Complex> > (creal->GetMatrixPtr());
 	break;
       case 4:
-	cm = new Real2ComplexMatrix<Vec<4,double>,Vec<4,Complex> > (&creal->GetMatrix());
+	cm = new Real2ComplexMatrix<Vec<4,double>,Vec<4,Complex> > (creal->GetMatrixPtr());
 	break;
       default:
 	cout << "Error: dimension " << dim << " for complex preconditioner not supported!" << endl;
@@ -716,16 +779,6 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
 
 
 
-  ChebychevPreconditioner :: ChebychevPreconditioner (PDE * apde, Flags & aflags, const string aname)
-    : Preconditioner(apde,aflags,aname)
-  {
-    steps = int (flags.GetNumFlag ("steps",10.));
-    cm = 0;
-    csimple = apde->GetPreconditioner (flags.GetStringFlag ("csimple",""));
-    bfa = apde->GetBilinearForm (flags.GetStringFlag ("bilinearform",""));
-    test = flags.GetDefineFlag ("test");
-  }
-
   ChebychevPreconditioner :: ~ChebychevPreconditioner()
   { 
     ;
@@ -768,14 +821,6 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
   ////////////////////////////////////////////////////////////////////////////////
   // added 08/19/2003, FB
 
-  NonsymmetricPreconditioner :: NonsymmetricPreconditioner (PDE * apde, Flags & aflags, const string aname)
-    : Preconditioner(apde,aflags,aname)
-  {
-    dim = int (flags.GetNumFlag ("dim", 1));
-    cm = 0;
-    cbase = apde->GetPreconditioner (flags.GetStringFlag ("basepreconditioner",""));
-  }
-
   NonsymmetricPreconditioner :: ~NonsymmetricPreconditioner()
   { 
     ;
@@ -823,24 +868,6 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
   ////////////////////////////////////////////////////////////////////////////////
 
 
-  CommutingAMGPreconditioner :: CommutingAMGPreconditioner (PDE * apde, Flags & aflags, const string aname)
-    : Preconditioner (apde,aflags,aname), pde(apde)
-  {
-    bfa = pde->GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
-    while (&bfa->GetLowOrderBilinearForm())
-      bfa = &bfa->GetLowOrderBilinearForm();
-
-    coefse = pde->GetCoefficientFunction (flags.GetStringFlag ("coefse", ""),1);    
-    coefe = pde->GetCoefficientFunction (flags.GetStringFlag ("coefe", ""),1);    
-    coeff = pde->GetCoefficientFunction (flags.GetStringFlag ("coeff", ""),1);    
-
-    hcurl = dynamic_cast<const NedelecFESpace*> (&bfa->GetFESpace()) != 0;
-    levels = int (flags.GetNumFlag ("levels", 10));
-    coarsegrid = flags.GetDefineFlag ("coarsegrid");
-
-    amg = NULL;
-  }
-
   CommutingAMGPreconditioner :: ~CommutingAMGPreconditioner ()
   {
     delete amg;
@@ -851,20 +878,21 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
     cout << "Update amg" << endl;
 
 #ifdef PARALLEL
-    if ( this->on_proc != MyMPI_GetId() && this->on_proc != -1)
+    auto comm = ma->GetCommunicator();
+    if ( this->on_proc != comm.Rank() && this->on_proc != -1)
       {
 	amg = NULL;
 	return;
       }
 #endif
 
-    const MeshAccess & ma = pde->GetMeshAccess();
+    shared_ptr<MeshAccess> ma = nullptr; //pde->GetMeshAccess();
 
-    int nedge = ma.GetNEdges(); 
-    int nface = ma.GetNFaces(); 
-    int nel = ma.GetNE();
+    size_t nedge = ma->GetNEdges(); 
+    size_t nface = ma->GetNFaces(); 
+    size_t nel = ma->GetNE();
 
-    if (coarsegrid && ma.GetNLevels() > 1)
+    if (coarsegrid && ma->GetNLevels() > 1)
       return;
       
 
@@ -874,19 +902,18 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
 
     Array<INT<2> > e2v (nedge);
     e2v = INT<2> (-1, -1);
-    for (int i = 0; i < nedge; i++)
-      if (ma.GetClusterRepEdge (i) >= 0)
-	ma.GetEdgePNums (i, e2v[i][0], e2v[i][1]);
+    for (size_t i = 0; i < nedge; i++)
+      if (ma->GetClusterRepEdge (i) >= 0)
+	e2v[i] = ma->GetEdgePNums (i);
 
     // cout << "get faces" << endl;
 
     Array<INT<4> > f2v (nface);
-    Array<int> fpnums;
     for (int i = 0; i < nface; i++)
       {
-	if (ma.GetClusterRepFace (i) >= 0)
+	if (ma->GetClusterRepFace (i) >= 0)
 	  {
-	    ma.GetFacePNums (i, fpnums);
+	    auto fpnums = ma->GetFacePNums (i);
 	    
 	    f2v[i][3] = -1;
 	    for (int j = 0; j < fpnums.Size(); j++)
@@ -914,8 +941,8 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
       if (e2v[i][0] != -1)
 	{
 	  Vec<3> v = 
-	    ma.GetPoint<3> (e2v[i][0]) -
-	    ma.GetPoint<3> (e2v[i][1]);
+	    ma->GetPoint<3> (e2v[i][0]) -
+	    ma->GetPoint<3> (e2v[i][1]);
 	  hi[i] = L2Norm (v);
 	}
     
@@ -924,15 +951,15 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
 	ai.SetSize(nface);
 	for (int i = 0; i < nface; i++)
 	  {
-	    Vec<3> p1 = ma.GetPoint<3> (f2v[i][0]);
-	    Vec<3> p2 = ma.GetPoint<3> (f2v[i][1]);
-	    Vec<3> p3 = ma.GetPoint<3> (f2v[i][2]);
+	    Vec<3> p1 = ma->GetPoint<3> (f2v[i][0]);
+	    Vec<3> p2 = ma->GetPoint<3> (f2v[i][1]);
+	    Vec<3> p3 = ma->GetPoint<3> (f2v[i][2]);
 	    Vec<3> vn = Cross (Vec<3> (p2-p1), Vec<3> (p3-p1));
 	    ai[i] = L2Norm (vn);
 	  }
       }
 
-    Array<int> ednums(12), fanums(12);
+    // Array<int> fanums(12);
     LocalHeap lh (10000, "CommutingAMG");
     IntegrationPoint ip(0, 0, 0, 0);
 
@@ -942,39 +969,41 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
     weightf = 0;
     for (int i = 0; i < nel; i++)
       {
+        ElementId ei(VOL,i);
 	HeapReset hr(lh);
-	ma.GetElEdges (i, ednums);
+	auto ednums = ma->GetElEdges (ei);
 
-	ElementTransformation & eltrans = ma.GetTrafo (i, false, lh);
+	ElementTransformation & eltrans = ma->GetTrafo (ei, lh);
 	MappedIntegrationPoint<3,3> sip(ip, eltrans);
 
-	double vol = ma.ElementVolume (i);
-	double vale = Evaluate (*coefe, sip);
+	double vol = ma->ElementVolume (i);
+	double vale = ngfem::Evaluate (*coefe, sip);
 
 	for (int j = 0; j < ednums.Size(); j++)
 	  weighte[ednums[j]] += vale * vol / sqr (hi[ednums[j]]);
 
 	if (hcurl)
 	  {
-	    ma.GetElFaces (i, fanums);
-	    double valf = Evaluate (*coeff, sip);
+	    auto fanums = ma->GetElFaces (ei);
+	    double valf = ngfem::Evaluate (*coeff, sip);
 	    for (int j = 0; j < fanums.Size(); j++)
 	      weightf[fanums[j]] += valf * vol / sqr (ai[fanums[j]]);
 	  }
       }
 
-    int nsel = ma.GetNSE();
+    int nsel = ma->GetNSE();
     if (coefse)
       for (int i = 0; i < nsel; i++)
 	{
+          ElementId sei(BND,i);
 	  HeapReset hr(lh);
-	  ma.GetSElEdges (i, ednums);
-	  ElementTransformation & eltrans = ma.GetTrafo (i, true, lh);
+	  auto ednums = ma->GetElEdges (sei);
+	  ElementTransformation & eltrans = ma->GetTrafo (sei, lh);
 
 	  MappedIntegrationPoint<2,3> sip(ip, eltrans);
 
-	  double vol = ma.SurfaceElementVolume (i);
-	  double vale = Evaluate (*coefse, sip);
+	  double vol = ma->SurfaceElementVolume (i);
+	  double vale = ngfem::Evaluate (*coefse, sip);
 
 	  for (int j = 0; j < ednums.Size(); j++)
 	    weighte[ednums[j]] += vale * vol / sqr (hi[ednums[j]]);
@@ -986,10 +1015,10 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
     timer_coarse.Start();
     Array< Vec<3> > vertices;
 
-    int nv = ma.GetNV();
+    int nv = ma->GetNV();
     vertices.SetSize(nv);
     for (int i = 0; i < nv; i++)
-      ma.GetPoint(i, vertices[i]);
+      ma->GetPoint(i, vertices[i]);
 
     CommutingAMG * amgmat;
     if (hcurl)
@@ -1041,41 +1070,33 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
 
   PreconditionerClasses::PreconditionerInfo::
   PreconditionerInfo (const string & aname,
-		      Preconditioner* (*acreator)(const PDE & pde, const Flags & flags, const string & name))
-    : name(aname), creator(acreator) 
-  {
-    ;
-  }
-  
-  PreconditionerClasses :: PreconditionerClasses ()
+                      function<shared_ptr<Preconditioner>(shared_ptr<BilinearForm>,const Flags &,const string)> acreatorbf,
+                      DocInfo adocinfo)
+    : name(aname), creatorbf(acreatorbf), docinfo(adocinfo)
   {
     ;
   }
 
-  PreconditionerClasses :: ~PreconditionerClasses()
-  {
-    for(int i=0; i<prea.Size(); i++)
-      delete prea[i];
-  }
+  /*
+  PreconditionerClasses :: PreconditionerClasses () { }
+  PreconditionerClasses :: ~PreconditionerClasses() { }
+  */
   
   void PreconditionerClasses :: 
   AddPreconditioner (const string & aname,
-		     Preconditioner* (*acreator)(const PDE & pde, 
-						 const Flags & flags,
-						 const string & name))
+                     function<shared_ptr<Preconditioner>(shared_ptr<BilinearForm>,const Flags &,const string)> acreatorbf,
+                     DocInfo docinfo)
   {
-    prea.Append (new PreconditionerInfo(aname, acreator));
+    prea.Append (make_unique<PreconditionerInfo>(aname, acreatorbf, docinfo));
   }
 
   const PreconditionerClasses::PreconditionerInfo * 
   PreconditionerClasses::GetPreconditioner(const string & name)
   {
     for (int i = 0; i < prea.Size(); i++)
-      {
-	if (name == prea[i]->name)
-	  return prea[i];
-      }
-    return 0;
+      if (name == prea[i]->name)
+        return prea[i].get();
+    return nullptr;
   }
 
   void PreconditionerClasses :: Print (ostream & ost) const
@@ -1087,6 +1108,11 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
       ost << setw(20) << prea[i]->name << endl;
   }
 
+  void PreconditionerClasses :: Cleanup ()
+  {
+      prea.DeleteAll();
+  }
+
  
   PreconditionerClasses & GetPreconditionerClasses ()
   {
@@ -1096,6 +1122,7 @@ ComplexPreconditioner :: ComplexPreconditioner (PDE * apde, Flags & aflags, cons
 
   RegisterPreconditioner<MGPreconditioner> registerMG("multigrid");
   RegisterPreconditioner<DirectPreconditioner> registerDirect("direct");
+  RegisterPreconditioner<LocalPreconditioner> registerlocal("local");
 
 }
 
