@@ -32,10 +32,10 @@ namespace ngfem
   {
   protected:
     double alpha;   // interior penalyty
-    CoefficientFunction *coef_lam;
+    shared_ptr<CoefficientFunction> coef_lam;
   public:
-    DGInnerFacet_LaplaceIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetBilinearFormIntegrator(coeffs)
+    DGInnerFacet_LaplaceIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+    // : FacetBilinearFormIntegrator(coeffs)
     { 
       coef_lam  = coeffs[0];
       if (dgtype!=DG_FORMULATIONS::BO)
@@ -43,29 +43,30 @@ namespace ngfem
     }
 
     virtual ~DGInnerFacet_LaplaceIntegrator () { ; }
+
+    virtual VorB VB() const
+    { return VOL; }
+    virtual xbool IsSymmetric () const { return true; }
     
-    virtual bool BoundaryForm () const 
-    { return 0; }
-    
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGInnerFacet_LaplaceIntegrator (coeffs);
     }
     
     virtual void CalcElementMatrix (const FiniteElement & fel,
                                         const ElementTransformation & eltrans, 
-                                        FlatMatrix<double> & elmat,
+                                        FlatMatrix<double> elmat,
                                         LocalHeap & lh) const
     {
       throw Exception("DGInnerFacet_LaplaceIntegrator::CalcElementMatrix - not implemented!");
     }
     
     virtual void CalcFacetMatrix (const FiniteElement & volumefel1, int LocalFacetNr1,
-			 const ElementTransformation & eltrans1, FlatArray<int> & ElVertices1,
-			 const FiniteElement & volumefel2, int LocalFacetNr2,
-			 const ElementTransformation & eltrans2, FlatArray<int> & ElVertices2,
-                         FlatMatrix<double> & elmat,
-                         LocalHeap & lh) const
+                                  const ElementTransformation & eltrans1, FlatArray<int> & ElVertices1,
+                                  const FiniteElement & volumefel2, int LocalFacetNr2,
+                                  const ElementTransformation & eltrans2, FlatArray<int> & ElVertices2,
+                                  FlatMatrix<double> elmat,
+                                  LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("DGInnerFacet_LaplaceIntegrator");
 
@@ -83,7 +84,7 @@ namespace ngfem
       fel2_l2 = dynamic_cast<const ScalarFiniteElement<D>*> (&volumefel2);
       eltype2 = volumefel2.ElementType();
       int nd2 = fel2_l2->GetNDof();
-      int maxorder = max(fel1_l2->Order(),fel2_l2->Order());
+      int maxorder = max2(fel1_l2->Order(),fel2_l2->Order());
       
       elmat = 0.0;
 
@@ -195,28 +196,32 @@ namespace ngfem
   class ConvectionIntegrator : public BilinearFormIntegrator
   {
   protected:
-    CoefficientFunction * coef_conv[D];
+    // shared_ptr<CoefficientFunction>  coef_conv[D];
+    DVec<D> coef_conv;
   public:
-    ConvectionIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : BilinearFormIntegrator()
+    ConvectionIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+      : BilinearFormIntegrator(), coef_conv (coeffs)
     { 
+      /*
       for (int j = 0; j < D; j++)
         coef_conv[j] = coeffs[j];
+      */
     }
 
     virtual ~ConvectionIntegrator () { ; }
 
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new ConvectionIntegrator (coeffs);
     }
-    
-    virtual bool BoundaryForm () const 
-    { return 0; }
+
+    virtual VorB VB() const
+    { return VOL; }
+    virtual xbool IsSymmetric () const { return false; }
 
     virtual void CalcElementMatrix (const FiniteElement & fel,
                                         const ElementTransformation & eltrans, 
-                                        FlatMatrix<double> & elmat,
+                                        FlatMatrix<double> elmat,
                                         LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("ConvectionIntegrator");
@@ -235,7 +240,7 @@ namespace ngfem
       
       FlatMatrixFixHeight<2> bmat(nd, lh);
       FlatMatrixFixHeight<2> dbmat(nd, lh);
-      Mat<2> dmat;
+      // Mat<2> dmat;
           
       const IntegrationRule & ir_vol =
         SelectIntegrationRule (eltype, 2*fel_l2.Order());
@@ -247,9 +252,12 @@ namespace ngfem
 	HeapReset hr(lh);
 	const MappedIntegrationPoint<D,D> sip(ir_vol[l], eltrans);
 	Vec<D> conv;
+        coef_conv.GenerateVector (fel_l2, sip, conv, lh);
+        /*
 	for (int j = 0; j < D; j++)
 	  conv(j) = coef_conv[j]->Evaluate(sip);
-	
+        */
+
 	fel_l2.CalcShape (sip.IP(), shape);
 	fel_l2.CalcMappedDShape (sip, dshape);
 	
@@ -257,7 +265,7 @@ namespace ngfem
 
 	conv_dshape *= sip.GetJacobiDet() * ir_vol[l].Weight();
 	
-	elmat  -= conv_dshape * Trans (shape);
+        elmat  -= conv_dshape * Trans (shape);
       }
     }//end of CalcElementMatrix
     
@@ -270,10 +278,10 @@ namespace ngfem
   class DGInnerFacet_ConvectionIntegrator : public FacetBilinearFormIntegrator
   {
   protected:
-    Array<CoefficientFunction *> coef_b;
+    Array<shared_ptr<CoefficientFunction> > coef_b;
   public:
-    DGInnerFacet_ConvectionIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetBilinearFormIntegrator(coeffs)
+    DGInnerFacet_ConvectionIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+    // : FacetBilinearFormIntegrator(coeffs)
     { 
       coef_b.SetSize(D);
       for (int i=0; i<D; i++)
@@ -282,18 +290,19 @@ namespace ngfem
 
     virtual ~DGInnerFacet_ConvectionIntegrator () { ; }
     
-    virtual bool BoundaryForm () const 
-    { return 0; }
+    virtual VorB VB() const
+    { return VOL; }
+    virtual xbool IsSymmetric () const { return false; }
     
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGInnerFacet_ConvectionIntegrator (coeffs);
     }
     
     virtual void CalcElementMatrix (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatMatrix<double> & elmat,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatMatrix<double> elmat,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGInnerFacet_ConvectionIntegrator::CalcElementMatrix - not implemented!");
     }
@@ -302,8 +311,8 @@ namespace ngfem
 			 const ElementTransformation & eltrans1, FlatArray<int> & ElVertices1,
 			 const FiniteElement & volumefel2, int LocalFacetNr2,
 			 const ElementTransformation & eltrans2, FlatArray<int> & ElVertices2,
-                         FlatMatrix<double> & elmat,
-                         LocalHeap & lh) const
+                         FlatMatrix<double> elmat,
+                         LocalHeap & lh) const 
     {
       static int timer = NgProfiler::CreateTimer ("DGInnerFacet_ConvectionIntegrator");
       if (LocalFacetNr2==-1) throw Exception ("DGInnerFacet_ConvectionIntegrator: LocalFacetNr2==-1");
@@ -318,7 +327,7 @@ namespace ngfem
       ELEMENT_TYPE eltype2 = volumefel2.ElementType();
       eltype2 = volumefel2.ElementType();
       int nd2 = fel2_l2->GetNDof();
-      int maxorder = max(fel1_l2->Order(),fel2_l2->Order());
+      int maxorder = max2(fel1_l2->Order(),fel2_l2->Order());
       
       elmat = 0.0;
 
@@ -407,10 +416,10 @@ namespace ngfem
   {
   protected:
     double alpha;   // interior penalyty
-    Array<CoefficientFunction *> coef_b;
+    Array<shared_ptr<CoefficientFunction> > coef_b;
   public:
-    DGBoundaryFacet_ConvectionIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetBilinearFormIntegrator(coeffs)
+    DGBoundaryFacet_ConvectionIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+    // : FacetBilinearFormIntegrator(coeffs)
     { 
       coef_b.SetSize(D);
       for (int i=0; i<D; i++)
@@ -418,28 +427,29 @@ namespace ngfem
     }
 
     virtual ~DGBoundaryFacet_ConvectionIntegrator () { ; }
-    
-    virtual bool BoundaryForm () const 
-    { return 1; }
+
+    virtual VorB VB() const
+    { return BND; }
+    virtual xbool IsSymmetric () const { return false; }
   
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGBoundaryFacet_ConvectionIntegrator (coeffs);
     }
     
     virtual void CalcElementMatrix (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatMatrix<double> & elmat,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatMatrix<double> elmat,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGBoundaryFacet_ConvectionIntegrator::CalcElementMatrix - not implemented!");
     }
     
     virtual void CalcFacetMatrix (const FiniteElement & volumefel, int LocalFacetNr,
-			 const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
- 			 const ElementTransformation & seltrans,
-                         FlatMatrix<double> & elmat,
-                         LocalHeap & lh) const
+                                  const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
+                                  const ElementTransformation & seltrans, FlatArray<int> & SElVertices,
+                                  FlatMatrix<double> elmat,
+                                  LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("DGBoundaryFacet_ConvectionIntegrator");
       NgProfiler::RegionTimer reg (timer);
@@ -507,39 +517,40 @@ namespace ngfem
   {
   protected:
     double alpha;   // interior penalyty
-    CoefficientFunction *coef_lam;
+    shared_ptr<CoefficientFunction> coef_lam;
   public:
-    DGBoundaryFacet_LaplaceIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetBilinearFormIntegrator(coeffs)
+    DGBoundaryFacet_LaplaceIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+    // : FacetBilinearFormIntegrator(coeffs)
     { 
       coef_lam  = coeffs[0];
       if (dgtype!=DG_FORMULATIONS::BO)
 	alpha = coeffs[1] -> EvaluateConst();
     }
-    
-    virtual bool BoundaryForm () const 
-    { return 1; }
+
+    virtual VorB VB() const
+    { return BND; }
+    virtual xbool IsSymmetric () const { return true; }
 
     virtual ~DGBoundaryFacet_LaplaceIntegrator () { ; }
 
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGBoundaryFacet_LaplaceIntegrator (coeffs);
     }
     
     virtual void CalcElementMatrix (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatMatrix<double> & elmat,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatMatrix<double> elmat,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGBoundaryFacet_LaplaceIntegrator::CalcElementMatrix - not implemented!");
     }
     
     virtual void CalcFacetMatrix (const FiniteElement & volumefel, int LocalFacetNr,
-			 const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
- 			 const ElementTransformation & seltrans,
-                         FlatMatrix<double> & elmat,
-                         LocalHeap & lh) const
+                                  const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
+                                  const ElementTransformation & seltrans, FlatArray<int> & SElVertices,
+                                  FlatMatrix<double> elmat,
+                                  LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("DGBoundaryFacet_LaplaceIntegrator boundary");
       NgProfiler::RegionTimer reg (timer);
@@ -638,31 +649,32 @@ namespace ngfem
   {
   protected:
     double alpha;   // interior penalyty
-    CoefficientFunction *coef_lam;
-    CoefficientFunction *coef_dir;
+    shared_ptr<CoefficientFunction> coef_lam;
+    shared_ptr<CoefficientFunction> coef_dir;
   public:
-    DGFacet_DirichletBoundaryIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetLinearFormIntegrator(coeffs)
+    DGFacet_DirichletBoundaryIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+      : FacetLinearFormIntegrator( /* coeffs */)
     { 
       coef_lam  = coeffs[0];
       coef_dir  = coeffs[1];
       alpha = coeffs[2] -> EvaluateConst();
     }
-    
-    virtual bool BoundaryForm () const 
-    { return 1; }
+
+
+    virtual VorB VB() const
+    { return BND; }
     
     virtual ~DGFacet_DirichletBoundaryIntegrator () { ; }
 
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGFacet_DirichletBoundaryIntegrator (coeffs);
     }
     
     virtual void CalcElementVector (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatVector<double> & elvec,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatVector<double> elvec,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGFacet_DirichletBoundaryIntegrator::CalcElementVector - not implemented!");
     }
@@ -670,7 +682,7 @@ namespace ngfem
     virtual void CalcFacetVector (const FiniteElement & volumefel, int LocalFacetNr,
 			 const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
 			 const ElementTransformation & seltrans,
-                         FlatVector<double> & elvec, LocalHeap & lh) const
+                         FlatVector<double> elvec, LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("DGFacet_DirichletBoundaryIntegrator");
       NgProfiler::RegionTimer reg (timer);
@@ -756,28 +768,28 @@ namespace ngfem
   class DGFacet_NeumannBoundaryIntegrator : public FacetLinearFormIntegrator
   {
   protected:
-    CoefficientFunction *coef_lam;
+    shared_ptr<CoefficientFunction> coef_lam;
   public:
-    DGFacet_NeumannBoundaryIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetLinearFormIntegrator(coeffs)
+    DGFacet_NeumannBoundaryIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+      : FacetLinearFormIntegrator( /* coeffs */ )
     { 
       coef_lam  = coeffs[0];
     }
-    
-    virtual bool BoundaryForm () const 
-    { return 1; }
+
+    virtual VorB VB() const
+    { return BND; }
 
     virtual ~DGFacet_NeumannBoundaryIntegrator () { ; }
 
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGFacet_NeumannBoundaryIntegrator (coeffs);
     }
     
     virtual void CalcElementVector (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatVector<double> & elvec,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatVector<double> elvec,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGFacet_NeumannBoundaryIntegrator::CalcElementVector - not implemented!");
     }
@@ -785,7 +797,7 @@ namespace ngfem
     virtual void CalcFacetVector (const FiniteElement & volumefel, int LocalFacetNr,
 			 const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
 			 const ElementTransformation & seltrans,
-                         FlatVector<double> & elvec, LocalHeap & lh) const
+                         FlatVector<double> elvec, LocalHeap & lh) const
     {
       static int timer = NgProfiler::CreateTimer ("DGFacet_NeumannBoundaryIntegrator");
 
@@ -847,11 +859,11 @@ namespace ngfem
   class DGFacet_ConvectionDirichletBoundaryIntegrator : public FacetLinearFormIntegrator
   {
   protected:
-    CoefficientFunction *coef_rob;
-    Array<CoefficientFunction *> coef_b;
+    shared_ptr<CoefficientFunction> coef_rob;
+    Array<shared_ptr<CoefficientFunction> > coef_b;
   public:
-    DGFacet_ConvectionDirichletBoundaryIntegrator (Array<CoefficientFunction*> & coeffs) 
-      : FacetLinearFormIntegrator(coeffs)
+    DGFacet_ConvectionDirichletBoundaryIntegrator (const Array<shared_ptr<CoefficientFunction>> & coeffs) 
+      : FacetLinearFormIntegrator( /* coeffs */)
     { 
       coef_b.SetSize(D);
       for (int i=0; i<D; i++)
@@ -859,20 +871,20 @@ namespace ngfem
       coef_rob = coeffs[D];
     }
 
-    virtual bool BoundaryForm () const 
-    { return 1; }
+    virtual VorB VB() const
+    { return BND; }
     
     virtual ~DGFacet_ConvectionDirichletBoundaryIntegrator () { ; }
 
-    static Integrator * Create (Array<CoefficientFunction*> & coeffs)
+    static Integrator * Create (const Array<shared_ptr<CoefficientFunction>> & coeffs)
     {
       return new DGFacet_ConvectionDirichletBoundaryIntegrator (coeffs);
     }
     
     virtual void CalcElementVector (const FiniteElement & fel,
-                                        const ElementTransformation & eltrans, 
-                                        FlatVector<double> & elvec,
-                                        LocalHeap & lh) const
+                                    const ElementTransformation & eltrans, 
+                                    FlatVector<double> elvec,
+                                    LocalHeap & lh) const
     {
       throw Exception("DGFacet_ConvectionDirichletBoundaryIntegrator::CalcElementVector - not implemented!");
     }
@@ -880,7 +892,7 @@ namespace ngfem
     virtual void CalcFacetVector (const FiniteElement & volumefel, int LocalFacetNr,
 			 const ElementTransformation & eltrans, FlatArray<int> & ElVertices,
 			 const ElementTransformation & seltrans,
-                         FlatVector<double> & elvec, LocalHeap & lh) const
+                         FlatVector<double> elvec, LocalHeap & lh) const
     {
        static int timer = NgProfiler::CreateTimer ("DGFacet_ConvectionDirichletBoundaryIntegrator");
       NgProfiler::RegionTimer reg (timer);
