@@ -1,22 +1,25 @@
+#define FILE_HDIVHOFE_CPP
+
 #include <fem.hpp>
-#include <thdivfe_impl.hpp>
 #include <hdivhofe.hpp>
+/*
+#include <thdivfe_impl.hpp>
 #include <hdivhofe_impl.hpp>
 #include <hdivhofefo.hpp>
-
+*/
 
 namespace ngfem
 {  
-
    
   //------------------------------------------------------------------------
   // HDivHighOrderFiniteElement
   //------------------------------------------------------------------------
 
+  /*
   template <int D>
   HDivHighOrderFiniteElement<D> ::
   HDivHighOrderFiniteElement (ELEMENT_TYPE aeltype)
-    : HDivFiniteElement<D> (aeltype, -1, -1)
+    : HDivFiniteElement<D> (-1, -1)
   {
     for (int i = 0; i < 8; i++)
       vnums[i] = i;
@@ -50,7 +53,6 @@ namespace ngfem
     for (int i = 0; i < of.Size(); i++)
       order_face[i] = INT<2>(of[i],of[i]);
     ComputeNDof();
-   
   }
   
   template <int D>
@@ -64,34 +66,42 @@ namespace ngfem
 
   template <int D>
   void HDivHighOrderFiniteElement<D>::
-  SetOrderInner (int oi)
-  {
-    order_inner = INT<3>(oi,oi,oi);
-    ComputeNDof();
-
-  }
-  
-  template <int D>
-  void HDivHighOrderFiniteElement<D>::
-  SetOrderInner (INT<3> oi)
+  SetOrderInner (INT<D> oi)
   {
     order_inner = oi;
     ComputeNDof();
   }
 
+  template <int D>
+  void HDivHighOrderFiniteElement<D> :: Print (ostream & ost) const
+  {
+    ELEMENT_TYPE et = this->ElementType();
+    ost << "HDivHighOrderFiniteElement<" << ElementTopology::GetElementName(et) << ">:" << endl;
+    if (D == 2)
+      for (int j = 0; j < ElementTopology::GetNEdges(et); j++)
+	ost << "order_edge[" << j << "] = " << order_edge[j] << endl;
+    else
+      for (int j = 0; j < ElementTopology::GetNFaces(et); j++)
+	ost << "order_face[" << j << "] = " << order_face[j] << endl;
+  }
+  */
+
 
   //------------------------------------------------------------------------
   // HDivHighOrderNormalFiniteElement
   //------------------------------------------------------------------------
+  /*
   template <int D>
   HDivHighOrderNormalFiniteElement<D> ::
-  HDivHighOrderNormalFiniteElement (ELEMENT_TYPE aeltype)
-    : HDivNormalFiniteElement<D> (aeltype, -1, -1)
+  HDivHighOrderNormalFiniteElement ()
+    : HDivNormalFiniteElement<D> (-1, -1)
   {
-    for (int i = 0; i < 4; i++)
-      vnums[i] = i;
+    // for (int i = 0; i < 4; i++)
+    // vnums[i] = i;
+    ;
   }
-
+  */
+  /*
   template <int D>
   void HDivHighOrderNormalFiniteElement<D>::
   SetVertexNumbers (FlatArray<int> & avnums)
@@ -100,7 +110,9 @@ namespace ngfem
     for (int i = 0; i < avnums.Size(); i++)
       vnums[i] = avnums[i];
   }
+  */
 
+  /*
   template <int D>
   void HDivHighOrderNormalFiniteElement<D>::
   SetOrderInner (int oi)
@@ -114,13 +126,14 @@ namespace ngfem
   {
     order_inner = oi;
   }
-
+  */
+  
   //------------------------------------------------------------------------
   // HDivHighOrderNormalSegm
   //------------------------------------------------------------------------
   template <class T_ORTHOPOL>
   HDivHighOrderNormalSegm<T_ORTHOPOL> :: HDivHighOrderNormalSegm (int aorder)
-    : HDivHighOrderNormalFiniteElement<1>(ET_SEGM)
+  // : HDivHighOrderNormalFiniteElement<1>()
   {
     order_inner = INT<2>(aorder,aorder);
     ComputeNDof();
@@ -133,35 +146,33 @@ namespace ngfem
     order = order_inner[0];
   }
 
+  /*
   template <class T_ORTHOPOL>
   void HDivHighOrderNormalSegm<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip,
 							 FlatVector<> shape) const
   {
-    double x = ip(0);
+    AutoDiff<1> x (ip(0), 0);
+    AutoDiff<1> lam[2] = { x, 1-x };
 
-    int p = order_inner[0];
+    ArrayMem<AutoDiff<1>,10> adpol1(order);
+	
+    INT<2> e = ET_trait<ET_SEGM>::GetEdgeSort (0, vnums);	  
+    
+    shape[0] = -lam[e[0]].DValue(0);
 
-    int fac = 1;
-    if (vnums[0] > vnums[1])
-      {
-	fac *= -1;
-	x=1-x;
+    int p = order_inner[0]; 
+    // LegendrePolynomial::
+    IntLegNoBubble::
+      EvalMult (p-1, 
+                lam[e[1]]-lam[e[0]], lam[e[0]]*lam[e[1]], adpol1);
+    
+    for(int j = 0; j < p; j++) 	      
+      shape[j+1] = -adpol1[j].DValue(0);
       }
+  */
 
 
-    ArrayMem<double, 10> rec_pol(p+1), leg(2);
 
-    MatrixFixWidth<2> DExt(p);
-    T_ORTHOPOL::CalcTrigExtDeriv(p+1, 1-2*x, 0, DExt);
-
-    //IntegratedLegendrePolynomial (order_inner, 1-2*x, rec_pol);
-    // LegendrePolynomial (order_inner, 2*x-1, rec_pol);
-    //T_ORTHOPOL::CalcDeriv(order_inner+1, 1-2*x, rec_pol);
-    shape(0) = -1.*fac;
-    for(int j=1; j<=p;j++)
-      shape(j) = 2.* fac * DExt(j-1,0);
-
-  }
   template class HDivHighOrderNormalSegm<IntegratedLegendreMonomialExt>;
   template class HDivHighOrderNormalSegm<TrigExtensionMonomial>;
   //template class HDivHighOrderNormalSegm<TrigExtensionOptimal>;
@@ -172,7 +183,7 @@ namespace ngfem
   //------------------------------------------------------------------------
   template <class T_ORTHOPOL>
   HDivHighOrderNormalQuad<T_ORTHOPOL> :: HDivHighOrderNormalQuad (int aorder)
-    : HDivHighOrderNormalFiniteElement<2>(ET_QUAD)
+  // : HDivHighOrderNormalFiniteElement<2>()
   {
     order_inner = INT<2>(aorder,aorder);
     ComputeNDof();
@@ -182,10 +193,11 @@ namespace ngfem
   void HDivHighOrderNormalQuad<T_ORTHOPOL> :: ComputeNDof()
   {
     ndof = (order_inner[0] < 0) ? 0 : (1 + order_inner[0]*order_inner[1] + order_inner[0] + order_inner[1]);
-    order = max(order_inner[0],order_inner[1]);
+    order = max2(order_inner[0],order_inner[1]);
     order++; // order used for numerical integration
   }
 
+#ifdef OLD
   template <class T_ORTHOPOL>
   void HDivHighOrderNormalQuad<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip,
 							 FlatVector<> shape) const
@@ -201,10 +213,9 @@ namespace ngfem
     int ii = 1;
 
     INT<2> p = order_inner;
-    // int pp = max(p[0],p[1]); 
+    // int pp = max2(p[0],p[1]); 
     
     ArrayMem<AutoDiff<2>,20> pol_xi(p[0]+1), pol_eta(p[1]+1);
-    AutoDiff<2> polprod;
 
     int fmax = 0;
     for (int j = 1; j < 4; j++)
@@ -226,8 +237,12 @@ namespace ngfem
 
     shape(0) = fac;
 
+    /*
     T_ORTHOPOL::Calc(p[0]+1, xi,pol_xi);
     T_ORTHOPOL::Calc(p[1]+1,eta,pol_eta);
+    */
+    IntLegNoBubble::EvalMult (p[0]-1, xi, 1-xi*xi, pol_xi);
+    IntLegNoBubble::EvalMult (p[1]-1, eta, 1-eta*eta, pol_eta);
 
     // Typ 1
     for (int k = 0; k < p[0]; k++)
@@ -240,7 +255,7 @@ namespace ngfem
     for (int k = 0; k < p[1]; k++)
       shape(ii++)   = -xi.DValue(0)*pol_eta[k].DValue(1) + xi.DValue(1)*pol_eta[k].DValue(0);
   }
-
+#endif 
 
   template class HDivHighOrderNormalQuad<IntegratedLegendreMonomialExt>;
   template class HDivHighOrderNormalQuad<TrigExtensionMonomial>;
@@ -253,7 +268,7 @@ namespace ngfem
 
   template <class T_ORTHOPOL>
   HDivHighOrderNormalTrig<T_ORTHOPOL> :: HDivHighOrderNormalTrig (int aorder)
-    : HDivHighOrderNormalFiniteElement<2>(ET_TRIG)
+  // : HDivHighOrderNormalFiniteElement<2>()
   {
     order_inner = INT<2>(aorder,aorder);
     ComputeNDof();
@@ -268,6 +283,7 @@ namespace ngfem
     order++;
   }
 
+#ifdef OLD
   template <class T_ORTHOPOL>
   void HDivHighOrderNormalTrig<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip,
 							 FlatVector<> shape) const
@@ -287,16 +303,15 @@ namespace ngfem
     dlami(2,0) = -1.;
     dlami(2,1) = -1.;
 
-    int ii, i, j, k, l, is, ie, iop;
+    int ii, is, ie, iop;
 
     int p = order_inner[0];
     ii = 1;
 
     int fav[3];
-    for(i=0;i<3;i++) fav[i] = i;
+    for(int i=0;i<3;i++) fav[i] = i;
 
-
-    //Sort vertices  first edge op minimal vertex
+    //Sort vertices first edge op minimal vertex
     int fswap = 1;
     if(vnums[fav[0]] > vnums[fav[1]]) { swap(fav[0],fav[1]); fswap *= -1; }
     if(vnums[fav[1]] > vnums[fav[2]]) { swap(fav[1],fav[2]); fswap *= -1; }
@@ -309,7 +324,7 @@ namespace ngfem
     AutoDiff<2> lo = lami[iop];
 
     //AutoDiff<3> lsle = lami[is]*lami[ie];
-    for (j = 0; j < 2; j++)
+    for (int j = 0; j < 2; j++)
       {
         ls.DValue(j) = dlami(is,j);
         le.DValue(j) = dlami(ie,j);
@@ -317,19 +332,21 @@ namespace ngfem
       }
 
     Vec<2> nedelec;
-    for (j = 0; j < 2; j++)
+    for (int j = 0; j < 2; j++)
       nedelec(j) = ls.Value()*le.DValue(j) - le.Value()*ls.DValue(j);
 
-    AutoDiff<2> lsle=ls*le;
-
     // RT_0-normal low order shapes
     shape(0) = fswap;
 
+
+    /*
+    AutoDiff<2> lsle=ls*le;
     Vec<2> grad1, grad2;
     ArrayMem<AutoDiff<2>, 100> ad_rec_pol1(p);
     ArrayMem<AutoDiff<2>, 100> ad_rec_pol2(p);
 
     ScaledLegendrePolynomial(p-1, le-ls, 1-lo, ad_rec_pol1);
+    // IntLegNoBubble::EvalScaled (p-1, le-ls, 1-lo, ad_rec_pol1); 
     LegendrePolynomial(p-1, 2*lo-1, ad_rec_pol2);
 
     for (k = 0; k <= p-1; k++)
@@ -361,7 +378,48 @@ namespace ngfem
           grad2(j) = ad_rec_pol2[k].DValue(j);
         shape(ii) = (grad2(0)*nedelec(1) - grad2(1)*nedelec(0)) + ad_rec_pol2[k].Value()*curlned;
       }       
+    */
+
+    ArrayMem<AutoDiff<2>, 20> adpol1(p);
+    ArrayMem<AutoDiff<2>, 20> adpol2(p);
+
+    IntLegNoBubble::EvalScaledMult (p-1, le-ls, le+ls, ls*le, adpol1); 
+    /*
+    // Typ 1
+    for (int k = 0; k <= p-1; k++)
+      {
+        IntegratedJacobiPolynomialAlpha jac(2*k+3);
+        jac.EvalMult(p-1-k, 2*lo-1, lo, adpol2);
+	for (int l = 0; l <= p-1-k; l++, ii++)
+          shape(ii) = Cross (adpol2[l], adpol1[k]).DValue(0);
+      }
+    */
+
+    // Vector<> hshape(shape.Size());
+    // ArrayMem<double, 20> adpol2b(p);
+    // hshape = 0.0;
+    // int iii = 1;
+    for (int k = 0; k <= p-1; k++)
+      {
+        JacobiPolynomialAlpha jac(2*k+3);
+	double factor = Cross (lo, adpol1[k]).DValue(0);
+        jac.EvalMult(p-1-k, 2*lo.Value()-1, factor, shape+ii);
+	ii += p-k;
+      }
+    // cout << "shape = " << endl << shape << endl << " hshape = " << hshape << endl;
+
+    IntegratedJacobiPolynomialAlpha jac(3);
+    jac.EvalMult(p-1, 2*lo-1, lo, adpol2);
+
+    // Typ 2
+    double curlned;
+    curlned = 2.* (ls.DValue(0)*le.DValue(1) - ls.DValue(1)*le.DValue(0));
+    for (int k = 0; k <= p-1; k++, ii++)
+      shape(ii) = adpol2[k].DValue(0)*nedelec(1) - adpol2[k].DValue(1)*nedelec(0) 
+        + adpol2[k].Value()*curlned;
   }
+#endif
+
   template class HDivHighOrderNormalTrig<IntegratedLegendreMonomialExt>;
   template class HDivHighOrderNormalTrig<TrigExtensionMonomial>;
   // template class HDivHighOrderNormalTrig<TrigExtensionOptimal>;
@@ -376,13 +434,8 @@ namespace ngfem
 
 
 
-
-
-
-
-
   template <ELEMENT_TYPE ET>
-  void T_HDivHighOrderFiniteElement<ET> :: 
+  void HDivHighOrderFE<ET> :: 
   ComputeNDof()
   {
     if (DIM == 2)
@@ -390,9 +443,15 @@ namespace ngfem
         if (only_ho_div)
 	  {
 	    if (ET == ET_TRIG)
-	      ndof = order_inner[0]*(order_inner[0]+1)/2 - 1;
+              {
+                ndof = order_inner[0]*(order_inner[0]+1)/2 - 1;
+                order = order_inner[0];
+              }
 	    else
-	      ndof = order_inner[0]*order_inner[1] + order_inner[0] + order_inner[1];
+              {
+                ndof = order_inner[0]*order_inner[1] + order_inner[0] + order_inner[1];
+                order = max2(order_inner[0], order_inner[1])+1;
+              }
 	    return;
 	  }
         else
@@ -400,7 +459,7 @@ namespace ngfem
 	    ndof = ET_trait<ET>::N_EDGE;
 	    
 	    for(int i = 0; i < ET_trait<ET>::N_EDGE; i++)
-	      ndof += order_edge[i];
+	      ndof += order_facet[i][0];
 	    
 	    if (ET == ET_TRIG)
 	      {
@@ -411,6 +470,8 @@ namespace ngfem
 		    else
 		      ndof += order_inner[0]*order_inner[0]-1;
 		  }
+                if (RT && order_inner[0] > 0)
+		  ndof += order_inner[0] + 1;
 	      }
 	    else
 	      {  // quad
@@ -426,13 +487,15 @@ namespace ngfem
 
         order = 0; 
         for (int i = 0; i < ET_trait<ET>::N_EDGE; i++)
-          if (order_edge[i] > order)
-            order = order_edge[i];
+          if (order_facet[i][0] > order)
+            order = order_facet[i][0];
         
         for (int j = 0; j < 2; j++)
           if (order_inner[j] > order) 
             order = order_inner[0];
-        order++;       // ????
+
+        if (ET != ET_TRIG) order++;
+        if(RT) order ++;
       }
     else
       {
@@ -459,6 +522,9 @@ namespace ngfem
                   ndof += (order_inner[0]-1)*(order_inner[0]+1)*(order_inner[0]+2)/2;
               */
               break;
+            case ET_HEX:
+              ndof = 3*(p+1)*(p+1)*p;
+              break; 
             }
         }
         else
@@ -467,7 +533,7 @@ namespace ngfem
 
           for(int i = 0; i < ET_trait<ET>::N_FACE; i++)
             {
-              INT<2> p = order_face[i];
+              INT<2> p = order_facet[i];
               if (ET_trait<ET>::FaceType(i) == ET_TRIG)
                 ndof += (p[0]*p[0]+3*p[0])/2;
               else
@@ -485,6 +551,8 @@ namespace ngfem
                 ndof += pc*(pc+1)*(pc-1)/3 + pc*(pc-1)/2;
               if(p > 1 && !ho_div_free) 
                 ndof += p*(p+1)*(p-1)/6 + p*(p-1)/2 + p-1;
+	      if(RT && p >= 1)
+		ndof += (p+1)*(p+2)/2;
               break;
 
             case ET_PRISM:
@@ -502,6 +570,11 @@ namespace ngfem
               */
                 }
               break;
+            case ET_HEX:
+              ndof += 3*(p+1)*(p+1)*p;       
+              if (ho_div_free)     
+                ndof -= p*p*p+3*p*p+3*p;
+              break; 
             }
         }
 	// cout << "ndof, tot = " << ndof << endl;
@@ -509,7 +582,7 @@ namespace ngfem
         order = 0; 
         for (int i = 0; i < ET_trait<ET>::N_FACE; i++)
           {
-            int p = max(order_face[i][0], order_face[i][1]);
+            int p = max2(order_facet[i][0], order_facet[i][1]);
             if (p > order) order = p;
           }
 
@@ -517,16 +590,145 @@ namespace ngfem
         if (pi > order) order = pi;
 
 	if (ET != ET_TET) order++;
+	
+	if(RT) order ++;
       }
   }
 
 
+  template <ELEMENT_TYPE ET>
+  tuple<int,int,int,int> HDivHighOrderFE<ET> :: 
+  GetNDofVEFC() const
+  {
+    int nv=0, ne = 0, nf = 0, nc = 0;
+    if (DIM == 2)
+      {
+        // if (only_ho_div)
+	//   {
+	//     if (ET == ET_TRIG)
+        //       {
+        //         ndof = order_inner[0]*(order_inner[0]+1)/2 - 1;
+        //         order = order_inner[0];
+        //       }
+	//     else
+        //       {
+        //         ndof = order_inner[0]*order_inner[1] + order_inner[0] + order_inner[1];
+        //         order = max2(order_inner[0], order_inner[1])+1;
+        //       }
+	//     return;
+	//   }
+        // else
+	  {
+	    ne = ET_trait<ET>::N_EDGE;
+	    
+	    for(int i = 0; i < ET_trait<ET>::N_EDGE; i++)
+	      ne += order_facet[i][0];
+	    
+	    if (ET == ET_TRIG)
+	      {
+		if (order_inner[0] > 1)
+		  { 
+		    if (ho_div_free)
+		      nf += order_inner[0]*(order_inner[0]-1)/2;
+		    else
+		      nf += order_inner[0]*order_inner[0]-1;
+		  }
+                if (RT && order_inner[0] > 0)
+		  nf += order_inner[0] + 1;
+	      }
+	    else
+	      {  // quad
+		INT<2> p(order_inner[0], order_inner[1]);
+		
+		int ni = ho_div_free
+		  ? p[0]*p[1] 
+		  : 2*p[0]*p[1] + p[0] + p[1];
+		
+		nf += ni; 
+	      }
+	  }
+      }
+    else
+      {
 
+        int p = order_inner[0];
+        int pc = order_inner[0]; // should be order_inner_curl!!!  
+        int pz = p; // order_inner[2];
 
+        // if (only_ho_div){
+        //   switch (ET)
+        //     {   
+        //     case ET_TRIG: case ET_QUAD: // for the compiler
+        //       break;
+        //     case ET_TET: 
+        //         ndof = p*(p+1)*(p-1)/6 + p*(p-1)/2 + p-1;
+        //       break;
+        //     case ET_PRISM:
+        //       if (order_inner[0]>0 )
+        //         ndof = (p+1)*(p+2)*(pz+1)/2 - 1;
+        //       /*
+        //           // inner_dof horizontal
+        //           ndof += (order_inner[0]+1)*(3*(order_inner[0]-1)+(order_inner[0]-2)*(order_inner[0]-1));
+        //           // inner dof vertical
+        //           ndof += (order_inner[0]-1)*(order_inner[0]+1)*(order_inner[0]+2)/2;
+        //       */
+        //       break;
+        //     case ET_HEX:
+        //       ndof = 3*(p+1)*(p+1)*p;
+        //       break; 
+        //     }
+        // }
+        // else
+        {
+          nf = ET_trait<ET>::N_FACE;
+
+          for(int i = 0; i < ET_trait<ET>::N_FACE; i++)
+            {
+              INT<2> p = order_facet[i];
+              if (ET_trait<ET>::FaceType(i) == ET_TRIG)
+                nf += (p[0]*p[0]+3*p[0])/2;
+              else
+                nf +=  p[0]*p[1] + p[0] + p[1];
+            }
+
+          switch (ET)
+            {   
+            case ET_TRIG: case ET_QUAD: // for the compiler
+              break;
+
+            case ET_TET: 
+              if(pc > 1) 
+                nc += pc*(pc+1)*(pc-1)/3 + pc*(pc-1)/2;
+              if(p > 1 && !ho_div_free) 
+                nc += p*(p+1)*(p-1)/6 + p*(p-1)/2 + p-1;
+	      if(RT && p >= 1)
+		nc += (p+1)*(p+2)/2;
+              break;
+
+            case ET_PRISM:
+              // SZ: ATTENTION PRISM up to now only using for order_inner[0] !!  
+              if (order_inner[0]>0 )
+                {
+                  nc += (p+2)*p*(pz+1) + (p+1)*(p+2)*pz/2;
+                  if (ho_div_free)
+                    nc -= (p+1)*(p+2)*(pz+1)/2 - 1;
+                }
+              break;
+            case ET_HEX:
+              nc += 3*(p+1)*(p+1)*p;       
+              if (ho_div_free)     
+                nc -= p*p*p+3*p*p+3*p;
+              break; 
+            }
+        }
+      }
+
+    return { nv, ne, nf, nc };
+  }
 
 
   template <ELEMENT_TYPE ET>
-  void T_HDivHighOrderFiniteElement<ET> :: 
+  void HDivHighOrderFE<ET> :: 
   GetFacetDofs(int i, Array<int> & dnums) const
   {
     dnums.SetSize(0);
@@ -538,10 +740,10 @@ namespace ngfem
 	int nf = 0;
 	switch (ElementTopology::GetFacetType(ET,j))
 	  {
-	  case ET_SEGM: nf = order_edge[j]; break;
+	  case ET_SEGM: nf = order_facet[j][0]; break;
 	    // case ET_TRIG: nf = (sqr (order_face[0])+3*order_face[0])/2; break;
-	  case ET_TRIG: nf = (order_face[j][0]+1)*(order_face[j][0]+2)/2-1; break;
-	  case ET_QUAD: nf = (order_face[j][0]+1)*(order_face[j][1]+1)-1; break;
+	  case ET_TRIG: nf = (order_facet[j][0]+1)*(order_facet[j][0]+2)/2-1; break;
+	  case ET_QUAD: nf = (order_facet[j][0]+1)*(order_facet[j][1]+1)-1; break;
 	  default:
 	    throw Exception("what kind of face is that ?");
 	  }
@@ -558,14 +760,6 @@ namespace ngfem
   }
 
 
-
-
-  
-
-
-
-
-
   //------------------------------------------------------------------------
   // HDivHighOrderTrig
   //------------------------------------------------------------------------
@@ -727,7 +921,7 @@ namespace ngfem
         
           T_FACESHAPES::CalcSplitted (p+2, xi, eta, zeta, adpol1, adpol2); 
 
-          // Compability WITH TRIG!! 
+          // Compatibility WITH TRIG!! 
           for (int k = 0; k < adpol1.Size(); k++)
             adpol1[k] *= 0.5; 
             
@@ -746,7 +940,7 @@ namespace ngfem
     // cell-based shapes 
     int p = order_inner[0];
     int pc = order_inner[0]; // should be order_inner_curl  
-    int pp = max(p,pc); 
+    int pp = max2(p,pc); 
     if ( pp >= 2 )
       {
         T_INNERSHAPES::CalcSplitted(pp+2, lami[0]-lami[3], lami[1], lami[2], adpol1, adpol2, adpol3 );
@@ -967,7 +1161,7 @@ namespace ngfem
 	AutoDiff<1> zeta = muz[f]-muz[f2];
 
 
-        int pp = max(p[0],p[1]); 
+        int pp = max2(p[0],p[1]); 
 	T_ORTHOPOL::CalcTrigExt(pp+1,xi,eta,adpolxy1);
 	T_ORTHOPOL::Calc(pp+1,zeta,adpolz);
 
@@ -1402,6 +1596,8 @@ namespace ngfem
   //------------------------------------------------------------------------
 
 
+#ifdef HDIVHEX
+
   HDivHighOrderFE<ET_HEX> :: HDivHighOrderFE (int aorder)
     : HDivHighOrderFiniteElement<3>(ET_HEX)
   {
@@ -1430,16 +1626,16 @@ namespace ngfem
     //3*p*(p+1)*(p+1);
     ndof += ndof_inner;
 
-    order = 0; // max(order_face_normal,order_inner);
+    order = 0; // max2(order_face_normal,order_inner);
     for (int i = 0; i < 6; i++)
       {
-        int pp = max(order_face[i][0],order_face[i][1]);
+        int pp = max2(order_face[i][0],order_face[i][1]);
 	if (pp > order)
 	  order = pp;
 
       }
-    int pp = max(order_inner[0],max(order_inner[1],order_inner[2]));
-    order=max(pp,order); 
+    int pp = max2(order_inner[0],max2(order_inner[1],order_inner[2]));
+    order=max2(pp,order); 
 
     if (order == 0) order = 1;
     order++; // integration order
@@ -1447,7 +1643,7 @@ namespace ngfem
 
 
   void HDivHighOrderFE<ET_HEX> :: CalcShape (const IntegrationPoint & ip,
-						  FlatMatrixFixWidth<3> shape) const
+                                             SliceMatrix<> shape) const
   {
 
     int i, j, k, l, m;
@@ -1627,7 +1823,7 @@ namespace ngfem
   }
 
   void HDivHighOrderFE<ET_HEX> :: CalcDivShape (const IntegrationPoint & ip,
-                                                FlatVector<> divshape) const
+                                                SliceVector<> divshape) const
   {
     int i, j, k; 
     AutoDiff<3> x (ip(0),0);
@@ -1745,29 +1941,22 @@ namespace ngfem
     for (int i=0; i<nf; i++)
       dnums.Append(base+i);
   }   
+#endif
 
 
-  template class  HDivHighOrderFiniteElement<2>;
-  template class  HDivHighOrderFiniteElement<3>;
+  // template class  HDivHighOrderFiniteElement<2>;
+  // template class  HDivHighOrderFiniteElement<3>;
   template class  HDivHighOrderNormalFiniteElement<1>;
   template class  HDivHighOrderNormalFiniteElement<2>;
 
+  /*
   template class T_HDivHighOrderFiniteElement<ET_TRIG>;
   template class T_HDivHighOrderFiniteElement<ET_QUAD>;
   template class T_HDivHighOrderFiniteElement<ET_TET>;
   template class T_HDivHighOrderFiniteElement<ET_PRISM>;
+  */
 
 
-//  template class HDivHighOrderFE<ET_TRIG>;
-//  template class HDivHighOrderFE<ET_QUAD>;
-  template class HDivHighOrderFE<ET_TET>;
-  template class HDivHighOrderFE<ET_PRISM>;
-
-  template class T_HDivFiniteElement<HDivHighOrderFE_Shape<ET_TRIG>, ET_TRIG>;
-  template class T_HDivFiniteElement<HDivHighOrderFE_Shape<ET_QUAD>, ET_QUAD>;
-  template class T_HDivFiniteElement<HDivHighOrderFE_Shape<ET_TET>, ET_TET>;
-  template class T_HDivFiniteElement<HDivHighOrderFE_Shape<ET_PRISM>, ET_PRISM>;
-
 
 
   template class HDivHighOrderFEFO<ET_TRIG,1>;
