@@ -35,12 +35,11 @@ namespace ngla
     SetMaxSteps (200); 
     SetInitialize (1);
     printrates = 0;
-    sh = NULL;
     useseed = false;
   }
   
 
-  KrylovSpaceSolver :: KrylovSpaceSolver (const BaseMatrix & aa)
+  KrylovSpaceSolver :: KrylovSpaceSolver (shared_ptr<BaseMatrix> aa)
   {
     //  SetSymmetric();
     
@@ -50,13 +49,12 @@ namespace ngla
     SetMaxSteps (200);
     SetInitialize (1);
     printrates = 0;
-    sh = NULL;
     useseed = false;
   }
 
 
 
-  KrylovSpaceSolver :: KrylovSpaceSolver (const BaseMatrix & aa, const BaseMatrix & ac)
+  KrylovSpaceSolver :: KrylovSpaceSolver (shared_ptr<BaseMatrix> aa, shared_ptr<BaseMatrix> ac)
   {
     //  SetSymmetric();
     
@@ -66,18 +64,10 @@ namespace ngla
     SetMaxSteps (200);
     SetInitialize (1);
     printrates = 0;
-    sh = NULL;
     useseed = false;
   }
 
- 
-  BaseVector * KrylovSpaceSolver :: CreateVector () const
-  {
-    return 0; // return a->CreateVector();
-  }
-
-
-  template <class SCAL>
+    template <class SCAL>
   void BruteInnerProduct(const BaseVector & a, const BaseVector & b, Vector<SCAL> & result, const int start = 0)
   {
     const SCAL * pa;
@@ -144,12 +134,11 @@ namespace ngla
     try
       {
 	// Solve A u = f
-	if(sh)
-	  sh->SetThreadPercentage(0);
- 
-	BaseVector & d = *f.CreateVector();
-	BaseVector & w = *f.CreateVector();
-	BaseVector & s = *f.CreateVector();
+        BaseStatusHandler::SetThreadPercentage(0);
+
+	auto d = f.CreateVector();
+	auto w = f.CreateVector();
+	auto s = f.CreateVector();
 
 	int n = 0;
 	Vector<SCAL> al(dim), be(dim), wd(dim), wdn(dim), kss(dim);
@@ -185,7 +174,7 @@ namespace ngla
 	double lerr = log(err);
 	
 
-	while (n++ < maxsteps && L2Norm(wdn) > err && !(sh && sh->ShouldTerminate()))
+	while (n++ < maxsteps && L2Norm(wdn) > err && !(BaseStatusHandler::ShouldTerminate()))
 	  {
 	    w = (*a) * s;
 
@@ -233,28 +222,29 @@ namespace ngla
 	    //s += w;
 
 	    if (printrates ) cout << IM(1) << n << " " << sqrt(L2Norm (wdn)) << endl;
-	    if(sh)
-	      sh->SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
+            BaseStatusHandler::SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
 						(lwstart-log(L2Norm(wdn)))/(lwstart-lerr)));
 	  } 
 	
 	const_cast<int&> (steps) = n;
 	
+        /*
 	delete &d;
 	delete &w;
 	delete &s;
+        */
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in CGSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in CGSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in CGSolver::Mult\n"));
+      }
   }
 
 
@@ -264,21 +254,22 @@ namespace ngla
     try
       {
 	// Solve A u = f
-	if(sh)
-	  sh->SetThreadPercentage(0);
+        BaseStatusHandler::SetThreadPercentage(0);
  
 	SCAL * pl;
 	const SCAL * pr;
 	int i;
 
-	BaseVector & d = *f.CreateVector();
+	auto d = f.CreateVector();
 
 	BaseMatrix * smalla;
-
+        /*
 	if(dynamic_cast< const SparseMatrixSymmetricTM<SCAL> *>(a))
 	  smalla = new SparseMatrixSymmetric<SCAL,SCAL>(*dynamic_cast< const SparseMatrixSymmetricTM<SCAL> *>(a));
-	else if (dynamic_cast< const SparseMatrixTM<SCAL> *>(a))
-	  smalla = new SparseMatrix<SCAL,SCAL>(*dynamic_cast< const SparseMatrixTM<SCAL> *>(a));
+	else
+        */
+        if (dynamic_cast< const SparseMatrixTM<SCAL> *>(a.get()))
+	  smalla = new SparseMatrix<SCAL,SCAL>(*dynamic_cast< const SparseMatrixTM<SCAL> *>(a.get()));
 	else
 	  throw Exception("Assumption about bilinearform wrong.");
 
@@ -375,7 +366,7 @@ namespace ngla
 	    
 
 
-	    while (n++ < maxsteps && Abs(wdn) > err[seed] && !(sh && sh->ShouldTerminate()))
+	    while (n++ < maxsteps && Abs(wdn) > err[seed] && !(BaseStatusHandler::ShouldTerminate()))
 	      {
 		//if(smalla)
 		w = (*smalla)  * s;
@@ -479,14 +470,12 @@ namespace ngla
 		s += w;
 
 		if (printrates ) cout << IM(1) << n << " (block " << seed+1 << ") " << sqrt (Abs (wdn)) << endl;
-		if(sh)
-		  sh->SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
+                BaseStatusHandler::SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
 						    (lwstart-log(Abs(wdn)))/(lwstart-lerr)));
 	      } 
 	  }
 	const_cast<int&> (steps) = n;
 	
-	delete &d;
 	/*
 	if(!smalla)
 	  {
@@ -497,16 +486,16 @@ namespace ngla
 	delete smalla;
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in CGSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in CGSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in CGSolver::Mult\n"));
+      }
   }
 
 
@@ -561,13 +550,13 @@ namespace ngla
     try
       {
 	// Solve A u = f
-	if(sh)
-	  sh->SetThreadPercentage(0);
+        BaseStatusHandler::SetThreadPercentage(0);
  
-	BaseVector & d = *f.CreateVector();
-	BaseVector & w = *f.CreateVector();
-	BaseVector & s = *f.CreateVector();
-
+        auto w = u.CreateVector();
+        auto s = u.CreateVector();
+        auto d = f.CreateVector();
+        auto as = f.CreateVector();
+        
 	int n = 0;
 	SCAL al, be, wd, wdn, kss;
 	double err;
@@ -580,6 +569,7 @@ namespace ngla
 	  {
 	    d = f - (*a) * u;
 	  }
+
 	if (c)
 	  w = (*c) * d;
 	else
@@ -599,25 +589,21 @@ namespace ngla
 	double lwstart = log(Abs(wdn));
 	double lerr = log(err);
 	
-	while (n++ < maxsteps && Abs(wdn) > err && !(sh && sh->ShouldTerminate()))
+	while (n++ < maxsteps && Abs(wdn) > err && !(BaseStatusHandler::ShouldTerminate()))
 	  {
-	    w = (*a) * s;
-
+	    as = (*a) * s;
 	    wd = wdn;
-
-	    kss = S_InnerProduct<IPTYPE> (s, w);
+	    kss = S_InnerProduct<IPTYPE> (s, as);
 	    if (kss == 0.0) break;
 	    
 	    al = wd / kss;
-	    
 	    u += al * s;
-	    d -= al * w;
-
+	    d -= al * as;
+            
 	    if (c)
 	      w = (*c) * d;
 	    else
 	      w = d;
-	    
 	    wdn = S_InnerProduct<IPTYPE> (d, w);
 
 	    be = wdn / wd;
@@ -626,28 +612,23 @@ namespace ngla
 	    s += w;
 
 	    if (printrates ) cout << IM(1) << n << " " << sqrt (Abs (wdn)) << endl;
-	    if ( sh )
-	      sh->SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
+            BaseStatusHandler::SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
 						(lwstart-log(Abs(wdn)))/(lwstart-lerr)));
 	  } 
 	
 	const_cast<int&> (steps) = n;
-	
-	delete &d;
-	delete &w;
-	delete &s;
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in CGSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in CGSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in CGSolver::Mult\n"));
+      }
   }
 
 
@@ -661,17 +642,16 @@ namespace ngla
     try
       {
 	// Solve A u = f
-	if(sh)
-	  sh->SetThreadPercentage(0);
+        BaseStatusHandler::SetThreadPercentage(0);
  
-	BaseVector & r = *f.CreateVector();
-	BaseVector & r_tilde = *f.CreateVector();
-	BaseVector & p = *f.CreateVector();
-	BaseVector & p_tilde = *f.CreateVector();
-	BaseVector & s = *f.CreateVector();
-	BaseVector & s_tilde = *f.CreateVector();
-	BaseVector & t = *f.CreateVector();
-	BaseVector & v = *f.CreateVector();
+	auto r = f.CreateVector();
+	auto r_tilde = f.CreateVector();
+	auto p = f.CreateVector();
+	auto p_tilde = f.CreateVector();
+	auto s = f.CreateVector();
+	auto s_tilde = f.CreateVector();
+	auto t = f.CreateVector();
+	auto v = f.CreateVector();
 
 	int n = 0;
 	SCAL rho_old, rho_new, beta, alpha, omega;
@@ -697,7 +677,8 @@ namespace ngla
 
 	v = (*a) * p_tilde;
 	alpha = rho_new / S_InnerProduct<IPTYPE> (r_tilde, v);
-	s = r - alpha * v;
+	s = r;
+	s -= alpha * v;
 
 	err_i = L2Norm(s);
 	if (c)
@@ -709,7 +690,8 @@ namespace ngla
 
 	omega = S_InnerProduct<IPTYPE> (t, s) / S_InnerProduct<IPTYPE> (t, t);
 	u += alpha * p_tilde + omega * s_tilde;
-	r = s - omega * t;
+	r = s;
+	r -= omega * t;
 
 	err_i = L2Norm(r);
 	if (printrates) cout << IM(1) << "0 " << err_i << endl;
@@ -724,12 +706,14 @@ namespace ngla
 	double lerr = log(err);
 	
 
-	while (n++ < maxsteps && err_i > err && !(sh && sh->ShouldTerminate()))
+	while (n++ < maxsteps && err_i > err && !(BaseStatusHandler::ShouldTerminate()))
 	  {
 	    rho_old = rho_new;
 	    rho_new = S_InnerProduct<IPTYPE>(r_tilde, r);
 	    beta = (rho_new / rho_old ) * ( alpha / omega );
-	    p = r + beta * ( p - omega * v );
+	    p = r;
+	    p += beta * p;
+	    p -= beta*omega * v;
 
 	    if (c)
 	      p_tilde = (*c) * p;
@@ -738,7 +722,8 @@ namespace ngla
 	    
 	    v = (*a) * p_tilde;
 	    alpha = rho_new / S_InnerProduct<IPTYPE> (r_tilde, v);
-	    s = r - alpha * v;
+	    s = r;
+	    s -= alpha * v;
 
 	    err_i = L2Norm(s);
 	    u += alpha * p_tilde;
@@ -757,38 +742,29 @@ namespace ngla
 	    
 	    omega = S_InnerProduct<IPTYPE> (t, s) / S_InnerProduct<IPTYPE> (t, t);
 	    u +=  omega * s_tilde;
-	    r = s - omega * t;
+	    r = s;
+	    r -= omega * t;
 
 	    err_i = L2Norm(r);
 
 	    if (printrates ) cout << IM(1) << n << " " << err_i << endl;
-	    if(sh)
-	      sh->SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
+            BaseStatusHandler::SetThreadPercentage(100.*max2(double(n)/double(maxsteps),
 						(lwstart-log(err_i))/(lwstart-lerr)));
 	  } 
 	
 	const_cast<int&> (steps) = n;
-	
-	delete &r;
-	delete &r_tilde;
-	delete &p;
-	delete &p_tilde;
-	delete &s;
-	delete &s_tilde;
-	delete &t;
-	delete &v;
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in BiCGStabSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in BiCGStabSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in BiCGStabSolver::Mult\n"));
+      }
   }
 
 
@@ -801,11 +777,10 @@ namespace ngla
   try
       {
 	// Solve A u = f
-	if(sh)
-	  sh->SetThreadPercentage(0);
+        BaseStatusHandler::SetThreadPercentage(0);
  
-	BaseVector & d = *f.CreateVector();
-	BaseVector & w = *f.CreateVector();
+	auto d = f.CreateVector();
+	auto w = f.CreateVector();
 
 	int n = 0;
 	double err, err0;
@@ -841,21 +816,18 @@ namespace ngla
           }
 
 	const_cast<int&> (steps) = n;
-	
-	delete &d;
-	delete &w;
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in SimpleIterationSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in SimpleIterationSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in SimpleIterationSolver::Mult\n"));
+      }
   }
 
 
@@ -887,13 +859,13 @@ namespace ngla
       {
 	// Solve A u = f
 
-	BaseVector & v = *f.CreateVector();
-	BaseVector & av = *f.CreateVector();
-	BaseVector & r = *f.CreateVector();
-	BaseVector & w = *f.CreateVector();
-	BaseVector & hv = *f.CreateVector();
+	auto v = f.CreateVector();
+	auto av = f.CreateVector();
+	auto r = f.CreateVector();
+	auto w = f.CreateVector();
+	auto hv = f.CreateVector();
 
-        Array<BaseVector*> vi(maxsteps);
+        Array<AutoVector> vi(maxsteps);
         Matrix<SCAL> h(maxsteps+1, maxsteps);
         Matrix<SCAL> h2(maxsteps+1, maxsteps);
         Vector<SCAL> gammai(maxsteps), ci(maxsteps), si(maxsteps);
@@ -935,8 +907,8 @@ namespace ngla
 	int j = -1;
 	while (j++ < maxsteps-2 && norm > err)
 	  {
-            vi[j] =  f.CreateVector();
-            *vi[j] = v;
+            vi[j].AssignPointer (f.CreateVector());
+            vi[j] = v;
 
             av = (*a) * v;
             if (c)
@@ -978,7 +950,7 @@ namespace ngla
           }
         
         j--;
-
+        cout << IM(5) << "gmres - Triangular matrix" << endl << h.Rows(0,j+2).Cols(0,j+2) << endl;
         Vector<SCAL> y(maxsteps);
         for (int i = j; i >= 0; i--)
           {
@@ -1026,26 +998,18 @@ namespace ngla
         for (int i = 0; i <= j; i++)
           x += us(i) * *vi[i];
         */
-
-
-
-	delete &v;
-        delete &av;
-        delete &w;
-	delete &r;
-	delete &hv;
       }
 
-    catch (exception & e)
-      {
-	throw Exception(e.what() +
-			string ("\ncaught in GMRESSolver::Mult\n"));
-      }
     catch (Exception & e)
       {
 	e.Append ("in caught in GMRESSolver::Mult\n");
 	throw;
       }
+    catch (exception & e)
+      {
+	throw Exception(e.what() +
+			string ("\ncaught in GMRESSolver::Mult\n"));
+      }
   }
 
 
@@ -1098,20 +1062,20 @@ void QMRSolver<SCAL> :: Mult (const BaseVector & b, BaseVector & x) const
       SCAL rho, rho_1, xi, gamma, gamma_1, theta, theta_1, eta, delta, ep=1.0, beta;
       
 
-      BaseVector & r = *b.CreateVector();
-      BaseVector & v_tld = *b.CreateVector();
-      BaseVector & y = *b.CreateVector();
-      BaseVector & w_tld = *b.CreateVector();
-      BaseVector & z = *b.CreateVector();
-      BaseVector & v = *b.CreateVector();
-      BaseVector & w = *b.CreateVector();
-      BaseVector & y_tld = *b.CreateVector();
-      BaseVector & z_tld = *b.CreateVector();
-      BaseVector & p = *b.CreateVector();
-      BaseVector & q = *b.CreateVector();
-      BaseVector & p_tld = *b.CreateVector();
-      BaseVector & d = *b.CreateVector();
-      BaseVector & s = *b.CreateVector();
+      auto r = b.CreateVector();
+      auto v_tld = b.CreateVector();
+      auto y = b.CreateVector();
+      auto w_tld = b.CreateVector();
+      auto z = b.CreateVector();
+      auto v = b.CreateVector();
+      auto w = b.CreateVector();
+      auto y_tld = b.CreateVector();
+      auto z_tld = b.CreateVector();
+      auto p = b.CreateVector();
+      auto q = b.CreateVector();
+      auto p_tld = b.CreateVector();
+      auto d = b.CreateVector();
+      auto s = b.CreateVector();
 
       double normb = b.L2Norm();
 
@@ -1324,20 +1288,20 @@ void QMRSolver<SCAL> :: Mult (const BaseVector & b, BaseVector & x) const
     }
 
   
-  catch (exception & e)
-    {
-      throw Exception(e.what() +
-		      string ("\ncaught in QMRSolver::Mult\n"));
-    }
 
   catch (Exception & e)
     {
       e.Append ("in caught in QMRSolver::Mult\n"); 
       throw;
     }
+  catch (exception & e)
+    {
+      throw Exception(e.what() +
+		      string ("\ncaught in QMRSolver::Mult\n"));
+    }
 }
   
-
+ 
   
   template class CGSolver<double>;
   template class CGSolver<Complex>;
