@@ -26,7 +26,7 @@ namespace ngla
     cout << "ne = " << ne << ", nv = " << nv << endl;
 
     jacobi = 0;
-    coarsemat = 0;
+    coarsemat = nullptr;
     inv = 0;
 
     if (nv < 20 || levels == 0)
@@ -224,9 +224,9 @@ namespace ngla
     delete prol;
     delete recAMG;
 
-    delete jacobi;
-    delete coarsemat;
-    delete inv;
+    // delete jacobi;
+    // delete coarsemat;
+    // delete inv;
   }
   
 
@@ -244,25 +244,25 @@ namespace ngla
 	  for (int j = 0; j < prol->GetRowIndices(i).Size(); j++)
 	    cnt[prol->GetRowIndices(i)[j]]++;
 	
-	Table<int> * smblocks = new Table<int>(cnt);
+	Table<int> smblocks(cnt);
 	cnt = 0;
 	for (int i = 0; i < prol->Height(); i++)
 	  for (int j = 0; j < prol->GetRowIndices(i).Size(); j++)
 	    {
 	      int jj = prol->GetRowIndices(i)[j];
-	      (*smblocks)[jj][cnt[jj]] = i;
+	      smblocks[jj][cnt[jj]] = i;
 	      cnt[jj]++;
 	    }
 	// *testout << "prol = " << endl << *prol << endl;
 	// *testout << "smoothing blocks = " << endl << *smblocks << endl;
 	
-	bjacobi = mat.CreateBlockJacobiPrecond (*smblocks);
+	bjacobi = mat.CreateBlockJacobiPrecond (make_shared<Table<int>> (smblocks));
       }
     jacobi = mat.CreateJacobiPrecond ();
 
     if (recAMG)
       {
-	coarsemat = mat.Restrict (*prol);
+	coarsemat = shared_ptr<BaseSparseMatrix>(mat.Restrict (*prol));
 	recAMG -> ComputeMatrices (*coarsemat);
 	inv = 0;
       }
@@ -274,9 +274,9 @@ namespace ngla
       }
   }
 
-  int AMG_H1 :: NZE() const
+  size_t AMG_H1 :: NZE() const
   {
-    int nze = pmat->NZE();
+    size_t nze = pmat->NZE();
     if (recAMG)
       nze += recAMG->NZE();
     return nze;
@@ -294,9 +294,9 @@ namespace ngla
 	return;
       }
 
-    BaseVector & hv = *pmat->CreateVector();
-    BaseVector & wc = *coarsemat->CreateVector();
-    BaseVector & dc = *coarsemat->CreateVector();
+    auto hv = pmat->CreateColVector();
+    auto wc = coarsemat->CreateColVector();
+    auto dc = coarsemat->CreateColVector();
 
     y = 0;
     jacobi->GSSmooth (y, x);
@@ -313,10 +313,6 @@ namespace ngla
       }
 
     jacobi->GSSmoothBack (y, x);
-
-    delete &hv;
-    delete &wc;
-    delete &dc;
   }
       
 
@@ -374,7 +370,7 @@ namespace ngla
     if (vertices.Size())   // visualization of faces
       {
         char buf[20];
-        sprintf (buf, "amg.level%d.surf", levels);
+        snprintf (buf, 20, "amg.level%d.surf", levels);
         cout << "filename = " << buf << endl;
         ofstream out (buf);
         out << "surfacemesh" << endl;
@@ -385,10 +381,11 @@ namespace ngla
               << vertices[i](2) << "\n";
 
         double minweight=1e10;
+        /*
         int cntf = 0;
         for (int i = 0; i < f2v.Size(); i++)
           if (weightf[i] < minweight) cntf++;
-
+        */
         out << f2v.Size() << endl;
         for (int i = 0; i < f2v.Size(); i++)
 	  if (weightf[i] < minweight) 
@@ -1104,12 +1101,12 @@ namespace ngla
     cout << "compute HCurl matrices" << endl;
 
     pmat = &mat;
-    coarsemat = mat.Restrict (*prol);
+    coarsemat = shared_ptr<BaseSparseMatrix>(mat.Restrict (*prol));
     jacobi = mat.CreateJacobiPrecond ();
 
     h1mat = mat.Restrict (*grad);
     //    dynamic_cast<SparseMatrixSymmetric<Mat<1,1> >&> (*h1mat) (0,0) += 1;
-    dynamic_cast<SparseMatrixSymmetricTM<double>&> (*h1mat)(0,0) += 1;
+    dynamic_cast<SparseMatrixTM<double>&> (*h1mat)(0,0) += 1;
 
     if (recAMG)
       {
@@ -1132,9 +1129,9 @@ namespace ngla
 
 
 
-  int AMG_HCurl :: NZE() const
+  size_t AMG_HCurl :: NZE() const
   {
-    int nze = pmat->NZE() + h1mat->NZE();
+    size_t nze = pmat->NZE() + h1mat->NZE();
     if (recAMG) 
       nze += recAMG->NZE() + h1AMG->NZE();
 
@@ -1151,13 +1148,13 @@ namespace ngla
 	return;
       }
 
-    BaseVector & hv = *pmat->CreateVector();
-    BaseVector & res = *pmat->CreateVector();
-    BaseVector & wc = *coarsemat->CreateVector();
-    BaseVector & dc = *coarsemat->CreateVector();
+    auto hv = pmat->CreateColVector();
+    auto res = pmat->CreateColVector();
+    auto wc = coarsemat->CreateColVector();
+    auto dc = coarsemat->CreateColVector();
 
-    BaseVector & wh1 = *h1mat->CreateVector();
-    BaseVector & dh1 = *h1mat->CreateVector();
+    auto wh1 = h1mat->CreateColVector();
+    auto dh1 = h1mat->CreateColVector();
     
     y = 0;
     res = x;
@@ -1192,14 +1189,6 @@ namespace ngla
 
     // for (int k = 0; k < 5; k++)
       jacobi->GSSmoothBack (y, x);
-
-
-    delete &dh1;
-    delete &wh1;
-    delete &dc;
-    delete &wc;
-    delete &res;
-    delete &hv;
   }
 }
 
