@@ -8,6 +8,8 @@
 #ifdef USE_MUMPS
 
 #include <la.hpp>
+#include "mumpsinverse.hpp"
+
 #include <comp.hpp>
 // #include <parallelngs.hpp>
 
@@ -28,23 +30,33 @@ namespace ngla
   template <class TM, class TV_ROW, class TV_COL>
   MumpsInverse<TM,TV_ROW,TV_COL> :: 
   MumpsInverse (const SparseMatrix<TM,TV_ROW,TV_COL> & a, 
-                const BitArray * ainner,
-                const Array<int> * acluster,
+                shared_ptr<BitArray> ainner,
+                shared_ptr<const Array<int>> acluster,
                 bool asymmetric)
+    : comm(MPI_COMM_NULL, false)
   { 
     static Timer timer ("Mumps Inverse");
     static Timer timer_analysis ("Mumps Inverse - analysis");
     static Timer timer_factor ("Mumps Inverse - factor");
     RegionTimer reg (timer);
 
-    VT_OFF();
 
     symmetric = asymmetric;
     inner = ainner;
     cluster = acluster;
 
-    int ntasks = MyMPI_GetNTasks();
-    int id = MyMPI_GetId();
+    auto pds = a.GetParallelDofs();
+    if ( (pds != nullptr) && // if we are on an "only-me comm", take it
+	 (pds->GetCommunicator().Size() == 1) )
+      { comm = pds->GetCommunicator(); }
+    else { // otherwise make an "only-me comm" from world-communicator
+      NgMPI_Comm wcomm(MPI_COMM_WORLD);
+      Array<int> procs = { wcomm.Rank() };
+      comm = wcomm.SubCommunicator(procs);
+    }
+
+    int ntasks = comm.Size();
+    int id = comm.Rank();
     
     if (id == 0)
       {
@@ -55,7 +67,7 @@ namespace ngla
 	    throw Exception("Invalid parameters inner/cluster. Thrown by MumpsInverse.");
 	  }
 	
-	if ( int( mat_traits<TM>::WIDTH) != int(mat_traits<TM>::HEIGHT) )
+	if ( int( ngbla::Width<TM>()) != int(ngbla::Height<TM>()) )
 	  {
 	    cout << "Mumps: Each entry in the square matrix has to be a square matrix!" << endl;
 	    throw Exception("No Square Matrix. Thrown by MumpsInverse.");
@@ -63,8 +75,8 @@ namespace ngla
       }
 
 
-    entrysize = mat_traits<TM>::HEIGHT; 
-    iscomplex = mat_traits<TM>::IS_COMPLEX;
+    entrysize = ngbla::Height<TM>();
+    iscomplex = ngbla::IsComplex<TM>();
 
 
     int * colstart = 0;
@@ -238,7 +250,7 @@ namespace ngla
     mumps_id.par = (ntasks == 1) ? 1 : 0;
     mumps_id.sym = symmetric ? 1 : 0;
     // mumps_id.comm_fortran=USE_COMM_WORLD;
-    mumps_id.comm_fortran = MPI_Comm_c2f (ngs_comm);
+    mumps_id.comm_fortran = MPI_Comm_c2f (comm);
     mumps_trait<TSCAL>::MumpsFunction (&mumps_id);
 
     // cout << IM(1) << "MUMPS version number is " << mumps_id.version_number << endl;
@@ -270,7 +282,7 @@ namespace ngla
     mumps_id.icntl[28]=0;  // 0..auto, 1..ptscotch 2..parmetis
 
     // mumps_id.comm_fortran=USE_COMM_WORLD;
-    mumps_id.comm_fortran = MPI_Comm_c2f (ngs_comm);
+    mumps_id.comm_fortran = MPI_Comm_c2f (comm);
     mumps_id.job = JOB_ANALYSIS;
 
 
@@ -297,7 +309,7 @@ namespace ngla
     if (id == 0)
       cout << "factor ... " << flush;
 
-    MPI_Barrier (ngs_comm);
+    MPI_Barrier (comm);
 
     timer_factor.Start();
     mumps_trait<TSCAL>::MumpsFunction (&mumps_id);
@@ -333,7 +345,6 @@ namespace ngla
     delete [] col_indices;
     delete [] row_indices;
     delete [] matrix;
-    VT_ON();
   }
   
   
@@ -342,12 +353,11 @@ namespace ngla
   void MumpsInverse<TM,TV_ROW,TV_COL> :: 
   Mult (const BaseVector & x, BaseVector & y) const
   {
-    int id = MyMPI_GetId();
+    int id = comm.Rank();
 
     static int timer = NgProfiler::CreateTimer ("Mumps mult inverse");
     NgProfiler::RegionTimer reg (timer);
 
-    VT_OFF();
 
     if (id == 0)
       {
@@ -384,7 +394,6 @@ namespace ngla
 	ncid.job = JOB_SOLVE;
 	mumps_trait<TSCAL>::MumpsFunction (&ncid);
       }
-    VT_ON();
   }
   
 
@@ -418,10 +427,11 @@ namespace ngla
   template <class TM, class TV>
   ParallelMumpsInverse<TM,TV> :: 
   ParallelMumpsInverse (const BaseSparseMatrix & ba, 
-			const BitArray * inner,
-			const Array<int> * cluster,
-			const ParallelDofs * pardofs,
+			shared_ptr<BitArray> inner,
+			shared_ptr<const Array<int>> cluster,
+			shared_ptr<ParallelDofs> pardofs,
 			bool asymmetric)
+    : BaseMatrix(pardofs)
   { 
     static Timer timer ("Mumps Inverse");
     static Timer timer_analysis ("Mumps Inverse - analysis");
@@ -438,8 +448,9 @@ namespace ngla
 
     cout << IM(1) << "Mumps Parallel inverse, symmetric = " << symmetric << endl;
 
-    int ntasks = MyMPI_GetNTasks();
-    int id = MyMPI_GetId();
+    NgMPI_Comm comm = pardofs->GetCommunicator();
+    int ntasks = comm.Size();
+    int id = comm.Rank();
 
     if (id == 0)
       {
@@ -450,7 +461,7 @@ namespace ngla
 	    throw Exception("Invalid parameters inner/cluster. Thrown by ParallelMumpsInverse.");
 	  }
 	
-	if ( int( mat_traits<TM>::WIDTH) != int(mat_traits<TM>::HEIGHT) )
+	if ( int(ngbla::Width<TM>()) != int(ngbla::Height<TM>()) )
 	  {
 	    cout << "Mumps: Each entry in the square matrix has to be a square matrix!" << endl;
 	    throw Exception("No Square Matrix. Thrown by ParallelMumpsInverse.");
@@ -468,8 +479,14 @@ namespace ngla
 
    // consistent enumeration (new version)
     
+    Array<int> global_nums;
+    int num_glob_dofs;
+    pardofs -> EnumerateGlobally (inner, global_nums, num_glob_dofs);
+
     int ndof = pardofs->GetNDofLocal();
-    
+
+
+    /*
     Array<int> global_nums(ndof);
     global_nums = -1;
     int num_master_dofs = 0;
@@ -495,7 +512,8 @@ namespace ngla
 	global_nums[i] += first_master_dof[id];
     
     ScatterDofData (global_nums, *pardofs);
-
+    */
+    
 
     // copy to old variables ...
     num_globdofs = num_glob_dofs;
@@ -505,6 +523,8 @@ namespace ngla
 	select.Append (row);
     
     
+    /*
+      // test dofs, only for ParallelMeshDofs
     const ParallelMeshDofs & pmdofs = dynamic_cast<const ParallelMeshDofs&> (*pardofs);
     for (int row = 0; row < ndof; row++)
       if (!inner || inner->Test(row))
@@ -518,10 +538,11 @@ namespace ngla
 	    pmdofs.GetMeshAccess().GetDistantProcs (node, procs);
 	    cout << "procs = " << procs << endl;
 	  }
+    */
 
 
-    entrysize = mat_traits<TM>::HEIGHT; 
-    iscomplex = mat_traits<TM>::IS_COMPLEX;
+    entrysize = ngbla::Height<TM>();
+    iscomplex = ngbla::IsComplex<TM>();
 
     Array<int> col_indices;
     Array<int> row_indices;
@@ -717,26 +738,29 @@ namespace ngla
 	  }
       }
     else
-      nze = 0;
+      {
+	height = 0;
+	nze = 0;
+      }
 
     /*
     for (int i = 0; i < 40; i++)
       mumps_id.icntl[i] = 0;
     */
 
-    VT_OFF();
 
     /*
     *testout << "mumps matrix: n = " << num_globdofs << ", nz = " << nze << endl;
+    *testout << "loc2glob = " << loc2glob << endl;
     for (int i = 0; i < nze; i++)
       *testout << "a(" << row_indices[i] << "," << col_indices[i] << ") = " << matrix[i] << endl;
       */
     
     mumps_id.job = JOB_INIT; 
     mumps_id.par = (ntasks == 1) ? 1 : 0;
-    mumps_id.sym = symmetric ? 1 : 0;
+    mumps_id.sym = symmetric ? 2 : 0;    // 1 .. spd, 2 .. general symmetric
     //mumps_id.comm_fortran=USE_COMM_WORLD;
-    mumps_id.comm_fortran = MPI_Comm_c2f (ngs_comm);
+    mumps_id.comm_fortran = MPI_Comm_c2f (comm);
 
 
     MumpsFunction (mumps_id);
@@ -750,17 +774,21 @@ namespace ngla
     mumps_id.irn_loc = row_indices.Addr(0);
     mumps_id.jcn_loc = col_indices.Addr(0);
 
+
     mumps_id.icntl[0]=-1; 
     mumps_id.icntl[1]=-1; 
     mumps_id.icntl[2]=-1; 
     mumps_id.icntl[3]=1;
 
-    /*
-    mumps_id.icntl[0]= 6; 
-    mumps_id.icntl[1]= 0; 
-    mumps_id.icntl[2]= 6; 
-    mumps_id.icntl[3]= 1;
-    */
+    if (getenv ("MUMPSMSG"))
+      {
+	int level = atoi (getenv("MUMPSMSG"));
+	mumps_id.icntl[0]= 6; 
+	mumps_id.icntl[1]= 6; 
+	mumps_id.icntl[2]= 6; 
+	mumps_id.icntl[3]= level;
+      }
+
 
     /*
     // mumps_id.icntl[7]=7;   // BUG (??) 0..min deg, 3..scotch 5..metis, 7..default
@@ -773,7 +801,7 @@ namespace ngla
     */
 
     mumps_id.icntl[6]=7;   // 0..min deg, 3..scotch 5..metis, 7..default
-    mumps_id.icntl[12]=1;  // 0..do use, 1..not using scalapck for root schur complement
+    mumps_id.icntl[12]=0;  // 0..do use, 1..not using scalapck for root schur complement
     mumps_id.icntl[13]=50; // memory increase (in %) due to error -9
     mumps_id.icntl[17]=3;  // parallel input
     mumps_id.icntl[27]=0;  // 0..default, 1..seq, 2..parallel analysis
@@ -781,7 +809,7 @@ namespace ngla
 
 
     // mumps_id.comm_fortran=USE_COMM_WORLD;
-    mumps_id.comm_fortran = MPI_Comm_c2f (ngs_comm);
+    mumps_id.comm_fortran = MPI_Comm_c2f (comm);
     mumps_id.job = JOB_ANALYSIS;
 
     cout << IM(1) << "analysis ... " << flush;
@@ -821,7 +849,6 @@ namespace ngla
 
     cout << IM(1) << " done " << endl;
     
-    VT_ON();
   }
   
   
@@ -836,10 +863,10 @@ namespace ngla
     x.Distribute();
     y.SetParallelStatus (CUMULATED);
 
-    VT_OFF();
 
-    int ntasks = MyMPI_GetNTasks();
-    int id = MyMPI_GetId();
+    NgMPI_Comm comm = paralleldofs->GetCommunicator();
+    int ntasks = comm.Size();
+    int id = comm.Rank();
 
 
     if (id != 0)
@@ -854,15 +881,15 @@ namespace ngla
 	for (int i = 0; i < select.Size(); i++)
 	  select_loc2glob[i] = loc2glob[select[i]];
 	
-	MyMPI_Send (select_loc2glob, 0);
-	MyMPI_Send (hx, 0);
+	comm.Send (select_loc2glob, 0, MPI_TAG_SOLVE);
+	comm.Send (hx, 0, MPI_TAG_SOLVE);
 
 	MUMPS_STRUC_C & ncid = const_cast<MUMPS_STRUC_C&> (mumps_id);
 	ncid.job = JOB_SOLVE;
 	mumps_trait<TSCAL>::MumpsFunction (&ncid);
 
-	MyMPI_Send (select_loc2glob, 0);
-	MyMPI_Recv (hx, 0);
+	comm.Send (select_loc2glob, 0, MPI_TAG_SOLVE);
+	comm.Recv (hx, 0, MPI_TAG_SOLVE);
 
 	y = 0;
 	for (int i = 0; i < select.Size(); i++)
@@ -877,8 +904,8 @@ namespace ngla
 	  {
 	    Array<int> loc2glob;
 	    Array<TV> hx;
-	    MyMPI_Recv (loc2glob, src);
-	    MyMPI_Recv (hx, src);
+	    comm.Recv (loc2glob, src, MPI_TAG_SOLVE);
+	    comm.Recv (hx, src, MPI_TAG_SOLVE);
 	    for (int j = 0; j < loc2glob.Size(); j++)
 	      rhs(loc2glob[j]) += hx[j];
 	  } 
@@ -894,16 +921,15 @@ namespace ngla
 	for (int src = 1; src < ntasks; src++)
 	  {
 	    Array<int> loc2glob;
-	    MyMPI_Recv (loc2glob, src);
+	    comm.Recv (loc2glob, src, MPI_TAG_SOLVE);
 	    Array<TV> hx(loc2glob.Size());
 
 	    for (int j = 0; j < loc2glob.Size(); j++)
 	      hx[j] = rhs(loc2glob[j]);
-	    MyMPI_Send (hx, src);
+	    comm.Send (hx, src, MPI_TAG_SOLVE);
 	  }
       }
 
-    VT_ON();
   }
   
 
@@ -915,8 +941,23 @@ namespace ngla
   }
 
 
+  template <class TM, class TV>
+  AutoVector ParallelMumpsInverse<TM,TV> :: CreateVector () const
+  {
+    return make_unique<ParallelVVector<TV>> (paralleldofs->GetNDofLocal(), paralleldofs);
+  }
 
+  template <class TM, class TV>
+  AutoVector ParallelMumpsInverse<TM,TV> :: CreateRowVector () const
+  {
+    return make_unique<ParallelVVector<TV>> (paralleldofs->GetNDofLocal(), paralleldofs);
+  }
 
+  template <class TM, class TV>
+  AutoVector ParallelMumpsInverse<TM,TV> :: CreateColVector () const
+  {
+    return make_unique<ParallelVVector<TV>> (paralleldofs->GetNDofLocal(), paralleldofs);
+  }
 
 
 
@@ -958,6 +999,26 @@ namespace ngla
   template class MumpsInverse<Mat<3,3,double> >;
   template class MumpsInverse<Mat<3,3,Complex> >;
 #endif
+#if MAX_SYS_DIM >= 4
+  template class MumpsInverse<Mat<4,4,double> >;
+  template class MumpsInverse<Mat<4,4,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 5
+  template class MumpsInverse<Mat<5,5,double> >;
+  template class MumpsInverse<Mat<5,5,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 6
+  template class MumpsInverse<Mat<6,6,double> >;
+  template class MumpsInverse<Mat<6,6,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 7
+  template class MumpsInverse<Mat<7,7,double> >;
+  template class MumpsInverse<Mat<7,7,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 8
+  template class MumpsInverse<Mat<8,8,double> >;
+  template class MumpsInverse<Mat<8,8,Complex> >;
+#endif
 
 
 
@@ -983,27 +1044,25 @@ namespace ngla
   template class ParallelMumpsInverse<Mat<3,3,Complex> >;
 #endif
 #if MAX_SYS_DIM >= 4
-  template class MumpsInverse<Mat<4,4,double> >;
-  template class MumpsInverse<Mat<4,4,Complex> >;
+  template class ParallelMumpsInverse<Mat<4,4,double> >;
+  template class ParallelMumpsInverse<Mat<4,4,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 5
+  template class ParallelMumpsInverse<Mat<5,5,double> >;
+  template class ParallelMumpsInverse<Mat<5,5,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 6
+  template class ParallelMumpsInverse<Mat<6,6,double> >;
+  template class ParallelMumpsInverse<Mat<6,6,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 7
+  template class ParallelMumpsInverse<Mat<7,7,double> >;
+  template class ParallelMumpsInverse<Mat<7,7,Complex> >;
+#endif
+#if MAX_SYS_DIM >= 8
+  template class ParallelMumpsInverse<Mat<8,8,double> >;
+  template class ParallelMumpsInverse<Mat<8,8,Complex> >;
 #endif
-  /*
-    #if MAX_SYS_DIM >= 5
-    template class MumpsInverse<Mat<5,5,double> >;
-    template class MumpsInverse<Mat<5,5,Complex> >;
-    #endif
-    #if MAX_SYS_DIM >= 6
-    template class MumpsInverse<Mat<6,6,double> >;
-    template class MumpsInverse<Mat<6,6,Complex> >;
-    #endif
-    #if MAX_SYS_DIM >= 7
-    template class MumpsInverse<Mat<7,7,double> >;
-    template class MumpsInverse<Mat<7,7,Complex> >;
-    #endif
-    #if MAX_SYS_DIM >= 8
-    template class MumpsInverse<Mat<8,8,double> >;
-    template class MumpsInverse<Mat<8,8,Complex> >;
-    #endif
-  */
 
   /*
     template class MumpsInverse<double, Vec<2,double>, Vec<2,double> >;
@@ -1063,5 +1122,3 @@ namespace ngla
 
 #endif
 
-
-
