@@ -5,48 +5,104 @@
 /* *************************************************************************/
 
 #include <la.hpp>
+#include "pardisoinverse.hpp"
 
-#ifdef USE_PARDISO
+#ifdef USE_MKL
+#include <mkl_service.h>
+#endif // USE_MKL
+
+// #include "/opt/intel/mkl/include/mkl_service.h"
 
 #define F77_FUNC(func)  func ## _
 
 using ngbla::integer;
 
-extern "C"
+extern "C" 
 {
   /* PARDISO prototype. */
 
-#ifdef USE_MKL
+#ifdef USE_PARDISO
 
+#ifdef USE_MKL
+#define MKL_PARDISO
+
+  void mkl_free_buffers (void);
   void F77_FUNC(pardiso)
     (void * pt, integer * maxfct, integer * mnum, integer * mtype, integer * phase, integer * n, 
      double * a, integer * ia, integer * ja, integer * perm, integer * nrhs, integer * iparam, 
      integer * msglvl, double * b, double * x, integer * error);
 
-#else
+#else // USE_MKL
 
 #ifdef USE_PARDISO400
 extern  integer F77_FUNC(pardisoinit)
     (void *, integer *, integer *, integer *, double *, integer *);
-#else
+#else // USE_PARDISO400
 extern  integer F77_FUNC(pardisoinit)
     (void *, integer *, integer *);
-#endif
+#endif // USE_PARDISO400
   integer F77_FUNC(pardiso)
     (void * pt, integer * maxfct, integer * mnum, integer * mtype, integer * phase, integer * n, 
      double * a, integer * ia, integer * ja, integer * perm, integer * nrhs, integer * iparam, 
      integer * msglvl, double * b, double * x, integer * error);
 
 #endif
+
+#else // USE_PARDISO
+  // Neither MKL nor PARDISO linked at compile-time
+  // check for MKL at run-time and set function pointers if available
+#define MKL_PARDISO
+
+  void (*mkl_free_buffers) (void) = nullptr;
+  void (*F77_FUNC(pardiso))
+    (void * pt, integer * maxfct, integer * mnum, integer * mtype, integer * phase, integer * n,
+     double * a, integer * ia, integer * ja, integer * perm, integer * nrhs, integer * iparam,
+     integer * msglvl, double * b, double * x, integer * error) = nullptr;
+
+#endif // USE_PARDISO
 }
 
 
-
+#ifdef USE_MKL
+namespace ngstd
+{
+    extern int mkl_max_threads;
+}
+#endif // USE_MKL
 
 
 
 namespace ngla
 {
+#ifdef USE_PARDISO
+  bool is_pardiso_available = true;
+#else
+#ifndef __EMSCRIPTEN__
+  static SharedLibrary libmkl;
+  static bool LoadMKL()
+  {
+      try
+      {
+#ifdef WIN32
+          libmkl.Load("mkl_rt.dll");
+#else
+          libmkl.Load("libmkl_rt.so");
+#endif
+          mkl_free_buffers = libmkl.GetFunction<decltype(mkl_free_buffers)>("mkl_free_buffers");
+          F77_FUNC(pardiso) = libmkl.GetFunction<decltype(pardiso_)>("pardiso_");
+          return mkl_free_buffers && F77_FUNC(pardiso);
+      }
+      catch(const std::runtime_error &)
+      {
+          return false;
+      }
+  };
+  bool is_pardiso_available = LoadMKL();
+#else // __EMSCRIPTEN__
+  bool is_pardiso_available = false;
+#endif // __EMSCRIPTEN__
+#endif
+
   int pardiso_msg = 0;
 
 
@@ -78,19 +134,32 @@ namespace ngla
 
   template<class TM>
   PardisoInverseTM<TM> :: 
-  PardisoInverseTM (const SparseMatrixTM<TM> & a, 
-		    const BitArray * ainner,
-		    const Array<int> * acluster,
+  PardisoInverseTM (shared_ptr<const SparseMatrixTM<TM>> a,
+		    shared_ptr<BitArray> ainner,
+		    shared_ptr<const Array<int>> acluster,
 		    int asymmetric)
     : SparseFactorization (a, ainner, acluster)
   { 
     static Timer timer("Pardiso Inverse");
     RegionTimer reg (timer);
+    GetMemoryTracer().SetName("PardisoInverseTM<" + Demangle(typeid(TM).name()) + ">");
+    GetMemoryTracer().Track(rowstart, "rowstart",
+                            indices, "indices",
+                            matrix, "matrix");
 
 
     if (getenv ("PARDISOMSG"))
       pardiso_msg = 1;
 
+    /*
+    cout << "set mkl_rt interface" << endl;
+#ifdef MKL_ILP64
+    MKL_Set_Interface_Layer(MKL_INTERFACE_ILP64);
+#else
+    MKL_Set_Interface_Layer(MKL_INTERFACE_LP64);
+#endif
+    MKL_Set_Threading_Layer (MKL_THREADING_GNU);
+    */
 
     print = bool (pardiso_msg); // false;
 
@@ -102,34 +171,36 @@ namespace ngla
     if (inner && cluster)
       throw Exception("PardisoInverse: Cannot use inner and cluster");
 
-    if ( (inner && inner->Size() < a.Height()) ||
-	 (cluster && cluster->Size() < a.Height() ) )
+    if ( (inner && inner->Size() < a->Height()) ||
+	 (cluster && cluster->Size() < a->Height() ) )
       {
 	cout << "PardisoInverse: Size of inner/cluster does not match matrix size!" << endl;
 	throw Exception("Invalid parameters inner/cluster. Thrown by PardisoInverse.");
       }
 
-    if ( int( mat_traits<TM>::WIDTH) != int(mat_traits<TM>::HEIGHT) )
+    // if ( int( mat_traits<TM>::WIDTH) != int(mat_traits<TM>::HEIGHT) )
+    if ( ngbla::Height<TM>() != ngbla::Width<TM>() )
       {
 	cout << "PardisoInverse: Each entry in the square matrix has to be a square matrix!" << endl;
 	throw Exception("No Square Matrix. Thrown by PardisoInverse.");
       }
 
 
-    entrysize = mat_traits<TM>::HEIGHT; 
-    height = a.Height() * entrysize;
+    entrysize = ngbla::Height<TM>(); // mat_traits<TM>::HEIGHT; 
+    height = a->Height() * entrysize;
 
-    *testout << "matrix.InverseTpye = " <<  a.GetInverseType() << endl;
-    spd = ( a.GetInverseType() == PARDISOSPD ) ? 1 : 0;
+    *testout << "matrix.InverseTpye = " <<  a->GetInverseType() << endl;
+    spd = ( a->GetInverseType() == PARDISOSPD ) ? 1 : 0;
 
-    integer maxfct = 1, mnum = 1, phase = 12, nrhs = 1, msglevel = print, error;
+    integer maxfct = 1, mnum = 1, phase = 12, nrhs = 1, msglevel = print, error = 0;
     integer * params = const_cast <integer*> (&hparams[0]);
 
     for (int i = 0; i < 64; i++)
       params[i] = 0;
 
     params[0] = 1; // no pardiso defaults
-    params[2] = 8; // 1 processor (?)
+    // params[2] = 8; // 1 processor (?)
+    params[2] = TaskManager::GetMaxThreads(); 
 
     params[1] = 0; // fill in 0..MDO, 2..metis
     params[3] = params[4] = params[5] = params[7] = params[8] = 
@@ -148,15 +219,16 @@ namespace ngla
 
     for (int i = 0; i < 128; i++) pt[i] = 0;
 
-#ifdef USE_MKL
+#ifdef MKL_PARDISO
     //    no init in MKL PARDISO
 #else
 
-    int retvalue;
+    integer retvalue;
 #ifdef USE_PARDISO400
     double dparm[64]; 
     integer solver = 0;
     F77_FUNC(pardisoinit) (pt,  &matrixtype, &solver, params, dparm, &retvalue); 
+    cout << "pardisoinit, retval = " << retvalue << endl;
 #else
     retvalue = F77_FUNC(pardisoinit) (pt,  &matrixtype, params); 
 #endif
@@ -167,11 +239,11 @@ namespace ngla
     SetMatrixType();
 
     if (inner)
-      GetPardisoMatrix (a, SubsetFree (*inner));
+      GetPardisoMatrix (*a, SubsetFree (*inner));
     else if (cluster)
-      GetPardisoMatrix (a, SubsetCluster (*cluster));
+      GetPardisoMatrix (*a, SubsetCluster (*cluster));
     else
-      GetPardisoMatrix (a, SubsetAll());
+      GetPardisoMatrix (*a, SubsetAll());
 
     nze = rowstart[compressed_height];
 
@@ -180,17 +252,29 @@ namespace ngla
     // time1 = clock();
     cout << IM(3) << "call pardiso ..." << flush;
 
-    // retvalue = 
-    F77_FUNC(pardiso) ( pt, &maxfct, &mnum, &matrixtype, &phase, &compressed_height, 
-			reinterpret_cast<double *>(&matrix[0]),
-			&rowstart[0], &indices[0], NULL, &nrhs, params, &msglevel,
+    if (task_manager) task_manager -> StopWorkers();
+
+#ifdef USE_MKL
+    mkl_set_num_threads(mkl_max_threads);
+#endif // USE_MKL
+
+    // retvalue =
+    if (matrix.Size() > 0)
+      F77_FUNC(pardiso) ( pt, &maxfct, &mnum, &matrixtype, &phase, &compressed_height, 
+			reinterpret_cast<double *>(matrix.Data()),
+			rowstart.Data(), indices.Data(), NULL, &nrhs, params, &msglevel,
 			NULL, NULL, &error );
+#ifdef USE_MKL
+    mkl_set_num_threads(1);
+#endif // USE_MKL
     
+    if (task_manager) task_manager -> StartWorkers();
+
     cout << IM(3) << " done" << endl;
 
     if ( error != 0 )
       {
-	cout << "Setup and Factorization: PARDISO returned error " << error << "!" << endl;
+	cout << IM(1) << "Setup and Factorization: PARDISO returned error " << error << "!" << endl;
 	
 	string errmsg;
 	switch (error)
@@ -214,30 +298,35 @@ namespace ngla
 	    }
 	  default: ;
 	  }
-
+	
 	cout << "symmetric = " << symmetric << endl;
 	cout << "spd = " << spd << endl;
 	cout << "compressed = " << compressed << endl;
 	cout << "inner = " << inner << endl;
 	cout << "cluster = " << cluster << endl;
-
-	ofstream err("pardiso.err");
-	err << "ngsolve-matrix = " << endl << a << endl;
-	err << "pardiso matrix = " << endl;
-	for (int i = 0; i < compressed_height; i++)
-	  {
-	    err << "Row " << i << " start " << rowstart[i] << ": ";
-	    if ( inner ) err << " free=" << inner->Test(i) << " ";
-	    if ( cluster ) err << " cluster=" << (*cluster)[i] << " ";
-	    for (int j = rowstart[i]; j < rowstart[i+1]; j++)
-	      err << "c=" << indices[j-1]-1 << ", v=" << matrix[j-1] << "   ";
-	    err << "\n";
-	  }
 	
-	cout << "wrote matrix to file 'pardiso.err', please check" << endl;
+	if (compressed_height < 1000)
+	  {
+	    ofstream err("pardiso.err");
+	    err << "ngsolve-matrix = " << endl << a << endl;
+	    err << "pardiso matrix = " << endl;
+	    for (int i = 0; i < compressed_height; i++)
+	      {
+		err << "Row " << i << " start " << rowstart[i] << ": ";
+		if ( inner ) err << " free=" << inner->Test(i) << " ";
+		if ( cluster ) err << " cluster=" << (*cluster)[i] << " ";
+		for (int j = rowstart[i]; j < rowstart[i+1]; j++)
+		  err << "c=" << indices[j-1]-1 << ", v=" << matrix[j-1] << "   ";
+		err << "\n";
+	      }
+	    cout << "wrote matrix to file 'pardiso.err', please check" << endl;
+	  }
 	throw Exception("PardisoInverse: Setup and Factorization failed.");
       }
 
+    memory_allocated_in_pardiso_lib = 1024*params[15];
+    GetMemoryTracer().Alloc(memory_allocated_in_pardiso_lib);
+
     /*
     (*testout) << endl << "Direct Solver: PARDISO by Schenk/Gaertner." << endl;
     (*testout) << "Matrix prepared for PARDISO in " <<
@@ -279,7 +368,7 @@ namespace ngla
 	  {
 	    if (!subset.Used(i)) continue;
 
-	    FlatArray<int> ind = a.GetRowIndices(i);
+	    FlatArray ind = a.GetRowIndices(i);
 	    for (int j = 0; j < ind.Size(); j++ )
 	      {
 		if (!subset.Used(i,ind[j])) continue;
@@ -311,7 +400,7 @@ namespace ngla
 	    if (!subset.Used(i)) continue;
 	    int ci = entrysize*icompress[i];
 
-	    FlatArray<int> ind = a.GetRowIndices(i);
+	    FlatArray ind = a.GetRowIndices(i);
 	    FlatVector<TM> values = a.GetRowValues(i);
 
 	    for (int j = 0; j < ind.Size(); j++)
@@ -339,7 +428,7 @@ namespace ngla
 	for (int i = 0; i < a.Height(); i++ )
 	  {
 	    if (!subset.Used(i)) continue;
-	    FlatArray<int> ind = a.GetRowIndices(i);
+	    FlatArray ind = a.GetRowIndices(i);
 
 	    int rowelems = 0;
 	    for (int j = 0; j < ind.Size(); j++)
@@ -358,7 +447,7 @@ namespace ngla
 	for (int i = 0; i < a.Height(); i++ )
 	  {
 	    if (!subset.Used(i)) continue;
-	    FlatArray<int> ind = a.GetRowIndices(i);
+	    FlatArray ind = a.GetRowIndices(i);
 	    FlatVector<TM> values = a.GetRowValues(i);
 
 	    int ci = entrysize*icompress[i];
@@ -387,13 +476,13 @@ namespace ngla
   void PardisoInverse<TM,TV_ROW,TV_COL> :: 
   Mult (const BaseVector & x, BaseVector & y) const
   {
-    static Timer timer("Pardiso Solve");
+    static Timer timer(string("Pardiso Solve, mat = ") + typeid(TM).name() + ", vec = " + typeid(TV_ROW).name());
     RegionTimer reg (timer);
 
     FlatVector<TVX> fx = x.FV<TVX> ();
     FlatVector<TVX> fy = y.FV<TVX> ();
 
-    integer maxfct = 1, mnum = 1, phase = 33, msglevel = 0, error;
+    integer maxfct = 1, mnum = 1, phase = 33, msglevel = 0, error = 0;
     integer nrhs = fx.Size() / (height/entrysize);
     
     if (fx.Size() != fy.Size()) 
@@ -413,39 +502,67 @@ namespace ngla
 
     integer * params = const_cast <integer*> (&hparams[0]);
 
-    if (compressed)
+    if(task_manager)
+      task_manager->SuspendWorkers(1000);
+
+#ifdef USE_MKL
+    mkl_set_num_threads(mkl_max_threads);
+#endif // USE_MKL
+
+    if (matrix.Size() > 0)
       {
-	Matrix<TVX> hx(nrhs, compress.Size());
-	Matrix<TVX> hy(nrhs, compress.Size());
-	hx = mx.Cols(compress);
-	F77_FUNC(pardiso) ( const_cast<integer*>(pt), &maxfct, &mnum, 
-			    const_cast<integer*>(&matrixtype),
-			    &phase, const_cast<integer*>(&compressed_height), 
-			    reinterpret_cast<double *>(&matrix[0]),
-			    &rowstart[0], &indices[0],
-			    NULL, &nrhs, params, &msglevel,
-			    reinterpret_cast<double *>(&hx(0,0)), 
-			    reinterpret_cast<double *>(&hy(0,0)), &error );
-	my = 0; 
-	my.Cols(compress) = hy;
-      }
-    else
-      {
-	F77_FUNC(pardiso) ( const_cast<integer *>(pt), &maxfct, &mnum, 
-			    const_cast<integer *>(&matrixtype),
-			    &phase, const_cast<integer *>(&compressed_height), 
-			    reinterpret_cast<double *>(&matrix[0]),
-			    &rowstart[0], &indices[0],
-			    NULL, &nrhs, params, &msglevel,
-			    static_cast<double *>(fx.Data()), 
-			    static_cast<double *>(fy.Data()), &error );
+        if (compressed)
+          {
+            Matrix<TVX> hx(nrhs, compress.Size());
+            Matrix<TVX> hy(nrhs, compress.Size());
+            hx = mx.Cols(compress);
+            F77_FUNC(pardiso) ( const_cast<integer*>(pt), &maxfct, &mnum, 
+                                const_cast<integer*>(&matrixtype),
+                                &phase, const_cast<integer*>(&compressed_height), 
+                                reinterpret_cast<double *>(matrix.Data()),
+                                rowstart.Data(), indices.Data(),
+                                NULL, &nrhs, params, &msglevel,
+                                reinterpret_cast<double *>(hx.Data()), 
+                                reinterpret_cast<double *>(hy.Data()), &error );
+            my = 0; 
+            my.Cols(compress) = hy;
+          }
+        else
+          {
+            F77_FUNC(pardiso) ( const_cast<integer *>(pt), &maxfct, &mnum, 
+                                const_cast<integer *>(&matrixtype),
+                                &phase, const_cast<integer *>(&compressed_height), 
+                                reinterpret_cast<double *>(matrix.Data()),
+                                rowstart.Data(), indices.Data(),
+                                NULL, &nrhs, params, &msglevel,
+                                static_cast<double *>((void*)fx.Data()), 
+                                static_cast<double *>((void*)fy.Data()), &error );
+          }
       }
 
+#ifdef USE_MKL
+    mkl_set_num_threads(1);
+#endif // USE_MKL
+
+    if(task_manager)
+      task_manager->ResumeWorkers();
+
     if ( error != 0 )
       cout << "Apply Inverse: PARDISO returned error " << error << "!" << endl;
   }
   
   
+  template<class TM, class TV_ROW, class TV_COL>
+  void PardisoInverse<TM,TV_ROW,TV_COL> ::
+  MultTrans (const BaseVector & x, BaseVector & y) const
+  {
+    const_cast<integer&>(hparams[11]) = 2; // Solve transposed matrix
+
+    // See https://software.intel.com/content/www/us/en/develop/documentation/onemkl-developer-reference-c/top/sparse-solver-routines/onemkl-pardiso-parallel-direct-sparse-solver-interface/pardiso-iparm-parameter.html
+    
+    Mult(x,y);
+    const_cast<integer&>(hparams[11]) = 0;
+  }
 
 
 
@@ -454,7 +571,7 @@ namespace ngla
   void PardisoInverse<double,Complex,Complex> :: 
   Mult (const BaseVector & x, BaseVector & y) const
   {
-    static Timer timer ("Pardiso Solve");
+    static Timer timer ("Pardiso Solve, mat = d, vec = c");
     RegionTimer reg (timer);
 
 
@@ -462,6 +579,9 @@ namespace ngla
     Matrix<> tx(2, compressed_height);
     if (compressed)
       tx = Trans (hx.Rows(compress));
+      // tx = Trans (hx).Cols(compress));
+    // for (int i = 0; i < compress.Size(); i++)
+    // tx.Col(i) = hx.Row(compress[i]);
     else
       tx = Trans (hx);
 
@@ -469,18 +589,32 @@ namespace ngla
     Matrix<> ty(2, compressed_height);
 
       
-    integer maxfct = 1, mnum = 1, phase = 33, msglevel = 0, error;
+    integer maxfct = 1, mnum = 1, phase = 33, msglevel = 0, error = 0;
     integer nrhs = 2;
 
     integer * params = const_cast <integer*> (&hparams[0]);
 
-    F77_FUNC(pardiso) ( const_cast<integer *>(pt), 
-			&maxfct, &mnum, const_cast<integer *>(&matrixtype),
-			&phase, const_cast<integer *>(&compressed_height), 
-			reinterpret_cast<double *>(&matrix[0]),
-			&rowstart[0], &indices[0],
-			NULL, &nrhs, params, &msglevel, &tx(0,0), &ty(0,0),
-			&error );
+    if(task_manager)
+      task_manager->SuspendWorkers(1000);
+
+#ifdef USE_MKL
+    mkl_set_num_threads(mkl_max_threads);
+#endif // USE_MKL
+
+    if (matrix.Size() > 0)
+      F77_FUNC(pardiso) ( const_cast<integer *>(pt), 
+                          &maxfct, &mnum, const_cast<integer *>(&matrixtype),
+                          &phase, const_cast<integer *>(&compressed_height), 
+                          reinterpret_cast<double *>(matrix.Data()),
+                          rowstart.Data(), indices.Data(),
+                          NULL, &nrhs, params, &msglevel, tx.Data(), ty.Data(),
+                          &error );
+#ifdef USE_MKL
+      mkl_set_num_threads(1);
+#endif // USE_MKL
+
+    if(task_manager)
+      task_manager->ResumeWorkers();
 
     if ( error != 0 )
       cout << "Apply Inverse: PARDISO returned error " << error << "!" << endl;
@@ -516,9 +650,16 @@ namespace ngla
     integer * params = const_cast <integer*> (&hparams[0]);
 
     //    cout << "call pardiso (clean up) ..." << endl;
+    if (task_manager) task_manager -> StopWorkers();
     F77_FUNC(pardiso) ( pt, &maxfct, &mnum, &matrixtype, &phase, &compressed_height, NULL,
-			&rowstart[0], &indices[0], NULL, &nrhs, params, &msglevel,
+			rowstart.Data(), indices.Data(), NULL, &nrhs, params, &msglevel,
 			NULL, NULL, &error );
+#ifdef MKL_PARDISO
+    mkl_free_buffers();
+#endif // MKL_PARDISO
+    GetMemoryTracer().Free(memory_allocated_in_pardiso_lib);
+    memory_allocated_in_pardiso_lib = 0;
+    if (task_manager) task_manager -> StartWorkers();
     if (error != 0)
       cout << "Clean Up: PARDISO returned error " << error << "!" << endl;
   }
@@ -531,7 +672,8 @@ namespace ngla
   template<class TM>
   void PardisoInverseTM<TM> :: SetMatrixType() // TM entry)
   {
-    if (mat_traits<TM>::IS_COMPLEX)
+    // if (mat_traits<TM>::IS_COMPLEX)
+    if (ngbla::IsComplex<TM>())
       {
 	if ( symmetric ) 
 	  matrixtype = 6;   // complex symmetric
@@ -553,7 +695,7 @@ namespace ngla
 
     if (print)
       cout << "spd = " << int(spd) << ", sym = " << int(symmetric) 
-	   << ", complex = " << int(mat_traits<TM>::IS_COMPLEX)
+	   << ", complex = " << int(ngbla::IsComplex<TM>())
 	   << ", matrixtype = " << matrixtype << endl;
     *testout << "pardiso matrixtype = " << matrixtype << endl;
   }
@@ -681,7 +823,3 @@ namespace ngla
 
 }
 
-
-
-#endif
-
