@@ -12,7 +12,6 @@
 
 namespace ngmg
 {
-  using namespace ngmg;
 
   Prolongation :: Prolongation()
   {
@@ -24,25 +23,153 @@ namespace ngmg
     ;
   }
 
-
-
-
-  SparseMatrix< double >* LinearProlongation :: CreateProlongationMatrix( int finelevel ) const
+  void Prolongation :: Update (const FESpace & fes)
   {
-    int i, nc, nf;
-    int parents[2];
-    Array< int > indicesPerRow( space.GetNDof() );
+    if (leveldofs.Size() < fes.GetMeshAccess()->GetNLevels())
+      leveldofs.Append (DofRange(fes.GetNDof(), fes.GetParallelDofs()));
+  }
 
-    if ( &( space.LowOrderFESpace() ) )
+  
+  DofRange Prolongation :: LevelDofs (int level) const
+  {
+    if (level < 0 || level >= leveldofs.Size())
+      throw Exception("Illegal level " + ToString(level) + " num levels = " + ToString(leveldofs.Size()));
+    return leveldofs[level];
+  }
+
+
+  LinearProlongation :: ~LinearProlongation() { ; }
+
+  
+  void LinearProlongation :: Update (const FESpace & fes)
+  {
+    /*
+      if (ma->GetNLevels() > nvlevel.Size())
+      nvlevel.Append (ma->GetNV());
+    */
+    Prolongation::Update(fes);
+    
+    nvlevel.SetSize(ma->GetNLevels());
+    for (auto i : Range(nvlevel))
+      nvlevel[i] = ma->GetNVLevel(i);
+    
+    if (nvlevel.Size() > 1)
       {
-        nf = space.LowOrderFESpace().GetNDofLevel( finelevel );
-        nc = space.LowOrderFESpace().GetNDofLevel( finelevel-1 );
+        // if we have a transitive dependency within one level
+        // we cannot trivially prolongate in parallel
+        int finelevel = nvlevel.Size()-1;
+        size_t nc = nvlevel[finelevel-1];
+        size_t nf = nvlevel[finelevel];
+        auto & mesh = *ma;        
+        ParallelFor (IntRange(nc, nf), [&mesh, nc, this] (size_t i)
+                     {
+                       auto parents = mesh.GetParentNodes (i);
+                       if (parents[0] >= nc || parents[1] >= nc)
+                         this->allow_parallel = false;
+                     });
+        
+      }
+  }
+
+  
+  void LinearProlongation :: ProlongateInline (int finelevel, BaseVector & v) const
+  {
+    static Timer t("Prolongate"); RegionTimer r(t);
+    size_t nc = nvlevel[finelevel-1];
+    size_t nf = nvlevel[finelevel];
+    
+    if (v.EntrySize() == 1)
+      {
+        FlatVector<> fv = v.FV<double>();        
+        fv.Range (nf, fv.Size()) = 0;
+        if (allow_parallel)
+          {
+            auto & mesh = *ma;
+            ParallelFor (IntRange(nc, nf), [fv, &mesh] (size_t i)
+                         {
+                           auto parents = mesh.GetParentNodes (i);
+                           fv(i) = 0.5 * (fv(parents[0]) + fv(parents[1]));
+                         });
+          }
+        else
+          for (size_t i = nc; i < nf; i++)
+            {
+              auto parents = ma->GetParentNodes (i);
+              fv(i) = 0.5 * (fv(parents[0]) + fv(parents[1]));
+            }
       }
     else
       {
-        nf = space.GetNDofLevel( finelevel );
-        nc = space.GetNDofLevel( finelevel-1 );
+        FlatSysVector<> sv = v.SV<double>();
+        sv.Range (nf, sv.Size()) = 0;
+        for (size_t i = nc; i < nf; i++)
+          {
+            auto parents = ma->GetParentNodes (i);
+            sv(i) = 0.5 * (sv(parents[0]) + sv(parents[1]));
+          }
       }
+  }
+
+
+  void LinearProlongation :: RestrictInline (int finelevel, BaseVector & v) const 
+  {
+    static Timer t("Restrict"); RegionTimer r(t);
+      
+    size_t nc = nvlevel[finelevel-1];
+    size_t nf = nvlevel[finelevel];
+
+
+    if (v.EntrySize() == 1)
+      {
+	FlatVector<> fv = v.FV<double>();
+	for (size_t i = nf; i-- > nc; )
+	  {
+	    auto parents = ma->GetParentNodes (i);
+	    fv(parents[0]) += 0.5 * fv(i);
+	    fv(parents[1]) += 0.5 * fv(i);
+	  }
+	fv.Range(nc, fv.Size()) = 0;          
+      }
+    else
+      {
+	FlatSysVector<> fv = v.SV<double>();
+	for (size_t i = nf; i-- > nc; )
+	  {
+	    auto parents = ma->GetParentNodes (i);
+	    fv(parents[0]) += 0.5 * fv(i);
+	    fv(parents[1]) += 0.5 * fv(i);
+	  }
+	fv.Range(nc, fv.Size()) = 0;
+      }
+    /*
+      for (size_t i = nf; i < fv.Size(); i++)
+      fv(i) = 0;  
+    */
+  }
+
+
+  shared_ptr<SparseMatrix< double >> LinearProlongation :: CreateProlongationMatrix( int finelevel ) const
+  {
+    int i;
+    int parents[2];
+
+    int nc = nvlevel[finelevel-1];
+    int nf = nvlevel[finelevel];
+
+    Array< int > indicesPerRow(nf);
+
+    /*
+      if (space.LowOrderFESpacePtr())
+      {
+      nf = space.LowOrderFESpacePtr()->GetNDofLevel( finelevel );
+      nc = space.LowOrderFESpacePtr()->GetNDofLevel( finelevel-1 );
+      }
+      else
+      {
+      nf = space.GetNDofLevel( finelevel );
+      nc = space.GetNDofLevel( finelevel-1 );
+      }
+    */
     
     // count entries per row
     indicesPerRow = 0;
@@ -50,7 +177,7 @@ namespace ngmg
       indicesPerRow[ i ]++;
     for( i=nc; i<nf; i++ )
       {
-        ma.GetParentNodes( i, parents );
+        ma->GetParentNodes( i, parents );
         if ( parents[ 0 ] != -1 )
           indicesPerRow[ i ]++;
         if ( parents[ 1 ] != -1 )
@@ -63,7 +190,7 @@ namespace ngmg
       mg.CreatePosition( i, i );
     for( i=nc; i<nf; i++ )
       {
-        ma.GetParentNodes( i, parents );
+        ma->GetParentNodes( i, parents );
         if ( parents[ 0 ] != -1 )
           mg.CreatePosition( i, parents[ 0 ] ); 
         if ( parents[ 1 ] != -1 )
@@ -71,12 +198,12 @@ namespace ngmg
       }
 
     // write prolongation matrix
-    SparseMatrix< double >* prol = new SparseMatrix< double >( mg, 1 );
+    shared_ptr<SparseMatrix< double >> prol = make_shared<SparseMatrix< double >> ( std::move(mg) );
     for( i=0; i<nc; i++ )
       (*prol)( i, i ) = 1;
     for( i=nc; i<nf; i++ )
       {
-        ma.GetParentNodes( i, parents );
+        ma->GetParentNodes( i, parents );
         if ( parents[ 0 ] != -1 )
           (*prol)( i, parents[ 0 ] ) = 0.5; 
         if ( parents[ 1 ] != -1 )
@@ -339,173 +466,230 @@ namespace ngmg
 
 
 
-  /*
-    ElementProlongation :: 
-    ElementProlongation(const MeshAccess & ama, const ElementFESpace & aspace)
-    : ma(ama), space(aspace)
-    {
-    ;
-    }
-
-    ElementProlongation :: ~ElementProlongation()
-    {
-    ;
-    }
-  
-    void ElementProlongation :: Update ()
-    {
-    ;
-    }
-
-    void ElementProlongation :: 
-    ProlongateInline (int finelevel, ngla::BaseVector & v) const
-    {
-    int i, j;
-    int parent;
-
-    int nc = space.GetNDofLevel (finelevel-1);
-    int nf = space.GetNDofLevel (finelevel);
-
-    BaseSystemVector & sv = dynamic_cast<BaseSystemVector&> (v);
-    int sdim = sv.SystemDim();
-
-    for (i = nc+1; i <= nf; i++)
-    {
-    parent = ma.GetParentElement (i);
-    for (j = 1; j <= sdim; j++)
-    sv.VElem(i, j) = sv.VElem(parent, j);
-    }
-    }
-
-    void ElementProlongation :: 
-    RestrictInline (int finelevel, ngla::BaseVector & v) const
-    {
-    int i, j;
-    int parent;
-
-    int nc = space.GetNDofLevel (finelevel-1);
-    int nf = space.GetNDofLevel (finelevel);
-
-    BaseSystemVector & sv = dynamic_cast<BaseSystemVector&> (v);
-    int sdim = sv.SystemDim();
-
-    for (i = nf; i > nc; i--)
-    {
-    parent = ma.GetParentElement (i);
-    for (j = 1; j <= sdim; j++)
-    {
-    sv.VElem(parent, j) += sv.VElem(i, j);
-    sv.VElem(i, j) = 0;
-    }
-    }
-    }
-  */
-
-
-
-
-
-
-
-
-
-  SurfaceElementProlongation :: 
-  SurfaceElementProlongation(const MeshAccess & ama, 
-			     const SurfaceElementFESpace & aspace)
-    : ma(ama), space(aspace)
+  ElementProlongation ::
+  ElementProlongation(const FESpace & aspace, VorB avb)
+    : ma(aspace.GetMeshAccess()), space(aspace), vb(avb)
   {
-    ;
+    if (space.GetOrder() != 0)
+      throw Exception("ElementProlongation needs space of order=0");
   }
 
-  SurfaceElementProlongation :: ~SurfaceElementProlongation()
+
+  ElementProlongation :: ~ElementProlongation() { }
+
+  void ElementProlongation ::Update (const FESpace & fes) 
   {
     ;
   }
   
-  void SurfaceElementProlongation :: Update ()
+  shared_ptr<SparseMatrix< double >> ElementProlongation ::
+  CreateProlongationMatrix( int finelevel ) const 
   {
-    ;
+    return nullptr;
   }
-
-  void SurfaceElementProlongation :: 
-  ProlongateInline (int finelevel, ngla::BaseVector & v) const
+  
+  ///
+  void ElementProlongation ::
+  ProlongateInline (int finelevel, BaseVector & v) const 
   {
-    cout << "SurfaceElementProlongation not implemented" << endl;
-    /*
-    //  (*testout) << "SurfaceElementProlongation" << endl;
-    //  (*testout) << "before: " << v << endl;
-    int i, j;
-    int parent;
-
-    int nc = space.GetNDofLevel (finelevel-1);
-    int nf = space.GetNDofLevel (finelevel);
-
-    BaseSystemVector & sv = dynamic_cast<BaseSystemVector&> (v);
-    int sdim = sv.SystemDim();
-
-
-    for (i = nc+1; i <= nf; i++)
-    {
-    parent = ma.GetParentSElement (i);
-    for (j = 1; j <= sdim; j++)
-    sv.VElem(i, j) = sv.VElem(parent, j);
-    }
-    */
-    //  (*testout) << "after: " << v << endl;
-  }
-
-  void SurfaceElementProlongation :: 
-  RestrictInline (int finelevel, ngla::BaseVector & v) const
-  {
-    /*
-      int i, j;
-      int parent;
-
-      int nc = space.GetNDofLevel (finelevel-1);
-      int nf = space.GetNDofLevel (finelevel);
-
-      BaseSystemVector & sv = dynamic_cast<BaseSystemVector&> (v);
-      int sdim = sv.SystemDim();
-
-      for (i = nf; i > nc; i--)
+    FlatSysVector<> fv = v.SV();
+    
+    size_t nc = space.GetNDofLevel (finelevel-1);
+    size_t nf = space.GetNDofLevel (finelevel);
+    
+    for (size_t i = nc; i < nf; i++)
       {
-      parent = ma.GetParentSElement (i);
-      for (j = 1; j <= sdim; j++)
-      {
-      sv.VElem(parent, j) += sv.VElem(i, j);
-      sv.VElem(i, j) = 0;
-      }
-      }
-    */
+	  auto parent = ma->GetParentElement (ElementId(vb,i));
+	  fv(i) = fv(parent.Nr());
+	}
+    
+    fv.Range (nf, fv.Size()) = 0;
   }
-
-
-
-
-
-  L2HoProlongation::
-  L2HoProlongation(const MeshAccess & ama, const Array<int> & afirst_dofs)
-    : ma(ama), first_dofs(afirst_dofs) 
-  { ; }
   
-  void L2HoProlongation::ProlongateInline (int finelevel, BaseVector & v) const
-  {   
+  void ElementProlongation ::
+  RestrictInline (int finelevel, BaseVector & v) const 
+  {
+    FlatSysVector<> fv = v.SV(); 
+    
+    size_t nc = space.GetNDofLevel (finelevel-1);
+    size_t nf = space.GetNDofLevel (finelevel);
+    
+    for (int i = nf-1; i >= nc; i--)
+      {
+        auto parent = ma->GetParentElement (ElementId(vb,i));
+        fv(parent.Nr()) += fv(i);
+        fv(i) = 0;
+      }
+  }
+  
+
+
+
+
+
+
+  void EdgeProlongation :: ProlongateInline (int finelevel, BaseVector & v) const
+  {
+    size_t nc = space.GetNDofLevel (finelevel-1);
+    size_t nf = space.GetNDofLevel (finelevel);
+    
+    if (v.EntrySize() == 1)
+      {
+	auto fv = v.FV<double>();
+	
+	for (size_t i = nf; i < fv.Size(); i++)
+	  fv(i) = 0;
+	
+	for (int k = 1; k <= 4; k++) // should not be necessary if new edges are always after old edges
+	  for (size_t i = nc; i < nf; i++)
+	    {
+	      int pa1 = space.ParentEdge1 (i);
+	      int pa2 = space.ParentEdge2 (i);
+	      
+	      fv(i) = 0;
+	      if (pa1 != -1)
+		{
+		  if (pa1 & 1)
+		    fv(i) += 0.5 * fv(pa1/2);
+		  else
+		    fv(i) -= 0.5 * fv(pa1/2);
+		}
+	      if (pa2 != -1)
+		{
+		  if (pa2 & 1)
+		    fv(i) += 0.5 * fv(pa2/2);
+		  else
+		    fv(i) -= 0.5 * fv(pa2/2);
+		}
+	    }
+
+	for (size_t i = 0; i < nf; i++)
+	  if (space.FineLevelOfEdge(i) < finelevel)
+	    fv(i) = 0;
+
+	return;
+      }
+	
+
+    
     FlatSysVector<> fv (v.Size(), v.EntrySize(), static_cast<double*>(v.Memory()));
-    
-    int ne = ma.GetNE();
-    int ndel = first_dofs[1];
-    
-    for (int i = 0; i <ne; i++)
-      {
-        int parent = ma.GetParentElement (i);
-        if(parent!=-1)
-          fv(ndel*i) = fv(ndel*parent);
-        for(int j = 1; j<ndel; j++)
-          fv(ndel*i+j) = 0;
-      }
+
+    int i, k;
+
+    for (i = nf; i < fv.Size(); i++)
+      fv(i) = 0;
+
+    for (k = 1; k <= 10; k++)
+      for (i = nc; i < nf; i++)
+	{
+	  int pa1 = space.ParentEdge1 (i);
+	  int pa2 = space.ParentEdge2 (i);
+	  
+	  fv(i) = 0;
+	  if (pa1 != -1)
+	    {
+	      if (pa1 & 1)
+		fv(i) += 0.5 * fv(pa1/2);
+	      else
+		fv(i) -= 0.5 * fv(pa1/2);
+	    }
+	  if (pa2 != -1)
+	    {
+	      if (pa2 & 1)
+		fv(i) += 0.5 * fv(pa2/2);
+	      else
+		fv(i) -= 0.5 * fv(pa2/2);
+	    }
+	}
+
+    for (i = 0; i < nf; i++)
+      if (space.FineLevelOfEdge(i) < finelevel)
+	fv(i) = 0;
   }
 
 
+  ///
+  void EdgeProlongation :: RestrictInline (int finelevel, BaseVector & v) const
+  {
+    size_t nc = space.GetNDofLevel (finelevel-1);
+    size_t nf = space.GetNDofLevel (finelevel);
+
+
+    if (v.EntrySize() == 1)
+      {
+	auto fv = v.FV<double>();
+
+	for (int i = 0; i < nf; i++)
+	  if (space.FineLevelOfEdge(i) < finelevel)
+	    fv(i) = 0;
+	
+	for (int k = 1; k <= 5; k++)
+	  for (int i = nf-1; i >= nc; i--)
+	    {
+	      int pa1 = space.ParentEdge1 (i);
+	      int pa2 = space.ParentEdge2 (i);
+	      
+	      if (pa1 != -1)
+		{
+		  if (pa1 & 1)
+		    fv(pa1/2) += 0.5 * fv(i);
+		  else
+		    fv(pa1/2) -= 0.5 * fv(i);
+		}
+	      if (pa2 != -1)
+		{
+		  if (pa2 & 1)
+		    fv(pa2/2) += 0.5 * fv(i);
+		  else
+		    fv(pa2/2) -= 0.5 * fv(i);
+		}
+	      fv(i) = 0;
+	    }
+	
+	for (int i = nf; i < fv.Size(); i++)
+	  fv(i) = 0;
+	return;
+      }
+    
+
+    FlatSysVector<> fv (v.Size(), v.EntrySize(), static_cast<double*>(v.Memory()));
+
+    for (int i = 0; i < nf; i++)
+      if (space.FineLevelOfEdge(i) < finelevel)
+	fv(i) = 0;
+	
+    for (int k = 1; k <= 10; k++)
+      for (int i = nf-1; i >= nc; i--)
+	{
+	  int pa1 = space.ParentEdge1 (i);
+	  int pa2 = space.ParentEdge2 (i);
+	  
+	  if (pa1 != -1)
+	    {
+	      if (pa1 & 1)
+		fv(pa1/2) += 0.5 * fv(i);
+	      else
+		fv(pa1/2) -= 0.5 * fv(i);
+	    }
+	  if (pa2 != -1)
+	    {
+	      if (pa2 & 1)
+		fv(pa2/2) += 0.5 * fv(i);
+	      else
+		fv(pa2/2) -= 0.5 * fv(i);
+	    }
+	  fv(i) = 0;
+	}
+
+    for (int i = nf; i < fv.Size(); i++)
+      fv(i) = 0;  
+  }
+
+
+
+
+
+  
 
   CompoundProlongation :: 
   CompoundProlongation(const CompoundFESpace * aspace)
@@ -513,19 +697,58 @@ namespace ngmg
 
   CompoundProlongation :: 
   CompoundProlongation(const CompoundFESpace * aspace,
-		       Array<const Prolongation*> & aprols)
+		       Array<shared_ptr<Prolongation>> & aprols)
     : space(aspace), prols(aprols) { ; }
 
   CompoundProlongation :: ~CompoundProlongation() { ; }
   
 
-  void CompoundProlongation :: Update ()
+  void CompoundProlongation :: Update (const FESpace & fespace)
   {
+    auto & cfes = dynamic_cast<const CompoundFESpace&> (fespace);
     for (int i = 0; i < prols.Size(); i++)
       if (prols[i])
-	const_cast<Prolongation*>(prols[i]) -> Update();
+	prols[i] -> Update(*cfes[i]);
   }
 
+  size_t CompoundProlongation :: GetNDofLevel (int level)
+  {
+    size_t nd = 0;
+    for (int i = 0; i < prols.Size(); i++)
+      if (prols[i])
+	nd += prols[i] -> GetNDofLevel(level);
+    return nd;
+  }
+  
+  shared_ptr<BitArray> CompoundProlongation :: GetInnerDofs (int finelevel) const
+  {
+    Array<shared_ptr<BitArray>> comp_inner;
+    bool has_inner = false;
+    for (auto prol : prols)
+      {
+        if (prol)
+          comp_inner.Append (prol->GetInnerDofs(finelevel));
+        else
+          comp_inner.Append (nullptr);
+        if (comp_inner.Last()) has_inner = true;
+      }
+    if (!has_inner) return nullptr;
+    
+    BitArray inner(space->GetNDofLevel(finelevel));
+    inner.Clear();
+    for (int i : Range(prols))
+      {
+        IntRange r = space->GetRange(i);
+        if (comp_inner[i])
+          for (int j : Range(r))
+            if (comp_inner[i]->Test(j))
+              inner.SetBit(r.First()+j);
+      }
+    return make_shared<BitArray>(inner);
+  }
+
+  
+
   void CompoundProlongation :: 
   ProlongateInline (int finelevel, BaseVector & v) const
   {
