@@ -43,15 +43,14 @@ namespace ngmg
   GSSmoother :: 
   GSSmoother  (const MeshAccess & ama,
 	       const BilinearForm & abiform)
-    : ma(ama), biform(abiform)
+    : /* ma(ama), */ biform(abiform)
   {
     Update();
   }
 
   GSSmoother :: ~GSSmoother()
   {
-    for (int i = 0; i < jac.Size(); i++)
-      delete jac[i];
+    ; // for (int i = 0; i < jac.Size(); i++) delete jac[i];
   }
 
   void GSSmoother :: Update (bool force_update)
@@ -60,9 +59,13 @@ namespace ngmg
     jac.SetSize (biform.GetNLevels());
     for (i = 0; i < biform.GetNLevels(); i++)
       {
-	if (&biform.GetMatrix(i))
-	  jac[i] = dynamic_cast<const BaseSparseMatrix&> (biform.GetMatrix(i))
-	    .CreateJacobiPrecond(biform.GetFESpace().GetFreeDofs());
+	if (biform.GetMatrixPtr(i))
+          {
+            jac[i] = dynamic_cast<const BaseSparseMatrix&> (*biform.GetMatrixPtr(i))
+              .CreateJacobiPrecond(biform.GetFESpace()->GetFreeDofs());
+            string name = "GSSmootherLevel" + ToString(i);
+            GetMemoryTracer().Track(*jac[i], name);
+          }
 	else
 	  jac[i] = NULL;
       }
@@ -142,9 +145,9 @@ namespace ngmg
     d = f - biform.GetMatrix(level) * u;
   }
   
-  BaseVector * GSSmoother :: CreateVector(int level) const
+  AutoVector GSSmoother :: CreateVector(int level) const
   {
-    return biform.GetMatrix(level).CreateVector();
+    return biform.GetMatrix(level).CreateColVector();
   }
 
 
@@ -162,8 +165,7 @@ namespace ngmg
 
   AnisotropicSmoother :: ~AnisotropicSmoother()
   {
-    for (int i = 0; i < jac.Size(); i++)
-      delete jac[i];
+    ; // for (int i = 0; i < jac.Size(); i++) delete jac[i];
   }
   
   void AnisotropicSmoother :: Update (bool force_update)
@@ -181,17 +183,19 @@ namespace ngmg
     for (j = 0; j < ma.GetNP(); j++)
       cnts[ma.GetClusterRepVertex(j)]++;
 
-    Table<int> * linecluster = new Table<int>(cnts);
+    Table<int> linecluster(cnts);
     
     cnts = 0;
     for (j = 0; j < ma.GetNP(); j++)
       {
 	int cl = ma.GetClusterRepVertex(j);
-	(linecluster)[cl][cnts[cl]] = j;
+	linecluster[cl][cnts[cl]] = j;
 	cnts[cl]++;
       }
     jac.Append  (dynamic_cast<const BaseSparseMatrix&> (biform.GetMatrix()) . 
-		 CreateBlockJacobiPrecond(*linecluster));
+		 CreateBlockJacobiPrecond(make_shared<Table<int>> (linecluster)));
+    string name = "AnisotropicSmootherLevel" + ToString(jac.Size()-1);
+    GetMemoryTracer().Track(*jac.Last(), name);
   }
 
 
@@ -221,10 +225,10 @@ namespace ngmg
     //    biform.GetMatrix (level).Residuum (u, f, d);
   }
   
-  BaseVector * AnisotropicSmoother :: 
+  AutoVector AnisotropicSmoother :: 
   CreateVector(int level) const
   {
-    return biform.GetMatrix(level).CreateVector();
+    return biform.GetMatrix(level).CreateColVector();
   }
 
 
@@ -676,7 +680,7 @@ namespace ngmg
   BlockSmoother :: 
   BlockSmoother  (const MeshAccess & ama,
 		  const BilinearForm & abiform, const Flags & aflags)
-    : Smoother(aflags), ma(ama), biform(abiform), constraint(NULL), direct(NULL)
+    : Smoother(aflags), biform(abiform), constraint(nullptr), direct(nullptr)
   {
     Update();
   }
@@ -685,32 +689,23 @@ namespace ngmg
   BlockSmoother  (const MeshAccess & ama,
 		  const BilinearForm & abiform,
 		  const LinearForm & aconstraint, const Flags & aflags)
-    : Smoother(aflags), ma(ama), biform(abiform), constraint(&aconstraint), direct(NULL)
+    : Smoother(aflags), /* ma(ama), */ biform(abiform), constraint(&aconstraint), direct(nullptr)
   {
     Update();
   }
 
   BlockSmoother :: ~BlockSmoother()
-  {
-    for (int i = 0; i < jac.Size(); i++)
-      delete jac[i];
-    for (int i = 0; i < inv.Size(); i++)
-      delete inv[i];
-    delete direct;
-  }
+  { ; }
   
   void BlockSmoother :: Update (bool force_update)
   {
     int level = biform.GetNLevels();
    
     if (level < 0) return;
-
     if(updateall)
       {
-	for (int i = 0; i < jac.Size(); i++)
-	  delete jac[i];
-	for (int i = 0; i < inv.Size(); i++)
-	  delete inv[i];
+	// for (int i = 0; i < jac.Size(); i++) delete jac[i];
+	// for (int i = 0; i < inv.Size(); i++) delete inv[i];
 
 	jac.DeleteAll();
 	inv.DeleteAll();
@@ -720,26 +715,38 @@ namespace ngmg
     
     if (biform.UsesEliminateInternal())
       flags.SetFlag("eliminate_internal");
-    Table<int> * it = biform.GetFESpace().CreateSmoothingBlocks(flags);
+
+    while(smoothing_blocks.Size() < level)
+      smoothing_blocks.Append(nullptr);
+
+    if (!smoothing_blocks.Last())
+      smoothing_blocks.Last() = biform.GetFESpace()->CreateSmoothingBlocks(flags);
 
 
     while (jac.Size() < level)
-      jac.Append(NULL);
+      jac.Append(nullptr);
 
 #ifndef PARALLELxxx
-    if (!constraint)
+    int startlevel = updateall ? 1 : level;
+    for(auto lvl : Range(startlevel,level+1))
       {
-	delete jac[level-1];
-	jac[level-1] = dynamic_cast<const BaseSparseMatrix&> 
-	  (biform.GetMatrix()).CreateBlockJacobiPrecond(*it);
-      }
-    else
-      {
-	jac[level-1] = dynamic_cast<const BaseSparseMatrix&> 
-	  (biform.GetMatrix()).CreateBlockJacobiPrecond(*it, &constraint->GetVector());
+        if (!constraint)
+          {
+            jac[lvl-1] = dynamic_cast<const BaseSparseMatrix&>
+              (biform.GetMatrix(lvl-1)).CreateBlockJacobiPrecond(smoothing_blocks[lvl-1]);
+          }
+        else
+          {
+            jac[lvl-1] = dynamic_cast<const BaseSparseMatrix&>
+              (biform.GetMatrix(lvl-1)).CreateBlockJacobiPrecond(smoothing_blocks[lvl-1], &constraint->GetVector());
+          }
+        string name = "BlockSmootherLevel" + ToString(lvl);
+        GetMemoryTracer().Track(*jac[lvl-1], name);
       }
 #else
 
+    if(updateall)
+      throw Exception("Not working with updateall!");
     bool isparallel = false;
     if ( id >= 1 ) isparallel = true;
 
@@ -764,17 +771,15 @@ namespace ngmg
 	if (!constraint)
 	  jac[level-1] = dynamic_cast<const BaseSparseMatrix&> 
 	    (biform.GetMatrix()).CreateBlockJacobiPrecond(*it);
+
 	else
-	  {
 	    jac[level-1] = dynamic_cast<const BaseSparseMatrix&> 
 	      (biform.GetMatrix()).CreateBlockJacobiPrecond(*it, &constraint->GetVector());
-	  }
-
       }
 #endif
 
     while (inv.Size() < level)
-      inv.Append(NULL);  
+      inv.Append(nullptr);
   
     //   BitArray * planedofs = biform.GetFESpace().CreateIntermediatePlanes();
     //   if (planedofs)
@@ -784,21 +789,26 @@ namespace ngmg
     //     }
     //   else
     //     {
-  
-    delete direct;
-    direct = biform.GetFESpace().CreateDirectSolverClusters(flags);
+
+    if(userdefined_direct)
+      direct = userdefined_direct;
+    else
+      direct = biform.GetFESpace()->CreateDirectSolverClusters(flags);
 
     if (direct)
       {
+        GetMemoryTracer().Track(*direct, "DirectSolverClusters");
 	if (biform.UsesEliminateInternal())
 	  {
-	    const FESpace & fes = biform.GetFESpace();
+	    const FESpace & fes = *biform.GetFESpace();
 	    for (int j = 0; j < direct->Size(); j++)
 	      if (fes.GetDofCouplingType(j) == LOCAL_DOF)
 		(*direct)[j] = 0;
 	  }
 	inv[level-1] = dynamic_cast<const BaseSparseMatrix&> 
 	  (biform.GetMatrix()).InverseMatrix (direct);
+        string name = "DirectSolverClustersInverse-Level" + ToString(level);
+        GetMemoryTracer().Track(*inv[level-1], name);
       }
   }
 
@@ -810,18 +820,16 @@ namespace ngmg
       jac[level] -> GSSmooth (u, f, steps);
     else
       {
-	BaseVector & d = *f.CreateVector();
-	BaseVector & w = *f.CreateVector();
+	auto d = f.CreateVector();
+	auto w = f.CreateVector();
 	for(int i=0;i<steps;i++)
 	  {
 	    jac[level] -> GSSmooth (u, f, 1); 
 	  
-	    d = f - biform.GetMatrix(level) * u;
-	    w = (*inv[level]) * d;
-	    u += w;
+	    *d = f - biform.GetMatrix(level) * u;
+	    *w = (*inv[level]) * *d;
+	    u += *w;
 	  }  
-	delete &w;
-	delete &d;
       }
   }
 
@@ -841,7 +849,7 @@ namespace ngmg
       }
     else
       {
-	SparseFactorization * scinv = dynamic_cast<SparseFactorization*> (inv[level]);
+	SparseFactorization * scinv = dynamic_cast<SparseFactorization*> (inv[level].get());
 	if (scinv)
 	  {
 	    for(int i=0;i<steps;i++)
@@ -881,31 +889,31 @@ namespace ngmg
       }
     else
       {
-	BaseVector & d = *f.CreateVector();
+	auto d = f.CreateVector();
 
-	SparseFactorization * scinv = dynamic_cast<SparseFactorization*> (inv[level]);
+	SparseFactorization * scinv = dynamic_cast<SparseFactorization*> (inv[level].get());
 	if (scinv)
 	  {
-	    d = f;
-	    biform.GetMatrix (level).MultAdd2 (-1, u, d);
+	    *d = f;
+	    biform.GetMatrix (level).MultAdd2 (-1, u, *d);
 	    for (int i = 0; i < steps; i++)
 	      {
 		if ( (!scinv->SmoothIsProjection()) || (i > 0) || (level > 0) )
-		  scinv -> Smooth (u, f, d);
-		jac[level] -> GSSmoothBackPartial (u, f, d); 
+		  scinv -> Smooth (u, f, *d);
+		jac[level] -> GSSmoothBackPartial (u, f, *d); 
 	      }
 	  }
 	else
 	  {
 	    for(int i=0;i<steps;i++)
 	      {
-		d = f - biform.GetMatrix(level) * u;
-		u += (*inv[level]) * d;
+		*d = f - biform.GetMatrix(level) * u;
+		u += (*inv[level]) * *d;
 		jac[level] -> GSSmoothBack (u, f, 1);
 	      }
 	  }
            
-	delete &d;
+	// delete &d;
       }
   }
   
@@ -922,15 +930,17 @@ namespace ngmg
     d = f - biform.GetMatrix (level) * u;
   }
   
-  BaseVector * BlockSmoother :: CreateVector(int level) const
+  AutoVector BlockSmoother :: CreateVector(int level) const
   {
-    return biform.GetMatrix(level).CreateVector();
+    return biform.GetMatrix(level).CreateColVector();
   }
 
-  void BlockSmoother :: MemoryUsage (Array<MemoryUsageStruct*> & mu) const
+  Array<MemoryUsage> BlockSmoother :: GetMemoryUsage () const
   {
+    Array<MemoryUsage> mu;
     for (int i = 0; i < jac.Size(); i++)
-      if (jac[i]) jac[i]->MemoryUsage (mu);
+      if (jac[i]) mu += jac[i]->GetMemoryUsage ();
+    return mu;
   }
 
 
@@ -942,20 +952,7 @@ namespace ngmg
 
 
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
+#ifdef XXX_OBSOLTE
   SmoothingPreconditioner :: 
   SmoothingPreconditioner (const Smoother & asmoother,
 			   int alevel)
@@ -975,11 +972,12 @@ namespace ngmg
     */
   }
   
-  BaseVector *   SmoothingPreconditioner :: CreateVector () const
+  AutoVector SmoothingPreconditioner :: CreateVector () const
   {
     //    return smoother.CreateVector(level);
-    return 0;
+    return shared_ptr<BaseVector>();
   }
-
+#endif
+  
 
 }
